/*
   OpenGUI - Drawing & Windowing library

   Copyright (C) 1996,2003  Marian Krivos

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   nezmar@atlas.sk

   base.h - base graphics support headers

*/

#ifndef __cplusplus
#error "OpenGUI 4.x and above supports only C++ compilers (not legacy C)"
#endif

#ifndef __BASE_H
#define __BASE_H

#ifdef __BORLANDC__
#pragma option push -b
#pragma option push -a4
#endif

#define FG_VERSION "4.1.3"
#define FG_MAJOR 	4
#define FG_MINOR 	1
#define FG_PATCH 	3

#include "internal.h"

#ifdef INDEX_COLORS
/**
	The unsigned integer that is equal in size with the one graphic pixel
	representation in VRAM, i.e. char for indexed colors, int for true colors (24 or 32 bit modes).
	@ingroup Types
*/
typedef unsigned char FGPixel;	// type of color (must correspond with bpp)
#endif

#if (DIRECT_COLORS==15) | (DIRECT_COLORS==16)
typedef unsigned short FGPixel;		// type of color (must correspond with bpp)
#endif

#ifdef TRUE_COLORS
typedef unsigned int FGPixel;	// type of color (must correspond with bpp)
#endif

typedef void (*blitter_t) (FGPixel *, FGPixel *, int, int, int, int);
typedef void (*blitter2_t) (FGPixel *, FGPixel *, int, int, int, int, unsigned);
typedef void (*blitter_a_t) (FGPixel *, FGPixel *, int, int, int, int, int, unsigned ppop);
typedef int  (*shapper_t) (int, int, int, int, FGPixel ink, unsigned ppop);
typedef void (*clipper_t) (int, int, int, int);
typedef void (*plotter_t) (FGPixel *ptr, int x, int y, int ww, FGPixel ink, unsigned ppop);
typedef void (*paletter_t) (unsigned, unsigned);
typedef int  (*mouser_t) (int, int, int, int&);

#define	CR		13
#define	ESC		27
#define	TAB		9
#define	LF		10

/**
	The ROPs. Intended for a drawing functions.
	@ingroup Enums
*/
enum ENUM_PPOP { _GSET, _GXOR, _GAND, _GOR, _GPLUS, _GMINUS, _GNOT, _GREPLACE, _GTRANSP, _GCOLORKEY } ;

/**
	Constants for apropriate videomodes.
	@ingroup Enums
*/
enum ENUM_VIDEOMODE {
	GTEXT,
	G320x200,
	G640x480,
	G800x600,
	G1024x768,
	G1280x1024,
	G1600x1200,
	GCUSTOM
};

#define	FG_BUTTON_LEFT		1
#define	FG_BUTTON_RIGHT		2
#define	FG_BUTTON_MIDDLE	4


/**
	Flags for use with DrawBuffer::bitblit()
	@addtogroup Defines
	@{
*/
#define BITBLIT_COPY			0x0000
#define BITBLIT_WITH_OPERATOR	0x0200
#define BITBLIT_WITH_ALPHA		0x0400
#define BITBLIT_WITH_CLIPPING	0x1000
#define BITBLIT_WITH_DECREMENT	0x2000

#ifdef INDEX_COLORS
//! Constant for the BLACK color
#define  CBLACK		((FGPixel) 0)
//! Constant for the DARK_GRAY color
#define  CDARK		((FGPixel) 1)
//! Constant for the another GRAY color
#define  CGRAYED	((FGPixel) 2)
//! Constant for the GRAY color #1
#define  CGRAY1		((FGPixel) 3)
//! Constant for the GRAY color #2
#define  CGRAY2		((FGPixel) 4)
//! Constant for the GRAY color #3
#define  CGRAY3		((FGPixel) 5)
//! Constant for the BLUE color
#define  CBLUE		((FGPixel) 6)
//! Constant for the LIGHT BLUE color
#define  CBLUELIGHT	((FGPixel) 7)
//! Constant for the GREEN color
#define  CGREEN		((FGPixel) 8)
//! Constant for the LIGHT GREEN color
#define  CGREENLIGHT ((FGPixel) 9)
//! Constant for the RED color
#define  CRED		((FGPixel) 10)
//! Constant for the LIGHT RED color
#define  CREDLIGHT	((FGPixel) 11)
//! Constant for the BROWN color
#define  CBROWN		((FGPixel) 12)
//! Constant for the YELLOW color
#define  CYELLOW	((FGPixel) 13)
//! Constant for the DUN WHITE color
#define  CWHITED	((FGPixel) 14)
//! Constant for the 100% WHITE color
#define  CWHITE		((FGPixel) 15)
#else
#define  CBLACK		((FGPixel) 0)
#define  CDARK		((FGPixel) FGDirectColor(0x444444U))
#define  CGRAYED	((FGPixel) FGDirectColor(0x585858U))
#define  CGRAY1		((FGPixel) FGDirectColor(0x7c8074U))
#define  CGRAY2		((FGPixel) FGDirectColor(0x687884U))
#define  CGRAY3		((FGPixel) FGDirectColor(0xa8b8acU))
#define  CBLUE		((FGPixel) FGDirectColor(0x000078U))
#define  CBLUELIGHT	((FGPixel) FGDirectColor(0x0000f8U))
#define  CGREEN		((FGPixel) FGDirectColor(0x007800U))
#define  CGREENLIGHT ((FGPixel) FGDirectColor(0x00f800U))
#define  CRED		((FGPixel) FGDirectColor(0xd44448U))
#define  CREDLIGHT	((FGPixel) FGDirectColor(0xf80000U))
#define  CBROWN		((FGPixel) FGDirectColor(0x684028U))
#define  CYELLOW	((FGPixel) FGDirectColor(0xfce000U))
#define  CWHITED	((FGPixel) FGDirectColor(0xd8d0b8U))
#define  CWHITE		((FGPixel) FGDirectColor(0xffffffU))
#endif // TRUE_COLORS

#ifdef INDEX_COLORS
/**
	Remap the indexed colors to the truecolor directly by palette color map.
	Useful macro when you port your 256 color modes app to the truecolors.
*/
#define FGCOLOR(i) ((FGPixel) i)
#else
#define FGCOLOR(i) ((FGPixel) FGDirectColor(_fgl_palette[i]))
#endif
/**
	@}
*/

extern unsigned int long _fgl_palette[256];

/**
	The all event types.
	@ingroup Enums
*/
enum ENUM_EVENTS  {
/**
a dummy event.
*/
NOEVENT,
/**
This events indicates key pressing (the keyboard input is going to the active window only).
System has some reserved key combination (ALT+X, CTRL+TAB etc), and these aren't possible
to testing. These reserved keys are converted to appropriate events (QUITEVENT, LOSTFOCUSEVENT & GETFOCUSEVENT).
<b>Used with:</b> Application and Window handlers.
*/
KEYEVENT,
/**
Internal event.
*/
MOUSEEVENT,
/**
The last event, that is sent only to the application procedure. This event can generated
in one of the two ways: by user pressing <ALT+X> or by calling AppDone().
<b>Used with:</b> Application handler only.
*/
QUITEVENT,
/**
This is a very important event that is generated by system when you activate
some Controls (Button, EditBox, ..). The member key contains the id
of current control. The callback handler of the object [you can assign
to each Control your callback that will be called at the right time] is
called after this event parsing into your procedure. The test of some controls:
<b>Used with:</b> Application and Window handlers.

	void WindowHandler(GuiEvent *p)
	{
		switch(p->GetType())
		{
			case ACCELEVENT:
			if (p->GetKey() == button1->GetId())
			{
				// some code
			}
		}
	}
*/
ACCELEVENT,
/**
Last event in the app life. You can delete your stuff after this, the window is deleted now.
<b>Used with:</b> Window handlers only.
*/
TERMINATEEVENT,
/**
Indicates any movement of the mouse (including button clicks). The members are
assigned as follows: If cursor is out of any window, member key is 0,
and items x & y contains the position of mouse cursor in screen co-ordination
([0,0] is top left corner). When the cursor is over some window, then event
is sent only to the active window. The key member is set to 0 if the window
is active, -1 when window with cursor is not active, and if the member key
is greater than 0, then value is an id of controls in window. The co-ordinations
for movement in case of window are relative to the window workspace (if cursor is
at the title of window, then y co-ordination will be negative!). For a good overview
of this, look at the events example.
<b>Used with:</b> Application and Window handlers.
*/
MOVEEVENT,
/**
This event is sent when a simple click on mouse's right button is detected. You must use window
flag WCLICKABLE to enable this feature. Position reported in the event
is in the screen co-ordinates, or relative to windows, according to placement.
<b>Used with:</b> Application and Window handlers.
*/
CLICKRIGHTEVENT,
/**
This event is sent when a simple click on mouse's left button is detected. You must use window
flag WCLICKABLE to enable this feature. Position reported in the event
is in the screen co-ordinates, or relative to windows, according to placement.
<b>Used with:</b> Application and Window handlers.
*/
CLICKLEFTEVENT,
/**
<b>Used with:</b> Application and Window handlers.
*/
DRAGLEFTEVENT,
/**
<b>Used with:</b> Application and Window handlers.
*/
DRAGRIGHTEVENT,
/**
	The window has got the input focus.
	<b>Used with:</b> Window handlers only.
*/
GETFOCUSEVENT,
/**
	The window lost the input focus.
	<b>Used with:</b> Window handlers only.
*/
LOSTFOCUSEVENT,
/**
The first event that is sent to the window when a window is created and it is drawn
at the screen. It is a right place for drawing controls and other stuff.
<b>Used with:</b> Window handlers only.
*/
INITEVENT,
/**
If you hold the mouse button at the Controls PUSHBUTTON a while,
system will repeatedly emit this event. See also the SetRepeatDelay() function.
<b>Used with:</b> Application and Window handlers.
*/
BUTTONHOLDEVENT,
/**
If the window is resize-able (switch WSIZEABLE), then you can by dragging
bottom right corner of windows, resize one. After each size change,
this event is generated. You may redraw window at this point (Controls
are draws automatically). The current size (of window workspace)
is available by calling GetWW() a GetHW().
<b>Used with:</b> Window handlers only.
*/
WINDOWRESIZEEVENT,
/**
The event is generated when window moves at new position.
<b>Used with:</b> Window handlers only.
*/
WINDOWMOVEEVENT,
/**
If you use switch APP_MAGNIFIER, you can select a rectangle area by
the mouse dragging. The members are assigned with [x,y] position
of top left corner and [w,h] the size of one. If you want exactly
first and end point position of dragging, you must use GetDragVector().
<b>Used with:</b> Application and Window handlers.
*/
STARTDRAGLEFTEVENT,
/**
If you use switch APP_MAGNIFIER, you can select a rectangle area by
the mouse dragging. The members are assigned with [x,y] position
of top left corner and [w,h] the size of one. If you want exactly
first and end point position of dragging, you must use GetDragVector().
<b>Used with:</b> Application and Window handlers.
*/
STARTDRAGRIGHTEVENT,
/**
The window is repainted. Place your drawing code here.
<b>Used with:</b> Window handlers only.
*/
REPAINTEVENT,
/**
<b>Used with:</b> Window handlers only.
*/
NOTIFYEVENT,
/**
<b>Used with:</b> Window handlers only.
*/
ICONIZEEVENT,
/**
<b>Used with:</b> Application handler only.
*/
CURSOROUTEVENT,
/**
This event is sent when a double click on mouse's left button is detected. You must use window
flag WCLICKABLE to enable this feature. Position reported in the event
is in the screen co-ordinates (app handler), or relative to windows, according to placement.
<b>Used with:</b> Application and Window handlers.
*/
DBLCLICKRIGHTEVENT,
/**
This event is sent when a double click on mouse's right button is detected. You must use window
flag WCLICKABLE to enable this feature. Position reported in the event
is in the screen co-ordinates (app handler), or relative to windows, according to placement.
<b>Used with:</b> Application and Window handlers.
*/
DBLCLICKLEFTEVENT,
//! NOTE> this must be at the end - used as terminator.
LASTEVENT
};

struct variable_record;
struct fonthdr;

/**
Internal structure to describe built-in bitmap fonts;
*/
struct FGFont
{
		char	name[24];
		int type, width, height;
		variable_record *var; // null or ptr
		FGPixel *fontimg;
	public:
		FGFont()
		{
			memset(this, 0, sizeof(*this));
		}
};

/**
The simple helper structure to store a rectangle's position and size.
*/
struct FGRect
{
	int x, y;
	unsigned int w, h;
	FGRect()
	{
		x=y=w=h=0;
	}
	FGRect(int xpos, int ypos, int width, int height)
	{
		x = xpos;
		y = ypos;
		w = width;
		h = height;
	}
};

#define MAX_FONTS		64

/**
	Predefined const for a built-in fixed and proportional fonts.
	@ingroup Enums
*/
enum ENUM_FONTS
{
	FONT0406, FONT0808, FONT0816, FONT1222, FONT1625, FONT2034, FONTSYS, FONTSYSMEDIUM, FONTSYSLIGHT, FONTLAST=(MAX_FONTS-1)
};

/**
 Describes the TrueType font provided by freetype library.
*/
class FontProperty
{
	public:
		/** for rendering */
		enum FONT_STYLE { ttfSolid, ttfShaded, ttfBlended };
	private:
		void	 	*_font;
		FONT_STYLE 	algorithm;
	public:
		/**
			Creates a TTF font face object for a Window text output.
			You can use a function Valid() to test for a valid font face.
			@param n a filename ("arial.ttf" by example).
			@param sz a size of font face in pixels.
		*/
		FontProperty(char *n, int sz);
		/**
		Unload the font face from memory.
		*/
		~FontProperty();
		/**
		Sets a 'bold' font style.
		*/
		void Bold(int value);
		/**
		Sets an 'underline' font style.
		*/
		void Underline(int value);
		/**
		Sets an 'italic' font style.
		*/
		void Italic(int value);
		/**
		Sets a solid rendering, fast but not very nice. Default.
		*/
		void Solid(void) { algorithm = ttfSolid; }
		/**
		Sets a shaded rendering, slow but very nice.
		*/
		void Shaded(void) { algorithm = ttfShaded; }
		/**
		Sets a blended rendering, slow but very nice.
		@bug still not implemented.
		*/
		void Blended(void) { algorithm = ttfBlended; } // not implemented yet
		/**
		Test the object for a valid data.
		@return true if object contains a valid font face.
		*/
		bool Valid(void) { return !!_font; }
		FONT_STYLE GetAlgorithm(void) { return algorithm; }
		void * GetFace(void) { return _font; }
		char * PrepareFont(variable_record *var, int& size, int count, int offset, char *name);
};

/**
	The OpenGUI interface to the font manager.
	It registers and deregisters variable in-memory or in-file fonts.
	These fonts maybe with fixed width or proportional width.
	Object provides these fonts to others.
*/
class FGFontManager
{
		static FGFont 	FGFonts[MAX_FONTS];
		static int 		font_counter;
		// register variable font from memory
		// img and data must be allocated with malloc
		static int register_font_var(variable_record *data, char *img, int size, char *desc);
		// register variable font from compiled font
		static int register_font_image(fonthdr * hdr, char *img, char *desc);
		static void deregister_fonts(void);
		static int _register_font(unsigned char *source, int width, int height, int count, int offset, variable_record * type, int index, char *desc);
	public:
		FGFontManager();
		~FGFontManager();

		static int register_font_ttf(char *filename, int points, int count=128-' ', int offset=' ', char *desc=0);
		static int register_font_fix(unsigned char *source, int width, int height, int count, int offset, char *desc);
		static int register_font_file(const char *source, char *desc);

		static int textwidth(int ID, char *txt);
		static int textwidth(int ID, char *txt, int cnt);
		static int charwidth(int ID, int c);

		/**
		Tests font ID and returns pointer to.
		@return a valid font record if exists one, or 0 if doesn't exist.
		@param ID font ID for prepared font
		*/
		static FGFont * get_font_ptr(int ID)
		{
			if (ID<0 || ID>font_counter)
				return 0;
			// test for defined fonts
			if (!FGFonts[ID].fontimg)
				return 0;
			return &FGFonts[ID];
		}
		/**
		@return returns the width in pixels for font ID or 0 if doesn't exists.
		for proportional fonts there is the width of bigger glyph in the font.
		@param ID font ID
		*/
		static inline int GetW(int ID)
		{
			return FGFonts[ID].width;
		}
		/**
		@return returns the height in pixels for font ID or 0 if doesn't exists.
		@param ID font ID
		*/
		static inline int GetH(int ID)
		{
			return FGFonts[ID].height;
		}
		/**
		@return returns the pointer to a structure for proportional
		font description or 0 if font is fixed size.
		@param ID font ID
		*/
		static inline variable_record * IsVariableFont(int ID)
		{
			return FGFonts[ID].var;
		}
		/**
		@return returns the font pixelmap or 0 if desn't exists.
		@param ID font ID
		*/
		static inline FGPixel * GetFontImage(int ID)
		{
			return FGFonts[ID].fontimg;
		}
		/**
		@return the font name (max. 23 chars in length+'\\0').
		@param ID font ID
		*/
		static inline char * GetFontName(int ID)
		{
			return FGFonts[ID].name;
		}
		/**
		@return the number of registered fonts in system.
		*/
		static inline int GetCounter(void)
		{
			return font_counter;
		}
};

extern FGFontManager fontman;

/**
	The structure describes a pattern for the patterned line drawing.
*/
struct Pattern
{
	FGPixel *data;
	int len, masked, reset, pos;
	FGPixel next(void);
	Pattern(FGPixel * pp, int l, int m, int r = 1):data(pp), len(l),
	  masked(m), reset(r), pos(0)
	{
	}
};

extern "C" {

#ifdef __DJGPP__
void		sound(int _frequency);
#define		nosound() sound(0)
#endif
#ifdef __QNX__
unsigned int delay(unsigned int);
void sound(unsigned);
void nosound(void);
#else
void delay(unsigned int);
#endif
#ifdef _WIN32
void sound(unsigned);
void nosound(void);
#endif

void outpb(unsigned, unsigned);
void outpw(unsigned, unsigned);
unsigned char inpb(unsigned);
unsigned short inpw(unsigned);
unsigned int inpl(unsigned);
void outpl(unsigned, unsigned);

extern Pattern PatternDot , PatternSlashDot , PatternSlashDotSlash , PatternSlash ;

//unsigned long long _rdtsc(void);
#ifndef __BORLANDC__
char *strupr(register char *s);
int stricmp(const char *s1, const char *s2);
int strnicmp(const char *s1, const char *s2, size_t cnt);
char *strlwr(char *_s);
#endif
/**
	@addtogroup Graphics
	@{
*/
int get_point(int x, int y);
void draw_hline(int x, int y, int w, FGPixel ink, unsigned ppop);
void draw_box(int x, int y, int w, int h, FGPixel ink, unsigned ppop);
void fill_box(int x, int y, int w, int h, FGPixel ink, unsigned ppop);
void draw_point(int x, int y, FGPixel color, unsigned ppop);
int  draw_line(int x, int y, int x1, int y1, FGPixel color, unsigned ppop);
#define draw_line(a,b,c,d,e,f) vector_draw_line(a,b,c,d,e,f)
int  drawto_line(int x, int y, FGPixel ink, unsigned ppop);
void draw_pattern_line(int x, int y, int x1, int y1, Pattern *, unsigned ppop);
void drawto_pattern_line(int x, int y, Pattern *, unsigned ppop);
void draw_pattern_box(int x, int y, int a, int b, Pattern *pat, unsigned ppop);
void draw_spline(int points[8], FGPixel ink, unsigned ppop);
void calc_spline(int points[8], int npts, int *x, int *y);
void moveto(int x, int y);
extern unsigned int areasize(int x, int y);

void _set_default_palette(void);
void _set_fgl_palette(void);
unsigned int get_palette(unsigned int i);
void _palette(unsigned, unsigned int);

int text_out(int x, int y, char *text, FGPixel ink, FGPixel paper, unsigned ppop, int font=FONT0816);
int get_area_size(int x, int y);
void fill_convex(int c, int a[][2], FGPixel ink, unsigned ppop);
void draw_convex(int c, int a[][2], FGPixel ink, unsigned ppop);
void get_block(int x, int y, int a, int b, FGPixel * p);
void put_block(int x, int y, int a, int b, FGPixel * p, unsigned ppop);
void set_clip_rect(int w, int h, int x, int y);

void fill_ellipse(int x, int y, int rx, int ry, FGPixel ink, unsigned ppop);
void draw_ellipse(int x, int y, int rx, int ry, FGPixel ink, unsigned ppop);
void draw_arc(int x, int y, double ang1, double ang2, int r, FGPixel ink, unsigned ppop);
extern void draw_circle(int x, int y, int r, FGPixel ink, unsigned ppop);
extern void fill_circle(int xs, int ys, int r, FGPixel ink, unsigned ppop);

extern void RamToVideo(FGPixel * Image, int x, int y, int xm, int ym, int w, int h, int cx_max, int cy_max, int opcia, unsigned ppop);
extern void VideoToRam(FGPixel * Image, int x, int y, int xm, int ym, int w, int h, int cx_max, int cy_max);
extern void RamToRam(int x, int y, int xmax, int ymax, int xdst, int ydst, int xmaxdst, int ymaxdst, int w, int h, FGPixel * from, FGPixel * to, int alpha, int ppop);
int gprintf(FGPixel ink, FGPixel paper, int x, int y, const char *format,...);
/** @} Graphics */

/**
	@addtogroup Colors
	@{
*/
int GetFreeColors(void);
void SetColorFuzzy(int a);
extern FGPixel CreateColor(int, int, int, int);
void DeleteColor(int);
void set_transpcolor(FGPixel);
/** @} Colors */

void FGmemcpy(FGPixel * to, FGPixel * from, unsigned c);
void FGmemset(FGPixel * to, FGPixel data, unsigned c);
void FGmemcpy2(FGPixel * to, FGPixel * from, unsigned c, unsigned ppop);
void FGmemset2(FGPixel * to, FGPixel data, unsigned c, unsigned ppop);

/**
	Clears the whole screen.
	@ingroup Misc
	@param c background color
*/
void clear_frame_buffer(FGPixel c);
int graph_set_mode(int mode);
/**
	Change from one to other graphics mode with its resolution.
	@ingroup Misc
	@param ww the new X axis resolution
	@param hh the new Y axis resolution
	@return TRUE if OK
*/
int graph_change_mode(int ww, int hh);
void cleanup(void);
int get_colordepth(void);
unsigned GetXRes(void);
unsigned GetYRes(void);
FGPixel * GetFrameBuffer(void);

/**
	For Buffered animation.
	@ingroup Enums
*/
enum ENUM_BUFFERING
{
	FG_DOUBLEBUFFER=2,
	FG_TRIPLEBUFFER,
	FG_QUADBUFFER
};

/**
	Returns absolute time in msec from the start of application. Useful for timing.
	@ingroup Misc
*/
unsigned long FGClock(void);
void Snd(int a, int b);
void Puk(void);

/**
	Returns the version string, by example: "4.1.0"
	@ingroup Misc
*/
const char *GetVer(void);
void UpdateRect(int x, int y, int xm, int ym, int w, int h);
int  GetPriv(void);

class FGDriver;

extern FGDriver *__fg_driver;

// draw clipped to screen line
extern shapper_t   vector_draw_line;

// draw clipped to screen rect
extern clipper_t   vector_clip_rect;

// draw clipped to screen plot
extern plotter_t   vector_draw_point;

// draw clipped to screen box
extern shapper_t   vector_fill_box;

// set 8 bit palette (shift needed for 6-bit palette)
// palette_8 is accepted
extern paletter_t  vector_palette;

// bitblit copy
extern blitter_t   vector_blit_copy;
extern blitter2_t   vector_blit_op;
extern blitter_a_t vector_blit_a;

/**
	This object is provided for common communication between
	any two OpenGUI graphics objects. By default is this unconnected.
	When you want connect the two objects to sending signal from one
	object to other, simply call the Connect() method. In called object
	you must overload virtual methods OnSignal(). When signal
	from one to other object is emited, there is one optional parameter
	that is transmited. You can set this parameter with Connect() or
	late by calling SetParam().
	@see FGCONNECT
*/
class FGConnector
{
		int 	parameter1;
		class 	FGConnector *sendTo;
	public:
		/**
		This is an trivial way to solve problem with unavailable related user defined
		parameter. You can use method SetParam() to any Window to assign some parameter
        of int type. Since you can read this parameter at other place by GetParam().
        If returned value is -1, value is not assigned.
        */
		void	SetParam(int a) { parameter1 = a; }
		/**
        This is an trivial way to solve problem with unavailable related user defined
		parameter. You can use method SetParam() to any Window to assign some parameter
        of int type. Since you can read this parameter at other place by GetParam().
		If returned value is -1, value is not assigned.
		*/
		int		GetParam(void)  const { return parameter1; }

		FGConnector()
		{
			SetParam(-1);
			sendTo = 0;
		}
		void 	Connect(FGConnector *s, int param)
		{
			sendTo = s;
			SetParam(param);
		}
		// do nothing
		virtual void OnSignal(FGConnector *sender, int value)
		{
		}
		void RunSignal(void)
		{
			if (sendTo)
				sendTo->OnSignal(this, parameter1);
		}
};

/**
 MACRO to connect one object with other. See class FGConnector for more.
 @ingroup Defines
*/
#define		FGCONNECT(sender, THIS, PARAM1) { (sender)->Connect((THIS), int(PARAM1)); }

/**
	The types of mutexes.
	@ingroup Enums
*/
enum ENUM_MUTEX_TYPE
{
	fastMutEx=1,
	recursiveMutEx
};

/**
	Synchronisation object - MUTEX. Used when multithrading
	support is enabled.
*/
class FGMutex
{
		int				type;
		int				count;
#ifdef _PTHREAD_H
		pthread_mutex_t m_mutex;
#endif
#ifdef _WIN32
		CRITICAL_SECTION m_mutex;
#endif
	public:
		//! Create a MUTEX. You can choose from fast or recursive one.
		FGMutex(ENUM_MUTEX_TYPE typ = fastMutEx)
#ifdef FG_THREADED
		;
#else
		{}
#endif
		~FGMutex()
#ifdef FG_THREADED
		;
#else
		{}
#endif
		//! Lock the object for others.
		void Lock()
#ifdef FG_THREADED
		;
#else
		{}
#endif
		//! Unlock the object for others.
		void Unlock()
#ifdef FG_THREADED
		;
#else
		{}
#endif
		//! Try lock the object for others. Returns false if one is already locked.
		bool TryLock()
#ifdef FG_THREADED
		;
#else
		{ return true; }
#endif
};

/**
	A base drawing object.
	This object is used as a parent for the all graphics stuff.
	You can use this one for in-memory drawing and after drawing
	complete update the screen with contents of this object.
*/
class DrawBuffer : public FGConnector
{
	public:
		enum Type {	WINDOW=0x4000, CHECKBUTTON, PUSHBUTTON, POINTBUTTON, EDITBOX, PROGRESSBAR,
			BASEMENU, BASEMENUITEM, MENUWINDOW, SLIDEBAR, MULTILINEEDIT, LISTBOX, LABEL,
			BMP_FILE, BMP_IMAGE, BMP_MEM, ROOTWINDOW, FGFRAMEBUFFER, COMBOBOX,
			BMP_NONE=0 };
		friend class GLSurface;
		friend class listbox;
	private:
		static	int ID_counter;
		int		id;
		FGMutex	atomic;
		inline void	lock(void)
		{
			atomic.Lock();
		}
		inline void	unlock(void)
		{
			atomic.Unlock();
		}
		void	_symetry(int xs, int ys, int x,	int	y, FGPixel color);
		void	_symetry2(int xs, int ys, int x, int y, FGPixel color);
		void	_init(int ww, int hh, Type t, FGPixel ii=CBLACK, FGPixel pp=CWHITE, char *nm=0);
	protected:
		int		w;
		int		h;
		int		xoff,yoff;	// coor
		int		wwrk,hwrk;	// sizes!!!
		unsigned long	status;
		FGPixel	*image;
		char 	*name;
		int		colorkey;
		union 	{ Type	type; int _asd; };
		FGPixel	_ink, _paper;
		int 	_font;	// kod fontu
		unsigned _ppop;
		unsigned char alpha;
		FGMutex	clip_lock;

		static	const char *empty_string;
		static	int oldx, oldy;

		FGPixel *	SetArray(FGPixel *array)
		{
			if (image) free(image);
			return image = array;
		}
		DrawBuffer(int ww, int hh, char * nm, Type t, FGPixel i, FGPixel p);
		void 	modify_point(FGPixel *ptr, FGPixel ink);
		void 	__line(int,int,int,int, FGPixel color);
		void 	SetStatus(long s) {	status = s;	}
		static  int GetCounter(void) { return ID_counter; }
		//! Clip coordinates and size to the client space.
		inline int clip(int& xr, int& yr, int& wr, int& hr)
		{
			if (xr<xoff)
			{
				wr -= (xoff-xr);
				xr = xoff;
			}
			if (yr<yoff)
			{
				hr -= (yoff-yr);
				yr = yoff;
			}
			if (xr+wr > wwrk)
			{
				wr = wwrk-(xr-xoff);
			}
			int yoff2 = h - (hwrk+yoff);
			if (yr+hr > h-yoff2)
			{
				hr = h-yr-yoff2;
			}
			return (wr<1 || hr<1) ? 0 : 1; // return if valid
		}
		inline int		clip_shape(int& xr, int& yr, int& wr, int& hr)
		{
			if (xr<0)
			{
				wr += xr;
				xr = 0;
			}
			if (yr<0)
			{
				hr += yr;
				yr = 0;
			}
			if (xr+wr>w)
			{
				wr = w-xr;
			}
			if (yr+hr>h)
			{
				hr = h-yr;
			}
			return (wr<1 || hr<1) ? 0 : 1; // return if valid
		}
	public:
		/**
		Calculate the absolute address of [x:y] pixel. There are no relative offsets used.
		*/
		inline FGPixel * CalcAddr(int x, int y)
		{
			return image + (w * y + x);
		}
		/**
		Resizes the buffer by relative sizes.
		*/
		void	Resize(int dx, int dy);
		/**
		Returns the status of object.
		*/
		long	GetStatus(void)  const	{ return status; }
//		void    Assign(int a, int b, void *c);
		//! Try get the font image and set requested font ID
		//! if success for the object as current.
		inline 	void set_font(int font_id)
		{
			if (FGFontManager::get_font_ptr(font_id))
				_font = font_id;
		}
		//! Returns a current object's font ID.
		inline 	int get_font(void)
		{
			return _font;;
		}
		//! Returns a width of the 'text' in pixels for current font.
		inline int	textwidth(char *text)
		{
			return FGFontManager::textwidth(_font, text);
		}
        //! Returns a width for the first 'cnt' characters
        //! of the 'text' in pixels for current font.
		inline int	textwidth(char *txt, int cnt)
		{
			return FGFontManager::textwidth(_font, txt, cnt);
		}
        //! Returns a width of the character 'c' in pixels for current font.
		inline int	charwidth(int c)
		{
			return FGFontManager::charwidth(_font, c);
		}
		//! Returns a width of one character for current font if one is fixed-size.
		inline int	GetFontW(void)
		{
			return FGFontManager::GetW(_font);
		}
        //! Returns a height of one character for current font if one is fixed-size.
		inline int	GetFontH(void)
		{
			return FGFontManager::GetH(_font);
		}
		//! Set the ROP for all drawing operations used with the object.
		//! See enum PPOP for more details.
		inline unsigned set_ppop(unsigned op)
		{
			unsigned oldp = _ppop;
			_ppop = op;
			return oldp;
		}
        //! Set the foreground color for drawing operations.
		//! This color will be used when you choose function
		//! that don't uses color as parameter immediatelly.
		inline 	void set_fcolor(FGPixel val) { _ink = val; }
		//! Set the background color for text drawing operations.
		//! This color will be used when you choose function
		//! that don't uses color as parameter immediatelly.
		inline 	void set_bcolor(FGPixel val) { _paper = val; }
		//! Set the foreground (ink) and the background (paper)
		//! color for drawing operations.
		inline 	void set_colors(FGPixel i, FGPixel p) { _ink = i; _paper = p; }
		//! Returns the ID of the object.
        /**
        The all controls have an exclusive ID. This value goes from 0 to ..,
		for each window and its control items. For example: when window
		contains 5 buttons, then these buttons have local id values
        from 0 to 4,according to the order of its creating. This is
		useful for fast and easy testing which button (when you use
        one call-back procedure for more buttons i.e.) especially
        with using GetLastActive();
		*/
		int		GetId(void)  const	{ return id; }
		//! Set position and size of the client space for the object (ww & hh are size).
		inline int set_clip_rect(int xx, int yy, int ww, int hh)
		{
			if (hh<1 || ww<1 || xx<0 || yy<0) return 0;
			xoff = xx;
			yoff = yy;
			if ((xx + ww)	> w) ww -= ((xx + ww) - w);
			if ((yy + hh)	> h) hh -= ((yy + hh) - h);
			wwrk = ww;
			hwrk = hh;
			return 1;
		}
        //! Returns current values of the client space of the object.
		inline void	get_clip_rect(int& x,int& y,int& w,int& h)
		{
			x = xoff;
			y = yoff;
			w = wwrk;
			h = hwrk;
		}
		DrawBuffer(int ww=0, int hh=0, Type=BMP_MEM, int color=CBLACK, FGPixel *buf=0);
		DrawBuffer(int ww, int hh, FGPixel *buf);

		virtual ~DrawBuffer();

		// copy block from one buffer to other (with ppop or alpha)
		// this is software blitter only.
		int			bitblit(int xdst, int ydst, int xsrc, int ysrc, int ww, int hh, class DrawBuffer *src, int opcia=BITBLIT_COPY);
		int			hline(int x, int y, int size, FGPixel color);
		int			vline(int x, int y, int size, FGPixel color);
		int			line(int x, int y, int x2, int y2);
		int			line(int x, int y, int x2, int y2, FGPixel color);
		int			lineto(int x, int y);
		int			lineto(int x, int y, FGPixel color);
		//! Sets the start position for first time used lineto().
		void		moveto(int x, int y) { oldx = x; oldy = y; }
		int			pline(int x1, int y1, int x2, int y2, Pattern *p);
		void		spline(int points[8]);
		void		spline(int points[8], FGPixel color);

		int			rect(int xdst, int ydst, int ww, int hh);
		int			rect(int xdst, int ydst, int ww, int hh, FGPixel ink);
		int			box(int xdst, int ydst, int ww, int hh);
		int			box(int xdst, int ydst, int ww, int hh, FGPixel i);

		int			text(int x, int y, const char *txt);
		int 		text(int x, int y, const char *txt, FGPixel i);
		int 		text(int x, int y, const char *txt, FGPixel i, FGPixel p);
#undef printf
		int			printf(int , int , const char * , ... );
		int			putpixel(int x, int y);
		int			putpixel(int x, int y, FGPixel color);
		/**
		 Returns the value of color at the position [x:y]
		 of the object or (-1) if coordinates are out of range.
		*/
		int			getpixel(int x, int y)
		{
			if (x<0	|| y<0 || x>=wwrk || y>=hwrk)
				return 0xFFFFFFFF;
			return *(image+(x+xoff)+(y+yoff)*w);
		}
		void		stretch(int _w, int _h);
		void		stretch(float ratiox, float ratioy);

		int			arc(int	x,	int	y,	double ang1, double ang2, int r);
		int			arc(int	x,	int	y,	double ang1, double ang2, int r, FGPixel color);
		int			ellipse(int x, int y, int rx, int ry);
		int			ellipse(int x, int y, int rx, int ry, FGPixel color);
		int			fellipse(int x, int y, int rx, int ry);
		int			fellipse(int x, int y, int rx, int ry, FGPixel color);
		int			circle(int	xs,	int	ys,	int	r);
		int			circle(int	xs,	int	ys,	int	r, FGPixel color);
		int			fcircle(int	xs,	int	ys,	int	r);
		int			fcircle(int	xs,	int	ys,	int	r, FGPixel color);
		void 		polygon(int n, int pt[][2]);
		void 		polygon(int n, int pt[][2], FGPixel color);
		FGRect		fpolygon(int n, int pt[][2]);
		FGRect		fpolygon(int n, int pt[][2], FGPixel color);
		FGRect		ftriangle(int x1, int y1, int x2, int y2, int x3, int y3);
		FGRect		ftriangle(int x1, int y1, int x2, int y2, int x3, int y3, FGPixel Ink);

		//! Returns the address of pixel data storage. There are pixels, from right to left, line by line.
		FGPixel		*GetArray(void) const { return image; }
		//! Returns an ALPHA (transparency) value of the object.
		int			GetAlpha(void) const { return alpha; }
		//! Returns a foreground color of the object.
		FGPixel	  	GetInk(void)  const { return (FGPixel)_ink; }
		//! Returns a background color of the object.
		FGPixel	  	GetPaper(void)  const { return	(FGPixel)_paper; }
		//! Returns a width of the object.
		int		  	GetW(void)  const { return	w; }
		//! Returns a height of the object.
		int		  	GetH(void)  const { return	h; }
		//! Returns the X coordinate of the client space of the object.
		int	   		GetXW(void) const { return xoff; }
		//! Returns the Y coordinate of the client space of the object.
		int	   		GetYW(void) const { return yoff; }
		//! Returns a client space width of the object.
		int			GetWW(void) const {	return wwrk; }
		//! Returns a client space height of the object.
		int			GetHW(void) const {	return hwrk; }
		//! Returns the type of the object (WINDOW, PUSHBUTTON ..)
		enum Type 	GetType(void)  const {	return type; }
		//! Returns the COLOR_KEY of the object (i.e. color that is used as transparent).
		int 		GetColorKey(void) const { return colorkey; }
		//! Returns the name of the object.
		char		*GetName(void)  const {	return name; }
        //! Set the COLOR_KEY of the object (i.e. color that will be transparent when bitblit
		//! operation will be performed with this object).
		void		SetColorKey(int ck)
		{
			colorkey = ck;
		}
		void	  	SetName(char *s);
        //! Set the level of transparency for the object when bitblit
        //! operation will be performed with this object. The level
        //! is number from 0 to 255, where 0 is completely transparent
        //! and 255 completely opaque.
		void		SetAlpha(int a) { alpha = a; }
		//! Clear the entire object image with this color.
		void		clear(int color) { FGmemset(image, color, w*h); }
};
/*
#define	SAVE_CLIPPING(THIS)	  \
	int _a,_b,_c,_d;          \
		clip_lock.Lock();     \
		_a = (THIS)->xoff;         \
		_b = (THIS)->yoff;         \
		_c = (THIS)->wwrk;         \
		_d = (THIS)->hwrk;         \
		(THIS)->xoff = 0;         \
		(THIS)->yoff = 0;         \
		(THIS)->wwrk = (THIS)->w;         \
		(THIS)->hwrk = (THIS)->h;

#define	SAVE_CLIPPING2(THIS)	  \
		clip_lock.Lock();     \
		_a = (THIS)->xoff;         \
		_b = (THIS)->yoff;         \
		_c = (THIS)->wwrk;         \
		_d = (THIS)->hwrk;         \
		(THIS)->xoff = 0;         \
		(THIS)->yoff = 0;         \
		(THIS)->wwrk = (THIS)->w;         \
		(THIS)->hwrk = (THIS)->h;

#define	RESTORE_CLIPPING(THIS)	 \
		(THIS)->xoff = _a;         \
		(THIS)->yoff = _b;         \
		(THIS)->wwrk = _c;         \
		(THIS)->hwrk = _d;			\
		clip_lock.Unlock();
*/

#define	SAVE_CLIPPING(THIS) int _a,_b,_c,_d; clip_lock.Lock(); _a = (THIS)->xoff; _b = (THIS)->yoff; _c = (THIS)->wwrk; _d = (THIS)->hwrk; (THIS)->xoff = 0; (THIS)->yoff = 0; (THIS)->wwrk = (THIS)->w; (THIS)->hwrk = (THIS)->h;
#define	SAVE_CLIPPING2(THIS) clip_lock.Lock(); _a = (THIS)->xoff; _b = (THIS)->yoff; _c = (THIS)->wwrk; _d = (THIS)->hwrk; (THIS)->xoff = 0; (THIS)->yoff = 0; (THIS)->wwrk = (THIS)->w; (THIS)->hwrk = (THIS)->h;
#define	RESTORE_CLIPPING(THIS)	(THIS)->xoff = _a; (THIS)->yoff = _b; (THIS)->wwrk = _c; (THIS)->hwrk = _d; clip_lock.Unlock(); 

/**
	Structure to provide user defined mouse cursor.
*/
struct FGMouseCursor
{
	FGPixel *bitmap;
	FGPixel *mask;
	int	xoff, yoff;
	int w,h;

	FGMouseCursor(FGPixel *a, FGPixel *b, int c, int d, int e, int f)
		: bitmap(a), mask(b), xoff(c), yoff(d), w(e), h(f)
	{
	}
};

/**
	Internal - one palette record
*/
struct FGPalette
{
		unsigned char b;
		unsigned char g;
		unsigned char r;
		unsigned char alfa;
	  public:
		FGPalette(void)
		{
			r = g = b = alfa = 0;
		}
		FGPalette(unsigned p)
		{
			*((unsigned *) this) = p;
		}
		FGPalette(unsigned char _r, unsigned char _g, unsigned char _b)
		{
			r = _r;
			g = _g;
			b = _b;
			alfa = 0;
		}
		unsigned int GetValue(void)
		{
			return *((unsigned *) this);
		}
};

extern FGMouseCursor IDC_NORMAL_SMALL;
extern FGMouseCursor IDC_NORMAL_LARGE;
extern FGMouseCursor *idc_normal;

} // extern "C"

#ifdef __BORLANDC__
#pragma option pop /* pop -a switch */
#pragma option pop /* pop -b */
#endif

#endif // base_h
