/*
   Copyright (C) 1996,2003  Marian Krivos

   nezmar@atlas.sk

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   bitmap.cc - bitmap engine

 */

#include <stdarg.h>

#include "config.h"
#include "fastgl.h"
#include "_fastgl.h"

#ifdef FG_JPEG
extern "C" {		// jpeglib.h forgot this... (Oron Peled)
#include <jpeglib.h>
#ifdef __BORLANDC__
#ifdef __linux__
#pragma link "libjpeg.so"
#else
#pragma link "libjpeg.lib"
#endif
#endif
}
#endif

#ifdef FG_PNG
#include <png.h>
#ifdef __BORLANDC__
#ifdef __linux__
#pragma link "libpng.so"
#pragma link "libz.so"
#else
#pragma link "libpng.lib"
#pragma link "zlib.lib"
#endif
#endif
#else
#include <setjmp.h>
#endif

#include "gif.h"

#define 	TRANSP_KEY			0xFE00FE
#define     MAX_LWZ_BITS        12
#define		INTERLACE			0x40
#define		LOCALCOLORMAP		0x80

#define BitSet(byte, bit)	(((byte) & (bit)) == (bit))
#define	ReadOK(file,buffer,len) (fread(buffer, len, 1, file) != 0)
#define MKINT(a,b)		(((b)<<8)|(a))
#define NEW(x)			((x *)malloc(sizeof(x)))

/***************************************************************************
*
*  ERROR()    --  should not return
*  INFO_MSG() --  info message, can be ignored
*
***************************************************************************/

#define INFO_MSG(fmt)
#define _ERROR(str) 	IError(str, 0), longjmp(setjmp_buffer, 1)

/***************************************************************************/

static int readColorMap(FILE *, int, unsigned char[GIF_MAXCOLORS][3]);
static int GetDataBlock(FILE *, unsigned char *);
static void readImage(FILE *, int, int, int, unsigned char *);

static jmp_buf setjmp_buffer;
int DrawBuffer::ID_counter = 1;		// incremental ID of all graph. objects
int DrawBuffer::oldx, DrawBuffer::oldy;

/* TRUEVISION-XFILE magic signature string */
static unsigned char magic[18] =
{
  0x54, 0x52, 0x55, 0x45, 0x56, 0x49, 0x53, 0x49, 0x4f,
  0x4e, 0x2d, 0x58, 0x46, 0x49, 0x4c, 0x45, 0x2e, 0x0
};

/***************************************************************************/

/*
void    DrawBuffer::Assign(int a, int b, void *c)
{
	wwrk = w = a;
	hwrk = h = b;
	image = (FGPixel *)c;
	xoff=yoff=0;
	type = BMP_IMAGE;
}
*/

#ifdef INDEX_COLORS
void Bitmap::DoPalette(int one)
{
	cUsedTable = new long[256];
	int sz = w * h, d, i;
	FGPixel *bm = GetArray();

	memset(cUsedTable, 0, sizeof(long) * 256);	// reset color table
	// calculate which colors is used in image
	for (i = 0; i < sz; i++)
	{
		cUsedTable[*bm++]++;
	}

	bm = GetArray();
	// allocating colors
	for (i = 0; i < 256; i++)
	{
		if (cUsedTable[i])
		{
			d = *((long *) ((char *) &rgb + i * one));
			if (i!=colorkey)
				cUsedTable[i] = (char) CreateColor((d >> 16)&0xff, (d >> 8)&0xff, d&0xff, -1); // CORRECTED
			else cUsedTable[i] = i;
		}
		else cUsedTable[i] = -1;
	}

	// remapping colors
	for (i = 0; i < sz; i++)
	{
		bm[i] = (char) cUsedTable[bm[i]];
	}
}
#endif

/* read_RLE8_compressed_image:
 *  For reading the 8 bit RLE compressed BMP image format.
 */
static void read_RLE8_compressed_image(FILE * f, unsigned char *bmp, int w, int h)
{
	unsigned char count, val, val0;
	int j, pos, line;
	int eolflag, eopicflag;

	eopicflag = 0;
	line = h - 1;

	while (eopicflag == 0)
	{
		pos = 0;				/* x position in bitmap */
		eolflag = 0;			/* end of line flag */

		while ((eolflag == 0) && (eopicflag == 0))
		{
			count = (char) getc(f);
			val = (char) getc(f);

			if (count > 0)
			{					/* repeat pixel count times */
				for (j = 0; j < count; j++)
				{
					bmp[line * w + pos] = val;
					pos++;
				}
			}
			else
			{
				switch (val)
				{
					case 0:		/* end of line flag */
						eolflag = 1;
						break;
					case 1:		/* end of picture flag */
						eopicflag = 1;
						break;
					case 2:		/* displace picture */
						count = (char) getc(f);
						val = (char) getc(f);
						pos += count;
						line += val;
						break;
					default:	/* read in absolute mode */
						for (j = 0; j < val; j++)
						{
							val0 = (char) getc(f);
							bmp[line * w + pos] = val0;
							pos++;
						}

						if (j % 2 == 1)
							/*val0 = (char) */getc(f);	/* align on word boundary */
						break;
				}
			}
			if (pos > w)
				eolflag = 1;
		}
		line--;
		if (line < 0)
			eopicflag = 1;
	}
}

/**
	Create a BITMAP object from in-memory BMP file image. Absolete.
*/
Bitmap::Bitmap(void *n) : DrawBuffer()
{
	long one, bits;
//	unsigned long c;
	struct in_image_bmp *hp;
	unsigned char *nn = ((unsigned char *) n) + sizeof(long);

	type = BMP_NONE;
	original_name = name = 0;
	cUsedTable = 0;
	hp = (in_image_bmp *) nn;	// skip size of file
	one = 4;
//	c = 256;					// compute numbers of used colors
	memcpy(&rgb, nn + sizeof(in_image_bmp), sizeof(rgb));
	wwrk = w = hp->cx;
	hwrk = h = hp->cy;
	bits = hp->bits;

	if (bits == 8)
	{
	}

	SetArray((FGPixel *) nn + sizeof(in_image_bmp) + sizeof(rgb));
	type = BMP_IMAGE;
#ifdef INDEX_COLORS
	DoPalette(one);
#endif
}

void Bitmap::turn_bitmap(int cnt, FGPixel *from, FGPixel *to)
{
	FGPixel *p = new FGPixel[w];

	for (int i = 0; i < cnt; i++)
	{
		memmove(p, to - (i * w), w);
		memmove(to - (i * w), from + (i * w), w);
		memmove(from + (i * w), p, w);
	}
	delete p;
}

static void copy_gif_palette(unsigned char *to, unsigned char *data, int cnt)
{
	int r,g,b;
	for(;cnt>0;cnt--)
	{
		r = *data++;
		g = *data++;
		b = *data++;
		*to++ = b;		
		*to++ = g;
		*to++ = r;
	}
}

/**
	Create a BITMAP object from one of supported image formats.
	Parameter 'n' is pointer to filename. You can test to fail
	after return. The member 'type' must not be BMP_NONE, else
	object is no valid. Supported file formats are BMP, TGA, GIF
	and PCX. You can enable JPG and PNG by editing 'config.mak'
	optionaly too. These two are supported by third party libs
	(libjpeg & libpng).
*/
Bitmap::Bitmap(char *n) : DrawBuffer()
{
	int ys, one, compressed = 0;
	unsigned int c, bits = 0;
	FILE *f=0;
	char nm[128], *n2 = n;
	int rest;
	long pixel;
	register unsigned char *src;
	char *gifp;
#ifndef INDEX_COLORS
	register FGPixel *bp;
	int sz, d, i;
	FGPixel *pom;
#endif

	type = BMP_NONE;
	init();
	if (!strstr(n, ".bmp") && !strstr(n, ".BMP")
		&& !strstr(n, ".gif") && !strstr(n, ".GIF")
		&& !strstr(n, ".pcx") && !strstr(n, ".PCX")
		&& !strstr(n, ".png") && !strstr(n, ".PNG")
		&& !strstr(n, ".jpg") && !strstr(n, ".JPG")
		&& !strstr(n, ".jpeg")&& !strstr(n, ".JPEG")
		&& !strstr(n, ".pcx") && !strstr(n, ".PCX")
		&& !strstr(n, ".tga") && !strstr(n, ".TGA")
)
	{
		strcpy(nm, n);
		strcat(nm, ".bmp");
		n = nm;
	}
	strcpy(nm, n);
	original_name = strdup(nm);
	if ((gifp=strstr(nm, ".gif")) || (gifp=strstr(nm, ".GIF")))
	{							// decode *.gif file
		if (gif(nm,gifp,one,bits,c)) return;
	}
	else if ((gifp=strstr(nm, ".pcx")) || (gifp=strstr(nm, ".PCX")))
	{							// decode *.gif file
		if(pcx(nm,gifp,one,bits,c)) return;
	}
	else if ((gifp=strstr(nm, ".tga")) || (gifp=strstr(nm, ".TGA")))
	{							// decode *.gif file
		if(tga(nm,gifp,one,bits,c)) return;
	}
#ifdef FG_PNG
	else
 	if ((gifp=strstr(nm, ".png")) || (gifp=strstr(nm, ".PNG")))
	{							// decode *.gif file
		if (png(nm,gifp,one,bits,c)) return;
	}
#endif
#ifdef FG_JPEG
	else
	if ((gifp=strstr(nm, ".jpg")) || (gifp=strstr(nm, ".JPG")) || (gifp=strstr(nm, ".jpeg")) || (gifp=strstr(nm, ".JPEG")))
	{							// decode *.gif file
		if (jpg(nm,gifp,one,bits,c)) return;
	}
#endif
	else
	{							// decode *.bmp file
		f = fopen(n, "rb");
		if (!f)
		{
			sprintf(nm, "Bitmap '%s' not found!", n2);
			IError(nm, 0);
			return;
		}
		name = new char[strlen(n) + 1];
		strcpy(name, n);
		if (sizeof(bmfh) != 10)
		{
			::printf("Ooops, your data structure is not good aligned!!!\a\n");
			assert(0);
			abort();			// 100% sure to stop
		}
		fread(&bmfh, sizeof(bmfh), 1, f);	// read first 10 bytes
		// return if no BMP
		if (bmfh.usType != (unsigned short) ('M' * 256 + 'B'))
		{
			char jj[128];
			sprintf(jj,"%s isn't supported image file!", n);
			fclose(f);
			IError(jj, 0);
			return;
		}

		fread(&bmi, sizeof(bmi), 1, f);		// read next 4 bytes
		memset(&bmih, 0, sizeof(bmih));
		fread(&bmih, 4, 1, f);	// read size of bitmap info header
		fread(&bmih.bmih1.cx, bmih.bmih1.size - 4, 1, f);	// if 0x0c then OS/2 bitmap
		if (bmih.bmih1.size == 0x0C)
			one = 3;
		else
			one = 4;

		c = (bmi.imageOffset - (bmih.bmih1.size + 14)) / one;	// compute numbers of used colors

		if (c == 0)
		{
#ifdef INDEX_COLORS
			fclose(f);
			IError("True colors BMP not supported!", 0);
			return;
#else
			w = bmih.bmih2.cx;
			h = bmih.bmih2.cy;
			bits = bmih.bmih2.cBitCount;
			if (bmih.bmih2.cPlanes != 1)
			{
				fclose(f);
				IError("This color depth BMP is not supported!", 0);
				return;
			}
#endif
		}
		else if (one == 3)
		{
			if ((compressed = bmih.bmih1.ulCompression) > 1)
			{
				IError("No this compressed BMP supported!", 0);
				return;
			}
			fread(&rgb, sizeof(RGB3), c, f);
			w = bmih.bmih1.cx;
			h = bmih.bmih1.cy;
			bits = bmih.bmih1.cBitCount;
			if (bmih.bmih1.cPlanes != 1)
			{
				fclose(f);
				IError("Multi-planes BMP not supported!", 0);
				return;
			}
		}
		else if (one == 4)
		{
			if ((compressed = bmih.bmih2.ulCompression) > 1)
			{
				IError("Unknown compress format!", 0);
				return;
			}
			fread(&rgb, sizeof(RGB4), c, f);
			w = bmih.bmih2.cx;
			h = bmih.bmih2.cy;
			bits = bmih.bmih2.cBitCount;
			if (bmih.bmih2.cPlanes != 1)
			{
				fclose(f);
				IError("Multi-planes BMP not supported!", 0);
				return;
			}
		}
		assert(SetArray((FGPixel *) calloc(areasize(w, h), 1)));
	}

	//
	// common image decoder
	//
	if (f) switch (bits)
	{
		case 1:
			for (ys = h - 1; ys >= 0; ys--)
			{
				src = (unsigned char *) GetArray() + ys * w;
				for (int xs = w / 8; xs > 0; xs--)
				{
					fread((unsigned char *) &pixel, 1, 1, f);
					*src++ = (pixel >> 7) & 1;
					*src++ = (pixel >> 6) & 1;
					*src++ = (pixel >> 5) & 1;
					*src++ = (pixel >> 4) & 1;
					*src++ = (pixel >> 3) & 1;
					*src++ = (pixel >> 2) & 1;
					*src++ = (pixel >> 1) & 1;
					*src++ = (pixel) & 1;
				}
				rest = w % 8;
				if (rest)
				{
					fread((unsigned char *) &pixel, 1, 1, f);
					*src++ = (pixel >> 7) & 1;
					if (--rest == 0)
						goto end_expand;
					*src++ = (pixel >> 6) & 1;
					if (--rest == 0)
						goto end_expand;
					*src++ = (pixel >> 5) & 1;
					if (--rest == 0)
						goto end_expand;
					*src++ = (pixel >> 4) & 1;
					if (--rest == 0)
						goto end_expand;
					*src++ = (pixel >> 3) & 1;
					if (--rest == 0)
						goto end_expand;
					*src++ = (pixel >> 2) & 1;
					if (--rest == 0)
						goto end_expand;
					*src++ = (pixel >> 1) & 1;
				}
			  end_expand:
				rest = (w * bits / 8) % 4;
				if (w % 8)
					rest++;
				while (rest & 3)
				{
					rest++;
					fgetc(f);
				}
			}
			break;
		case 4:
			for (ys = h - 1; ys >= 0; ys--)
			{
				src = (unsigned char *) GetArray() + ys * w;
				for (int xs = w / 2; xs > 0; xs--)
				{
					fread((unsigned char *) &pixel, 1, 1, f);
					*src++ = (pixel >> 4) & 15;
					*src++ = pixel & 15;
				}
				rest = w % 2;
				if (rest)
				{
					fread((unsigned char *) &pixel, 1, 1, f);
					*(src++) = (pixel >> 4) & 15;
				}
				rest = (w * bits / 8) % 4;
				if (w % 2)
					rest++;
				while (rest & 3)
				{
					rest++;
					fgetc(f);
				}
			}
			break;
		case 8:
			if (!compressed)
				for (ys = h - 1; ys >= 0; ys--)
				{
					fread((unsigned char *) GetArray() + ys * w, w, 1, f);
					rest = (w * bits / 8) % 4;
					while (rest & 3)
					{
						rest++;
						fgetc(f);
					}
				}
			else read_RLE8_compressed_image(f, (unsigned char *) GetArray(), w, h);
			break;
#ifndef INDEX_COLORS
		case 24:
			for (ys = h - 1; ys >= 0; ys--)
			{
				bp = (image + ys * w);
				for (int xs = 0; xs < w; xs++)
				{
					fread(&pixel, 3, 1, f);
					*bp = FGDirectColor(pixel);
					bp++;
				}
			}
			rest = (w * bits / 8) % 4;
			while (rest & 3)
			{
				rest++;
				fgetc(f);
			}
			break;
#endif
#ifdef DIRECT_COLORS
		case 16:
		case 15:
			for (ys = h - 1; ys >= 0; ys--)
			{
				bp = (image + ys * w);
				for (int xs = 0; xs < w; xs++)
				{
					fread(bp, 2, 1, f);
					bp++;
				}
			}
			rest = (w * bits / 8) % 4;
			while (rest & 3)
			{
				rest++;
				fgetc(f);
			}
			break;
#endif
		default:
			IError("Bad Bitmap format!", 0);
			fclose(f);
			free(GetArray());
			SetArray(0);
			return;
	}
	if (f) fclose(f);
#ifndef INDEX_COLORS
	if (bits<=8)
	{
		bp = (FGPixel *) calloc(areasize(w, h), 1);
		pom = bp;
		src = (unsigned char *) GetArray();
		sz = w * h;

		if (colorkey==-1) for (i = 0; i < sz; i++)
		{
			d = *((long *) ((char *) &rgb + *src * one));
			*pom = FGDirectColor(d);
			pom++;
			src++;
		}
		else
		{
			for (i = 0; i < sz; i++)
			{
				if (*src == colorkey) d = TRANSP_KEY;
				else d = *((long *) ((char *) &rgb + *src * one));
				*pom++ = FGDirectColor(d);
				src++;
			}
			colorkey = FGDirectColor(TRANSP_KEY);
		}
		free(image);
		image = bp;
	}
#endif
	type = BMP_FILE;
	wwrk = w;
	hwrk = h;
#ifdef INDEX_COLORS
	DoPalette(one);
#endif
}

#ifdef FG_JPEG
struct r_jpeg_error_mgr
{
	struct jpeg_error_mgr pub;
    jmp_buf envbuffer;
};

int fh_jpeg_id(char *name)
{
    FILE * fd;
    unsigned char id[10];
    fd=fopen(name,"rb");
		if(fd==0) return(0);
    fread(id,1,10,fd);
    fclose(fd);
    if(id[6]=='J' && id[7]=='F' && id[8]=='I' && id[9]=='F') return(1);
    if(id[0]==0xff && id[1]==0xd8 && id[2]==0xff) return(1);
    return(0);
}
			    

void jpeg_cb_error_exit(j_common_ptr cinfo)
{
    struct r_jpeg_error_mgr *mptr;
    mptr=(struct r_jpeg_error_mgr*) cinfo->err;
    (*cinfo->err->output_message) (cinfo);
    longjmp(mptr->envbuffer,1);
}

#endif

int Bitmap::jpg(char *n, char *gifp, int& one, unsigned int& bits, unsigned int& c)
{
#ifdef FG_JPEG
	struct jpeg_decompress_struct cinfo;
    struct jpeg_decompress_struct *ciptr;
    struct r_jpeg_error_mgr emgr;
    FGPixel *bp;
    FILE *fh;
    JSAMPLE *lb;

    ciptr=&cinfo;
	if(fh_jpeg_id(n)==0) return 1;
    if(!(fh=fopen(n,"rb"))) return(1);
    ciptr->err=jpeg_std_error(&emgr.pub);
    emgr.pub.error_exit=jpeg_cb_error_exit;
    if(setjmp(emgr.envbuffer)==1)
    {
		// FATAL ERROR - Free the object and return...
		jpeg_destroy_decompress(ciptr);
		fclose(fh);
		return(2);
    }
    
    jpeg_create_decompress(ciptr);
    jpeg_stdio_src(ciptr,fh);
    jpeg_read_header(ciptr,TRUE);
    ciptr->out_color_space=JCS_RGB;
    jpeg_start_decompress(ciptr);

    w = ciptr->output_width;
	h = ciptr->output_height;
    int cc = ciptr->output_components;
	c = 0;
	bits = 24;
	one = 0;
	strcpy(gifp, ".bmp");
	name = new char[strlen(n) + 1];
	strcpy(name, n);
	assert(SetArray((FGPixel *) calloc(areasize(w, h), 1)));
	bp = GetArray();

    if(cc==3)
    {
		unsigned char pixel[4];
		lb = (JSAMPLE *)(*ciptr->mem->alloc_small)(
			(j_common_ptr) ciptr,
			JPOOL_PERMANENT,
			cc*w);
		while (ciptr->output_scanline < ciptr->output_height)
		{
		    jpeg_read_scanlines(ciptr, &lb, 1);
			for (int xs = 0; xs < w; xs++)
			{
				pixel[2] = *(lb+3*xs);
				pixel[1] = *(lb+3*xs+1);
				pixel[0] = *(lb+3*xs+2);
				*bp = FGDirectColor(*(unsigned int *)(&pixel));
				bp++;
			}
		}
    }
    jpeg_finish_decompress(ciptr);
    jpeg_destroy_decompress(ciptr);
    fclose(fh);
	return 0;
#else
	return 1;
#endif
}

#ifdef FG_PNG

#define PNG_BYTES_TO_CHECK 4
#define min(x,y) ((x) < (y) ? (x) : (y))

int fh_png_id(char *name)
{
    FILE* fd;
    char id[4];
    fd=fopen(name,"rb");
	if(fd==0) return(0);
    fread(id,1,4,fd);
    fclose(fd);
    if(id[1]=='P' && id[2]=='N' && id[3]=='G') return(1);
    return(0);
}
#endif // PNG

int Bitmap::png(char *n, char *gifp, int& one, unsigned int& bits, unsigned int& c)
{
#ifdef FG_PNG
    png_structp png_ptr;
    png_infop info_ptr;
    png_uint_32 width, height;
    int i;
	int bit_depth, color_type, interlace_type;
	int number_passes,pass;
    char *rp;
    png_bytep rptr[2];
    FGPixel *fbptr;
    FILE *fh;

	if(fh_png_id(n)==0) return 1;
    if(!(fh=fopen(n,"rb"))) return(1);

    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);
    if (png_ptr == NULL) return(2);
    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL)
    {
		png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		fclose(fh); return(2);
    }
    rp=0;
    if (setjmp(png_ptr->jmpbuf))
    {
		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        if(rp) free(rp);
		fclose(fh); return(2);
    }

	png_init_io(png_ptr,fh);

    png_read_info(png_ptr, info_ptr);
    png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,&interlace_type, NULL, NULL);

    if (color_type == PNG_COLOR_TYPE_PALETTE)
		png_set_expand(png_ptr);
    if (bit_depth < 8)
		png_set_packing(png_ptr);
    if (color_type == PNG_COLOR_TYPE_GRAY || color_type== PNG_COLOR_TYPE_GRAY_ALPHA)
		png_set_gray_to_rgb(png_ptr);
    if( bit_depth==16)
		png_set_strip_16(png_ptr);
	png_set_filter(png_ptr, 0, PNG_NO_FILTERS);
/*
	If, for some reason, you don't need the alpha channel on an image,
	and you want to remove it rather than combining it with the background
	(but the image author certainly had in mind that you *would* combine
	it with the background, so that's what you should probably do):
*/
    if (color_type & PNG_COLOR_MASK_ALPHA)
		png_set_strip_alpha(png_ptr);
/*
	In PNG files, the alpha channel in an image
	is the level of opacity.  If you need the alpha channel in an image to
	be the level of transparency instead of opacity, you can invert the
	alpha channel (or the tRNS chunk data) after it's read, so that 0 is
	fully opaque and 255 (in 8-bit or paletted images) or 65535 (in 16-bit
	images) is fully transparent, with
*/
//    png_set_invert_alpha(png_ptr);

	number_passes = png_set_interlace_handling(png_ptr);
    png_read_update_info(png_ptr,info_ptr);
    
    rp=(char*) malloc(width*3);
    rptr[0]=(png_bytep) rp;
	w = width;
	h = height;
	c = 0;
	bits = 24;
	one = 0;
	strcpy(gifp, ".bmp");
	name = new char[strlen(n) + 1];
	strcpy(name, n);
	assert(SetArray((FGPixel *) calloc(areasize(w, h), 1)));
	unsigned char pixel[4];

	for (pass = 0; pass < number_passes; pass++)
	{
		fbptr=GetArray();
		for(i=0;i<(int)height;i++)
		{
    	    png_read_rows(png_ptr, rptr, NULL, 1);
			for (int xs = 0; xs < w; xs++)
			{
				pixel[2] = *(rp+3*xs);
				pixel[1] = *(rp+3*xs+1);
				pixel[0] = *(rp+3*xs+2);
				*fbptr = FGDirectColor(*(unsigned int *)(&pixel));
				fbptr++;
			}
		}
    }
    free(rp);
    png_read_end(png_ptr, info_ptr);
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    fclose(fh);
	return 0;
#else
	return 1;
#endif
}

int Bitmap::pcx(char *n, char *gifp, int& one, unsigned int& bits, unsigned int& c)
{
	char hdr[128];
	FILE *f=0;
	char nm[128];
	int counter, byte, data;
	register unsigned char *src;
	
		f = fopen(n, "rb");
		if (!f)
		{
			sprintf(nm, "Bitmap '%s' not found!", n);
			IError(nm, 0);
			return 1;
		}
		fread(hdr, 128, 1, f);
		if (hdr[0] != 0xA || hdr[1] != 5)
		{
error:
			sprintf(nm, "Bitmap '%s' isn't PCX 5.0 file!", n);
			IError(nm, 0);
			fclose(f);
			return 2;
		}
		w = *(short *)(hdr+66);
		h = *(short *)(hdr+10)+1;
		bits = hdr[3];
		assert(SetArray((FGPixel *) calloc(areasize(w, h), 1)));
		int sz = w * h;
		src = (unsigned char *)GetArray();
		unsigned char *end = src+sz;
		while (src < end)
		{
			byte = fgetc(f);
			if (0xC0 == (0xC0 & byte))
			{
				counter = 0x3F & byte;
				data = fgetc(f);
				while(counter--)
				{
					*src++ = data;
				}
			}
			else
			{
				*src++ = byte;
			}
		}
		byte = fgetc(f);
		if (byte != 0xC) goto error;
		fread(&rgb, 768, 1, f);
		copy_gif_palette((unsigned char *)&rgb, (unsigned char *)&rgb, 256);
		fclose(f);
//		f = 0;
		c = 256;
		one = 3;
		strcpy(gifp, ".bmp");
		name = new char[strlen(n) + 1];
		strcpy(name, n);
		return 0;
}

static int rle_read(FILE * fp, unsigned char * buffer, tga_info * info)
{
	static int repeat = 0;
	static int direct = 0;
	static unsigned char sample[4];
	int head;
	int x, k;

	for (x = 0; x < info->width; x++)
	{
		if (repeat == 0 && direct == 0)
		{
			head = getc(fp);

			if (head == EOF)
			{
				return EOF;
			}
			else if (head >= 128)
			{
				repeat = head - 127;

				if (fread(sample, info->bytes, 1, fp) < 1)
					return EOF;
			}
			else
			{
				direct = head + 1;
			}
		}

		if (repeat > 0)
		{
			for (k = 0; k < info->bytes; ++k)
			{
				buffer[k] = sample[k];
			}

			repeat--;
		}
		else					/* direct > 0 */
		{
			if (fread(buffer, info->bytes, 1, fp) < 1)
				return EOF;

			direct--;
		}

		buffer += info->bytes;
	}

	return 0;
}

static void flip_line(unsigned char * buffer, tga_info * info)
{
	unsigned char temp;
	unsigned char *alt;
	int x, s;

	alt = buffer + (info->bytes * (info->width - 1));

	for (x = 0; x * 2 <= info->width; x++)
	{
		for (s = 0; s < info->bytes; ++s)
		{
			temp = buffer[s];
			buffer[s] = alt[s];
			alt[s] = temp;
		}

		buffer += info->bytes;
		alt -= info->bytes;
	}
}

static void read_line(FILE * fp,
		  FGPixel * row, unsigned char * buffer, tga_info * info, unsigned *palette)
{
 	int width=info->width;
#ifndef INDEX_COLORS
	int x;
	unsigned char *dest = (unsigned char *)row;
#endif
	// read from the file
	if (info->imageCompression == TGA_COMP_RLE)
	{
		rle_read(fp, buffer, info);
	}
	else
	{
		fread(buffer, info->bytes, info->width, fp);
	}
	// flip
	if (info->flipHoriz)
	{
		flip_line(buffer, info);
	}

	switch(info->imageType)
	{
	case TGA_TYPE_COLOR:
#ifdef INDEX_COLORS
		IError("Not supported BPP", 1);
#endif
#ifdef DIRECT_COLORS
		if (info->bpp == 16)
		{
			for (x = 0; x < width; x++)
			{
				dest[0] = buffer[0];
				dest[1] = buffer[1];
#ifdef BPP16
				dest[1] <<= 1;
				dest[1] |= (dest[0]&0x80? 1:0);
				dest[0] = (dest[0]&0x1f)+((dest[0]&0x60)<<1);
#endif
				buffer+=2;
				dest+=2;
			}
		}
		else
		{
			int pad=info->bytes;
			for (x = 0; x < width; x++)
			{
				*((short *)dest) = FGDirectColor(*(unsigned *)buffer);
				buffer+=pad;
				dest+=2;
			}
		}
#endif
#ifdef TRUE_COLORS
		switch (info->bpp)
		{
		case 16:	// 16 > 32bit
			for (x = 0; x < width; x++)
			{
				dest[0] = ((buffer[0] << 3) & 0xf8);
				dest[0] += (dest[2] >> 5);

				dest[1] = ((buffer[0] & 0xe0) >> 2) + ((buffer[1] & 0x03) << 6);
				dest[1] += (dest[1] >> 5);

				dest[2] = ((buffer[1] << 1) & 0xf8);
				dest[2] += (dest[0] >> 5);

				buffer+=2;
				dest+=4;
			}
			break;
		case 24:	// 24 > 32bit
			for (x = 0; x < width; x++)
			{
				dest[0] = buffer[0];
				dest[1] = buffer[1];
				dest[2] = buffer[2];
				dest[3] = 0;
				buffer+=3;
				dest+=4;
			}
			break;
		case 32:  // ONE to ONE
			memcpy (row, buffer, info->width * info->bytes);
			break;
		}
#endif
		break;

	case TGA_TYPE_GRAY:
#ifdef INDEX_COLORS
		memcpy (row, buffer, info->width * info->bytes);
#endif
#ifdef DIRECT_COLORS
		for (x = 0; x < width; x++)
		{
			*((short *)dest) = FGDirectColor(palette[*buffer]);
			buffer++;
			dest+=2;
		}
#endif
#ifdef TRUE_COLORS
		for (x = 0; x < width; x++)
		{
			*(unsigned *)dest = palette[*buffer];
			buffer++;
			dest+=4;
		}
#endif
		break;
	case TGA_TYPE_MAPPED:
#ifdef TRUE_COLORS
			for (x = 0; x < width; x++)
			{
				*(unsigned *)dest = palette[*buffer];
				buffer++;
				dest+=4;
			}
#endif
#ifdef DIRECT_COLORS
		for (x = 0; x < width; x++)
		{
			*((short *)dest) = FGDirectColor(palette[*buffer]);
			buffer++;
			dest+=2;
		}
#endif
#ifdef INDEX_COLORS
		// this works only for BPP8, either this point will be never reached
		memcpy (row, buffer, width * info->bytes);
#endif
		break;
	}
}

//
// tga helper
//
static int ReadImage(FILE * fp, tga_info * info, char * filename, unsigned char *tga_cmap, FGPixel *data)
{
	unsigned char *buffer;
 	FGPixel *row;
	int i, y;

	unsigned int cmap_bytes;
	/* Handle colormap */

	if (info->colorMapType == 1)
	{
		memset(tga_cmap, 0, sizeof(tga_cmap));
		cmap_bytes = (info->colorMapSize + 7) / 8;
		if (cmap_bytes <= 4)
		{
			for (i=0; i<info->colorMapLength; i++)
				fread(tga_cmap+(i*4), cmap_bytes, 1, fp);
		}
		else
		{
			::printf("TGA: File is truncated or corrupted \"%s\"\n", filename);
			return -1;
		}
	}

	/* Allocate the data buffer, with file BPP */
	buffer = (unsigned char *) malloc (info->width * info->bytes);

	if (info->flipVert)
		row = data + (info->width * info->height);
	else
		row = data;

	for (y = 0; y < info->height; ++y)
	{
		if (info->flipVert)
			row -= info->width;

		// read the one line
		read_line(fp, row, buffer, info, (unsigned *)tga_cmap);
//::printf("row = %x\n",row);
		if (!info->flipVert)
			row += info->width;
	}
	free(buffer);
	return 0;
}

int Bitmap::tga(char *filename, char *gifp, int& one, unsigned int& bits, unsigned int& c)
{
	FILE *fp;
	char nm[128];
//	int counter, byte, data;
	register unsigned char *src;

	tga_info info;
	unsigned char header[18],
		footer[26],
		extension[495];
	long offset;

	memset(&info, 0, sizeof(info));
	fp = fopen(filename, "rb");
	if (!fp)
	{
		sprintf(nm, "Bitmap '%s' not found!", filename);
		IError(nm, 0);
		return 1;
	}

	if (!fseek(fp, -26L, SEEK_END))
	{							/* Is file big enough for a footer? */
		if (fread(footer, sizeof(footer), 1, fp) != 1)
		{
			::printf("TGA: Cannot read footer from %s\n", filename);
			return -1;
		}
		else if (memcmp(footer + 8, magic, sizeof(magic)) == 0)
		{
			/* Check the signature. */
			offset = footer[0] + (footer[1] * 256) + (footer[2] * 65536)
				+ (footer[3] * 16777216);

			if (fseek(fp, offset, SEEK_SET) ||
				fread(extension, sizeof(extension), 1, fp) != 1)
			{
				::printf("TGA: Cannot read extension from \"%s\"\n", filename);
				return -1;
			}

			/* Eventually actually handle version 2 TGA here */
		}
	}

	if (fseek(fp, 0, SEEK_SET) || fread(header, sizeof(header), 1, fp) != 1)
	{
		::printf("TGA: Cannot read header from \"%s\"\n", filename);
		return -1;
	}

	switch (header[2])
	{
		case 1:
			info.imageType = TGA_TYPE_MAPPED;
			info.imageCompression = TGA_COMP_NONE;
			break;
		case 2:
			info.imageType = TGA_TYPE_COLOR;
			info.imageCompression = TGA_COMP_NONE;
			break;
		case 3:
			info.imageType = TGA_TYPE_GRAY;
			info.imageCompression = TGA_COMP_NONE;
			break;

		case 9:
			info.imageType = TGA_TYPE_MAPPED;
			info.imageCompression = TGA_COMP_RLE;
			break;
		case 10:
			info.imageType = TGA_TYPE_COLOR;
			info.imageCompression = TGA_COMP_RLE;
			break;
		case 11:
			info.imageType = TGA_TYPE_GRAY;
			info.imageCompression = TGA_COMP_RLE;
			break;

		default:
			info.imageType = 0;
	}

	info.idLength = header[0];
	info.colorMapType = header[1];

	info.colorMapIndex = header[3] + header[4] * 256;
	info.colorMapLength = header[5] + header[6] * 256;
	info.colorMapSize = header[7];

	info.xOrigin = header[8] + header[9] * 256;
	info.yOrigin = header[10] + header[11] * 256;
	info.width = header[12] + header[13] * 256;
	info.height = header[14] + header[15] * 256;

	bits = info.bpp = header[16];
	info.bytes = (info.bpp + 7) / 8;
	info.alphaBits = header[17] & 0x0f;	/* Just the low 4 bits */
	info.flipHoriz = (header[17] & 0x10) ? 1 : 0;
	info.flipVert = (header[17] & 0x20) ? 0 : 1;

	switch (info.imageType)
	{
		case TGA_TYPE_MAPPED:
			if (info.bpp != 8)
			{
				::printf("TGA: Unhandled sub-format in \"%s\"\n", filename);
				return -1;
			}
			break;
		case TGA_TYPE_COLOR:
			if (info.alphaBits == 0 && info.bpp == 32)
			{
				::printf("TGA: Possibly incorrect alpha in \"%s\"\n",
						  filename);
				info.alphaBits = 8;
			}
			if (info.bpp != 16 && info.bpp != 24
				&& (info.alphaBits != 8 || info.bpp != 32))
			{
				::printf("TGA: Unhandled sub-format in \"%s\"\n", filename);
				return -1;
			}
			break;
		case TGA_TYPE_GRAY:
			if (info.alphaBits == 0 && info.bpp == 16)
			{
				::printf("TGA: Possibly incorrect alpha in \"%s\"\n",
						  filename);
				info.alphaBits = 8;
			}
			if (info.bpp != 8 && (info.alphaBits != 8 || info.bpp != 16))
			{
				::printf("TGA: Unhandled sub-format in \"%s\"\n", filename);
				return -1;
			}
			break;

		default:
			::printf("TGA: Unknown image type for \"%s\"\n", filename);
			return -1;
	}

	/* Plausible but unhandled formats */
	if (info.bytes * 8 != info.bpp)
	{
		::printf("TGA: No support yet for TGA with these parameters\n");
		return -1;
	}

	/* Check that we have a color map only when we need it. */
	if (info.imageType == TGA_TYPE_MAPPED && info.colorMapType != 1)
	{
		::printf("TGA: indexed image has invalid color map type %d\n",
				  info.colorMapType);
		return -1;
	}
	else if (info.imageType != TGA_TYPE_MAPPED && info.colorMapType != 0)
	{
		::printf("TGA: non-indexed image has invalid color map type %d\n",
				  info.colorMapType);
		return -1;
	}
#ifdef INDEX_COLORS
	else if (info.imageType == TGA_TYPE_COLOR)
	{
		::printf("TGA: non-indexed image not supported with BPP==8\n");
		return -1;
	}
#endif
	/* Skip the image ID field. */
	if (info.idLength && fseek(fp, info.idLength, SEEK_CUR))
	{
		::printf("TGA: File is truncated or corrupted \"%s\"\n", filename);
		return -1;
	}

	w = info.width;
	h = info.height;
	bits = info.bpp;

	if (info.bpp>8)
	{
		c = 0;
		one = 0;
	}
	else
	{
		c = 256;
		one = 4;
	}
	strcpy(gifp, ".bmp");
	name = new char[strlen(filename) + 1];
	strcpy(name, filename);
	
	assert(SetArray((FGPixel *) calloc(areasize(w, h), 1)));
	src = (unsigned char *)GetArray();

	// generate gray ramp
	if (info.imageType == TGA_TYPE_GRAY)
		for(int i=0; i<(int)c; i++)
			*(unsigned *)(&rgb.rgb4[i]) = i+(i<<8)+(i<<16);

	ReadImage(fp, &info, filename, (unsigned char *)&rgb, GetArray());

	// mark as already converted
	bits = 24;
	fclose(fp);
	return 0;
}

int Bitmap::gif(char *n, char *gifp, int& one, unsigned int& bits, unsigned int& c)
{
	GIFStream *gif=0;
	char nm[128];
	
		gif = GIFRead(n);
		if (gif)
		{
			w = gif->width;
			h = gif->height;
			one = 3;
			c = gif->cmapSize;
			bits = 8;
			if (c) copy_gif_palette((unsigned char *)&rgb, gif->cmapData[0], c);
			assert(SetArray((FGPixel *) malloc(areasize(w, h))));
			memset(GetArray(), gif->background, areasize(w, h));
			GIFData *d = gif->data;
			while(d)
			{
				if (d->type == gif_image)
				{
					if (d->data.image.cmapSize)
					{
						c = d->data.image.cmapSize;
						if (c) copy_gif_palette((unsigned char *)&rgb, d->data.image.cmapData[0], c);
					}
					colorkey = d->info.transparent;
					memcpy(GetArray(), d->data.image.data, d->width*d->height);
					free(d->data.image.data);
				}
				else if(d->type == gif_text || d->type == gif_comment)
				{
					if (d->data.comment.text) free(d->data.comment.text);
				}
				GIFData *pom = d;
				d = d->next;
				free(pom);
			}
			strcpy(gifp, ".bmp");
			name = new char[strlen(n) + 1];
			strcpy(name, n);
			free(gif);
		}
		else
		{
			sprintf(nm, "GIF file '%s' loading error!", n);
			IError(nm, 0);
			return 1;
		}
		return 0;
}

/**
	Perform image rectangle copy.

*/
int DrawBuffer::bitblit(int xdst, int ydst, int xsrc, int ysrc, int ww, int hh, DrawBuffer *src, int opcia)
{
	int ppp = _ppop, oldt = transp_color;
	int _ww=src->w, _hh=src->h;

	if (src->image == 0 || src->type == BMP_NONE || image==0)
		return 0;

	xsrc += src->xoff;
	ysrc += src->yoff;
	xdst += xoff;
	ydst += yoff;

	// compute the tile size
	if (src->clip(xsrc, ysrc, _ww, _hh)==0) return 0;
	if (ww<1 || hh<1) return 0;

	// test for transparency ... and force one
	if (src->GetAlpha())
		opcia = (src->GetAlpha() | BITBLIT_WITH_OPERATOR);

	if (src->GetColorKey() != -1)
	{
		ppp = _GCOLORKEY;
		transp_color = src->GetColorKey();
		opcia |= BITBLIT_WITH_OPERATOR;
	}
	for (int i = 0; i < hh;)
	{
		for (int j = 0; j < ww;)
		{
			int _x=xdst+j,
				_y=ydst+i,
				_w=j+_ww>ww?ww-j:_ww,
				_h=i+_hh>hh?hh-i:_hh;

			if (clip(_x, _y, _w, _h))
				RamToRam(xsrc+(_x-(xdst+j)), ysrc+(_y-(ydst+i)), src->w, src->h,
					_x, _y, w, h,
					_w, _h, src->image, image, opcia, ppp);
			j += _ww;
		}
		i += _hh;
	}
	transp_color = oldt;
	return 1;
}

Bitmap::~Bitmap()
{
#ifdef INDEX_COLORS
	if (cUsedTable)
	{
		for (int i = 0; i < 256; i++)
			if (cUsedTable[i] != -1)
				DeleteColor(cUsedTable[i]);
		delete cUsedTable;
	}
#endif
	if (original_name) free(original_name);
}

/**
	Create a BITMAP object in memory with specified size and color layout.
*/
Bitmap::Bitmap(int ww, int hh, FGPixel color)
: DrawBuffer(ww, hh, BMP_MEM, color)
{
	init();
}

/**
	Create a BITMAP object in memory as exact copy of the DrawBuffer object.
	This one has dimensions and contents the same as source buffer.
	The image has been copied from, that you can't expect that any next changes
	are appears in it. The application of one are when you want save
	the contents of a window to the BMP file by example.
*/
Bitmap::Bitmap(DrawBuffer * wPtr)
:DrawBuffer(wPtr->GetW(), wPtr->GetH())
{
	init();
	memcpy(image, wPtr->GetArray(), areasize(wPtr->GetW(), wPtr->GetH()));
}

void Bitmap::init(void)
{
	cUsedTable = 0;
	name = original_name = 0;
	memset(&bmih, 0, sizeof(bmih));
}

/**
	Save a BITMAP object into the current directory.
	The file will be standard Windows BMP file. Extension
	is added automagically.
	Return value is false if operation fail. On linux platform
	is file saved to PNG format also, if imagemagic is installed.
*/
int Bitmap::BitmapSave(char *nazov)
{
	if (GetType() == BMP_NONE)
		return 0;
	char *tmpname = new char[strlen(nazov) + 16];

#ifndef INDEX_COLORS
	int j, rest;
	char *bmp2;

#endif
	nazov = strdup(nazov);
#if defined (__MSDOS__) || defined (_WIN32)
	strlwr(nazov);
#endif
	if (strstr(nazov, ".bmp") || strstr(nazov, ".BMP"))
		strcpy(tmpname, nazov);
	else
		sprintf(tmpname, "%s.bmp", nazov);
	bmfh.usType = 'M' * 256 + 'B';
	bmfh.reserve1 = 0;
	bmfh.reserve2 = 0;
	bmi.reserve = 0;
	bmih.bmih2.size = 0x28;
	bmih.bmih2.cx = w;
	bmih.bmih2.cy = h;
	bmih.bmih2.cPlanes = 1;
	bmih.bmih2.ulCompression = 0;
	bmih.bmih2.cbImage = w * h;
	bmih.bmih2.cxResolution = 0;
	bmih.bmih2.cyResolution = 0;
#ifdef INDEX_COLORS
	bmi.imageOffset = 0x36 + 0x400;
	bmfh.cbSizeLow = w * h + 1024 + 0x36;
	bmfh.cbSizeHigh = (w * h + 1024 + 0x36) >> 16;
	bmih.bmih2.cclrUsed = 256;
	bmih.bmih2.cclrImportant = 256;
	bmih.bmih2.cBitCount = 8;
	memcpy(&rgb, &_fgl_palette, sizeof(rgb));
#else
	bmi.imageOffset = 0x36;
	bmfh.cbSizeLow = w * h + 0x36;
	bmfh.cbSizeHigh = (w * h + 0x36) >> 16;
	bmih.bmih2.cclrUsed = 0;
	bmih.bmih2.cclrImportant = 0;
	bmih.bmih2.cBitCount = 24;
#endif // index
	FILE *fp = fopen(tmpname, "wb");

	if (fp == 0)
	{
		IError("Bitmap not saved because of disk write error!", 0);
		delete tmpname;
		return 0;
	}
	assert(sizeof(bmfh) == 10);
	fwrite(&bmfh, sizeof(bmfh), 1, fp);
	fwrite(&bmi, sizeof(bmi), 1, fp);
	fwrite(&bmih, sizeof(bmih.bmih2), 1, fp);
#ifdef INDEX_COLORS
	fwrite(&rgb, sizeof(rgb.rgb4), 1, fp);
	turn_bitmap(h / 2, image, image + w * (h - 1));
	fwrite(image, w * h, 1, fp);
	turn_bitmap(h / 2, image, image + w * (h - 1));
#else // index
#if (FASTGL_BPP == 15 || FASTGL_BPP == 16)
	unsigned long cc, aa;
#endif
	for (int i = h - 1; i >= 0; i--)
	{
		bmp2 = (char *) (image + w * i);
		for (j = 0; j < w; j++)
		{
#ifdef TRUE_COLORS
			fwrite(bmp2, 3, 1, fp);
			bmp2 += 4;
#else
			cc = *(FGPixel *) bmp2;
#if (FASTGL_BPP == 15)
			aa = ((cc << 9) & 0xF80000) | ((cc << 6) & 0xF800) | ((cc << 3) & 0xF8);
#else
			aa = ((cc << 8) & 0xF80000) | ((cc << 5) & 0xFc00) | ((cc << 3) & 0xF8);
#endif // direct
			fwrite(&aa, 3, 1, fp);
			bmp2 += 2;
#endif // true
		}
#ifdef TRUE_COLORS
		rest = (w * 3) % 4;
#else
		rest = (w * 2) % 4;
#endif // true
		while (rest & 3)
		{
			rest++;
			fputc(0, fp);
		}
	}
#endif
	fclose(fp);
#ifdef __linux__
	char ss[64];
	sprintf(ss, "/usr/X11R6/bin/convert %s %s.png", tmpname, nazov);
	system(ss);
//	unlink(tmpname);
#endif
	free(nazov);
	delete tmpname;
	return 1;
}

/* +-------------------------------------------------------------------+ */
/* | Copyright 1990 - 1994, David Koblas. (koblas@netcom.com)          | */
/* |   Permission to use, copy, modify, and distribute this software   | */
/* |   and its documentation for any purpose and without fee is hereby | */
/* |   granted, provided that the above copyright notice appear in all | */
/* |   copies and that both that copyright notice and this permission  | */
/* |   notice appear in supporting documentation.  This software is    | */
/* |   provided "as is" without express or implied warranty.           | */
/* +-------------------------------------------------------------------+ */

GIFStream *GIFReadFP(FILE * fd)
{
	unsigned char buf[256];
	unsigned char c;
	GIFStream *stream=0;
	GIFData *cur, **end;
	GIF89info info;
	int resetInfo = TRUE;
	int n;

	if (fd == NULL)
		return NULL;

	if (setjmp(setjmp_buffer))
		goto out;

	if (!ReadOK(fd, buf, 6))
		_ERROR("error reading magic number");

	if (strncmp((char *) buf, "GIF", 3) != 0)
		_ERROR("not a GIF file");

	if ((strncmp((const char *) buf + 3, "87a", 3) != 0) &&
		(strncmp((const char *) buf + 3, "89a", 3) != 0))
		_ERROR("bad version number, not '87a' or '89a'");

	if (!ReadOK(fd, buf, 7))
		_ERROR("failed to read screen descriptor");

	stream = NEW(GIFStream);

	stream->width = MKINT(buf[0], buf[1]);
	stream->height = MKINT(buf[2], buf[3]);

	stream->cmapSize = 2 << (buf[4] & 0x07);
	stream->colorMapSize = stream->cmapSize;
	stream->colorResolution = ((int) (buf[4] & 0x70) >> 3) + 1;
	stream->background = buf[5];
	stream->aspectRatio = buf[6];

	stream->data = NULL;

	end = &stream->data;

	/*
	   **  Global colormap is present.
	 */
	if (BitSet(buf[4], LOCALCOLORMAP))
	{
		if (readColorMap(fd, stream->cmapSize, stream->cmapData))
			_ERROR("unable to get global colormap");
	}
	else
	{
		stream->cmapSize = 0;
		stream->background = -1;
	}

	if (stream->aspectRatio != 0 && stream->aspectRatio != 49)
	{
		float r;

		r = ((float) stream->aspectRatio + 15.0) / 64.0;
		INFO_MSG(("warning - non-square pixels; to fix do a 'pnmscale -%cscale %g'",
				  r < 1.0 ? 'x' : 'y',
				  r < 1.0 ? 1.0 / r : r));
	}

	while (ReadOK(fd, &c, 1) && c != ';')
	{
		if (resetInfo)
		{
			info.disposal = (GIFDisposalType) 0;
			info.inputFlag = 0;
			info.delayTime = 0;
			info.transparent = -1;
			resetInfo = FALSE;
		}
		cur = NULL;

		if (c == '!')
		{						/* Extension */
			if (!ReadOK(fd, &c, 1))
				_ERROR("EOF / read error on extention function code");
			if (c == 0xf9)
			{					/* graphic control */
				(void) GetDataBlock(fd, buf);
				info.disposal = (GIFDisposalType) ((buf[0] >> 2) & 0x7);
				info.inputFlag = (buf[0] >> 1) & 0x1;
				info.delayTime = MKINT(buf[1], buf[2]);
				if (BitSet(buf[0], 0x1))
					info.transparent = buf[3];

				while (GetDataBlock(fd, buf) != 0)
					;
			}
			else if (c == 0xfe || c == 0x01)
			{
				int len = 0;
				int size = 256;
				char *text;

				/*
				   **  Comment or Plain Text
				 */

				cur = NEW(GIFData);

				if (c == 0x01)
				{
					(void) GetDataBlock(fd, buf);

					cur->type = gif_text;
					cur->info = info;
					cur->x = MKINT(buf[0], buf[1]);
					cur->y = MKINT(buf[2], buf[3]);
					cur->width = MKINT(buf[4], buf[5]);
					cur->height = MKINT(buf[6], buf[7]);

					cur->data.text.cellWidth = buf[8];
					cur->data.text.cellHeight = buf[9];
					cur->data.text.fg = buf[10];
					cur->data.text.bg = buf[11];

					resetInfo = TRUE;
				}
				else
				{
					cur->type = gif_comment;
				}

				text = (char *) malloc(size);

				while ((n = GetDataBlock(fd, buf)) != 0)
				{
					if (n + len >= size)
						text = (char *) realloc(text, size += 256);
					memcpy(text + len, buf, n);
					len += n;
				}

				if (c == 0x01)
				{
					cur->data.text.len = len;
					cur->data.text.text = text;
				}
				else
				{
					cur->data.comment.len = len;
					cur->data.comment.text = text;
				}
			}
			else
			{
				/*
				   **  Unrecogonized extension, consume it.
				 */
				while (GetDataBlock(fd, buf) > 0)
					;
			}
		}
		else if (c == ',')
		{
			if (!ReadOK(fd, buf, 9))
				_ERROR("couldn't read left/top/width/height");

			cur = NEW(GIFData);

			cur->type = gif_image;
			cur->info = info;
			cur->x = MKINT(buf[0], buf[1]);
			cur->y = MKINT(buf[2], buf[3]);
			cur->width = MKINT(buf[4], buf[5]);
			cur->height = MKINT(buf[6], buf[7]);
			cur->data.image.cmapSize = 1 << ((buf[8] & 0x07) + 1);
			if (BitSet(buf[8], LOCALCOLORMAP))
			{
				if (readColorMap(fd, cur->data.image.cmapSize,
								 cur->data.image.cmapData))
					_ERROR("unable to get local colormap");
			}
			else
			{
				cur->data.image.cmapSize = 0;

			}
			cur->data.image.data = (unsigned char *) malloc(cur->width * cur->height);
			cur->data.image.interlaced = BitSet(buf[8], INTERLACE);
			readImage(fd, BitSet(buf[8], INTERLACE),
					  cur->width, cur->height, cur->data.image.data);

			resetInfo = TRUE;
		}
		else
		{
			INFO_MSG(("bogus character 0x%02x, ignoring", (int) c));
		}

		if (cur != NULL)
		{
			*end = cur;
			end = &cur->next;
			cur->next = NULL;
		}
	}

	if (c != ';')
		_ERROR("EOF / data stream");

  out:
	return stream;
}

GIFStream *GIFRead(char *file)
{
	FILE *fp = fopen(file, "rb");
	GIFStream *stream = NULL;

	if (fp != NULL)
	{
		stream = GIFReadFP(fp);
		fclose(fp);
	}
	return stream;
}

static int readColorMap(FILE * fd, int size,
						unsigned char data[GIF_MAXCOLORS][3])
{
	int i;
	unsigned char rgb[3 * GIF_MAXCOLORS];
	unsigned char *cp = rgb;

	if (!ReadOK(fd, rgb, size * 3))
		return TRUE;

	for (i = 0; i < size; i++)
	{
		data[i][0] = *cp++;
		data[i][1] = *cp++;
		data[i][2] = *cp++;
	}

	return FALSE;
}

/*
   **
 */

static int ZeroDataBlock = FALSE;

static int GetDataBlock(FILE * fd, unsigned char *buf)
{
	unsigned char count;

	if (!ReadOK(fd, &count, 1))
	{
		INFO_MSG(("error in getting DataBlock size"));
		return -1;
	}

	ZeroDataBlock = count == 0;
	if ((count != 0) && (!ReadOK(fd, buf, count)))
	{
		INFO_MSG(("error in reading DataBlock"));
		return -1;
	}
	return count;
}

/*
   **
   **
 */

/*
   **  Pulled out of nextCode
 */
static int curbit, lastbit, get_done, last_byte;
static int return_clear;

/*
   **  Out of nextLWZ
 */
static int stack[(1 << (MAX_LWZ_BITS)) * 2], *sp;
static int code_size, set_code_size;
static int max_code, max_code_size;
static int clear_code, end_code;

static void initLWZ(int input_code_size)
{
	set_code_size = input_code_size;
	code_size = set_code_size + 1;
	clear_code = 1 << set_code_size;
	end_code = clear_code + 1;
	max_code_size = 2 * clear_code;
	max_code = clear_code + 2;

	curbit = lastbit = 0;
	last_byte = 2;
	get_done = FALSE;

	return_clear = TRUE;

	sp = stack;
}

static int nextCode(FILE * fd, int code_size)
{
	static unsigned char buf[280];
	static int maskTbl[16] =
	{
		0x0000, 0x0001, 0x0003, 0x0007,
		0x000f, 0x001f, 0x003f, 0x007f,
		0x00ff, 0x01ff, 0x03ff, 0x07ff,
		0x0fff, 0x1fff, 0x3fff, 0x7fff,
	};
	int i, j, ret, end;

	if (return_clear)
	{
		return_clear = FALSE;
		return clear_code;
	}

	end = curbit + code_size;

	if (end >= lastbit)
	{
		int count;

		if (get_done)
		{
			if (curbit >= lastbit)
				_ERROR("ran off the end of my bits");
			return -1;
		}
		buf[0] = buf[last_byte - 2];
		buf[1] = buf[last_byte - 1];

		if ((count = GetDataBlock(fd, &buf[2])) == 0)
			get_done = TRUE;

		last_byte = 2 + count;
		curbit = (curbit - lastbit) + 16;
		lastbit = (2 + count) * 8;

		end = curbit + code_size;
	}

	j = end / 8;
	i = curbit / 8;

	if (i == j)
		ret = buf[i];
	else if (i + 1 == j)
		ret = buf[i] | (buf[i + 1] << 8);
	else
		ret = buf[i] | (buf[i + 1] << 8) | (buf[i + 2] << 16);

	ret = (ret >> (curbit % 8)) & maskTbl[code_size];

	curbit += code_size;

	return ret;
}

#define readLWZ(fd) ((sp > stack) ? *--sp : nextLWZ(fd))

static int nextLWZ(FILE * fd)
{
	static int table[2][(1 << MAX_LWZ_BITS)];
	static int firstcode, oldcode;
	int code, incode;
	register int i;

	while ((code = nextCode(fd, code_size)) >= 0)
	{
		if (code == clear_code)
		{
			for (i = 0; i < clear_code; ++i)
			{
				table[0][i] = 0;
				table[1][i] = i;
			}
			for (; i < (1 << MAX_LWZ_BITS); ++i)
				table[0][i] = table[1][i] = 0;
			code_size = set_code_size + 1;
			max_code_size = 2 * clear_code;
			max_code = clear_code + 2;
			sp = stack;
			do
			{
				firstcode = oldcode = nextCode(fd, code_size);
			}
			while (firstcode == clear_code);

			return firstcode;
		}
		if (code == end_code)
		{
			int count;
			unsigned char buf[260];

			if (ZeroDataBlock)
				return -2;

			while ((count = GetDataBlock(fd, buf)) > 0)
				;

//			if (count != 0)
//				INFO_MSG(("missing EOD in data stream"));
			return -2;
		}

		incode = code;

		if (code >= max_code)
		{
			*sp++ = firstcode;
			code = oldcode;
		}

		while (code >= clear_code)
		{
			*sp++ = table[1][code];
			if (code == table[0][code])
				_ERROR("circular table entry BIG ERROR");
			code = table[0][code];
		}

		*sp++ = firstcode = table[1][code];

		if ((code = max_code) < (1 << MAX_LWZ_BITS))
		{
			table[0][code] = oldcode;
			table[1][code] = firstcode;
			++max_code;
			if ((max_code >= max_code_size) &&
				(max_code_size < (1 << MAX_LWZ_BITS)))
			{
				max_code_size *= 2;
				++code_size;
			}
		}

		oldcode = incode;

		if (sp > stack)
			return *--sp;
	}
	return code;
}

static void readImage(FILE * fd, int interlace, int width, int height,
					  unsigned char *data)
{
	unsigned char *dp, c;
	int v, xpos = 0, ypos = 0;

	/*
	   **  Initialize the Compression routines
	 */
	if (!ReadOK(fd, &c, 1))
		_ERROR("EOF / read error on image data");

	initLWZ(c);

	if (verbose)
		INFO_MSG(("reading %d by %d%s GIF image",
				  width, height, interlace ? " interlaced" : ""));

	if (interlace)
	{
		int i;
		int pass = 0, step = 8;

		for (i = 0; i < height; i++)
		{
			dp = &data[width * ypos];
			for (xpos = 0; xpos < width; xpos++)
			{
				if ((v = readLWZ(fd)) < 0)
					goto fini;

				*dp++ = v;
			}
			if ((ypos += step) >= height)
			{
				do
				{
					if (pass++ > 0)
						step /= 2;
					ypos = step / 2;
				}
				while (ypos > height);
			}
		}
	}
	else
	{
		dp = data;
		for (ypos = 0; ypos < height; ypos++)
		{
			for (xpos = 0; xpos < width; xpos++)
			{
				if ((v = readLWZ(fd)) < 0)
					goto fini;

				*dp++ = v;
			}
		}
	}

  fini:
	if (readLWZ(fd) >= 0)
		INFO_MSG(("too much input data, ignoring extra..."));

	return;
}

DrawBuffer::DrawBuffer(int ws, int hs, char *nm, Type typ,  FGPixel i, FGPixel p)
{
	_init(ws, hs, typ, i, p, nm);
}

/**
	Create a DRAWBUFFER object.
	Parameter 'ww' and 'hh' are width and height of rectangle, 'buf'
	is pointed to the image data taht will be copied into created object.
*/
DrawBuffer::DrawBuffer(int ww, int hh, FGPixel * buf)
{
	_init(ww, hh, BMP_MEM, 0);
	Resize(0, 0);
	memcpy(image, buf, areasize(w,h));
}

/**                    
	Creates the object with specified size, type, color layout or already existing pixel data.
*/
DrawBuffer::DrawBuffer(int ww, int hh, Type t, int color, FGPixel * buf)
{
	_init(ww, hh, t, color);
	if (ww > 0 && hh > 0)
	{
		if (!buf)
		{
			Resize(0, 0);
			clear(color);
		}
		else image = buf;
	}
	else image = 0;
}

/**
	Release object's members.
*/
DrawBuffer::~DrawBuffer()
{
	if (name && name != empty_string) delete []name;
	if (type && image && type!=FGFRAMEBUFFER)
	{
		if (type != BMP_IMAGE)
			free(image);
	}
}

void DrawBuffer::_init(int ww, int hh, Type t, FGPixel ii,  FGPixel pp, char *nm)
{
	lock();
	id = ID_counter++;
	unlock();
	if (nm)
	{
		assert(name	 = new char[strlen(nm)+1]);
		strcpy(name,nm);
	}
	else
	{
		name = (char *)empty_string;
	}
	image = 0;
	alpha = 0;
	colorkey = -1;
	w = ww;
	h = hh;
	xoff=yoff=0;
	wwrk=ww;
	hwrk=hh;
	type = t;
	status = WHIDEN|WLOCKED;
	_ink = ii;
	_paper = pp;
	_ppop = _GSET;
	set_font(FONT0816);
}

inline void DrawBuffer::modify_point(FGPixel *ptr, FGPixel ink)
{
	switch (_ppop)
	{
		default:
		case 0:
			*ptr = ink;
			break;
		case 1:					// xor
			*ptr^= ink;
			break;
		case 2:					// and
			*ptr &= ink;
			break;
		case 3:					// or
			*ptr |= ink;
			break;
		case 4:					// plus
			*ptr += ink;
			break;
		case 5:					// minus
			*ptr -= ink;
			break;
		case 6:					// not
			*ptr ^= -1;
			break;
		case 7:					// not
			if (*ptr>=16) *ptr = ink;
			break;
	}
}

/**
	Draw a horizontal line from [x:y] with size 'c' and color 'ink'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::hline(int x, int y, int c, FGPixel ink)
{
	FGPixel *ptr;
	int	j,stuff=1;

	x += xoff;
	y += yoff;

	if (clip(x,y,c,stuff) == 0)
		return 0;

	ptr	= image	+ w * y + x;

	switch(_ppop)
	{
		default:
		case 0:	 // set
			FGmemset(ptr, ink, c);
			break;
		case 1:	 // xor
			for	(j = c;	j; j--)
			{
				*ptr++ ^= ink;
			}
			break;
		case 2:	 // and
			for	(j = c;	j; j--)
			{
				*ptr++ &= ink;
			}
			break;
		case 3:	 // or
			for	(j = c;	j; j--)
			{
				*ptr++ |= ink;
			}
			break;
		case 4:	 // plus
			for	(j = c;	j; j--)
			{
				*ptr++ += ink;
			}
			break;
		case 5:	 // minus
			for	(j = c;	j; j--)
			{
				*ptr++ -= ink;
			}
			break;
		case 6:	 // not
			for	(j = c;	j; j--)
			{
				*ptr++ ^= -1;
			}
			break;
		case 7:	 // replace
			for	(j = c;	j; j--)
			{
				if (*(ptr) >= 16) *ptr = ink;
				ptr++;
			}
			break;
	}
	return 1;
}

/**
	Draw a vertical line from [x:y] with size 'c' and color 'ink'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::vline(int x, int y, int c, FGPixel ink)
{
	FGPixel *ptr;
	int	i, stuff=1;

	x += xoff;
	y += yoff;

	if (clip(x,y,stuff,c) == 0)
		return 0;

	ptr	= image	+ w * y + x;

	for	(i = c;	i; i--)
	{
		modify_point(ptr, ink);
		ptr	+= w;
	}
	return 1;
}

/**
	Draw a pixel at position [x:y] with current object foreground color.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinates are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::putpixel(int x, int y)
{
	x += xoff;
	y += yoff;

	if (y <	yoff || x < xoff || y >= (hwrk+yoff) || x >= (wwrk+xoff))
		return 0;
	modify_point(image + w * y + x, _ink);
	return 1;
}

/**
	Draw a pixel at position [x:y] with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinates are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::putpixel(int x, int y, FGPixel color)
{
	x += xoff;
	y += yoff;

	if (y <	yoff || x < xoff || y >= (hwrk+yoff) || x >= (wwrk+xoff))
		return 0;
	modify_point(image + w * y + x, color);
	return 1;
}

// internal for circle draw to RAM
void	DrawBuffer::_symetry(int xs, int ys, int x,	int	y, FGPixel color)
{
	putpixel(xs + x, ys + y, color);
	putpixel(xs - x, ys + y, color);
	putpixel(xs + x, ys - y, color);
	putpixel(xs - x, ys - y, color);
	putpixel(xs + y, ys + x, color);
	putpixel(xs - y, ys + x, color);
	putpixel(xs + y, ys - x, color);
	putpixel(xs - y, ys - x, color);
}

// internal for circle fill to RAM
void	DrawBuffer::_symetry2(int xs, int ys, int x, int y, FGPixel color)
{
	hline(xs-x,ys+y,x+x+1, color);
	hline(xs-x,ys-y,x+x+1, color);
	hline(xs-y,ys+x,y+y+1, color);
	hline(xs-y,ys-x,y+y+1, color);
}

/**
	Draw a circle shape at position [x:y] with radius and with
	current object foreground color.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinates are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::circle(int	xs,	int	ys,	int	r)
{
	return circle(xs,ys,r,_ink);
}

/**
	Draw a circle shape at position [x:y] with radius and with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::circle(int	xs,	int	ys,	int	r, FGPixel color)
{
	int	x =	0;
	int	y =	r;
	int	p =	3 -	2 *	r;

	if ((ys	+ r) < 0 || (xs +	r) < 0
		||	(ys	- r) >=	h || (xs -	r) >= w
		||	r <	1) return 0;

	while (x < y)
	{
		_symetry(xs, ys, x,	y, color);
		if (p <	0)
			p += 4 * (x++) + 6;
		else
			p += 4 * ((x++)	- (y--)) + 10;
	}
	if (x == y)
		_symetry(xs, ys, x,	y, color);
	return 1;
}

/**
	Fill a circle shape at position [x:y] with radius and with
	current object foreground color.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::fcircle(int	xs,	int	ys,	int	r)
{
	return fcircle(xs,ys,r,_ink);
}

/**
	Fill a circle shape at position [x:y] with radius and with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::fcircle(int	xs,	int	ys,	int	r, FGPixel color)
{
	int	x =	0;
	int	y =	r;
	int	p =	3 -	2 *	r;

	if ((ys	+ r) < 0 || (xs +	r) < 0
		||	(ys	- r) >=	h || (xs -	r) >= w
		||	r <	1) return 0;

	while (x < y)
	{
		_symetry2(xs, ys, x, y, color);
		if (p <	0)
			p += 4 * (x++) + 6;
		else
			p += 4 * ((x++)	- (y--)) + 10;
	}
	if (x == y)
		_symetry2(xs, ys, x, y, color);
	return 1;
}

/**
	Draw an ellipse shape at position [x:y] with radius 'rx' & 'ry'and with
	current object foreground color.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::ellipse(int	x,	int	y,	int	rx, int ry)
{
	return ellipse(x,y,rx,ry,_ink);
}

/**
	Draw an ellipse shape at position [x:y] with radius 'rx' & 'ry'and with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::ellipse(int	x,	int	y,	int	rx, int ry, FGPixel color)
{
	if ((y	+ ry) < 0 || (x + rx) < 0 || (y - ry) >= h || (x	- rx) >= w || rx<1 || ry<1)
		return 0;
	int ix, iy;
	int h, i, j, k;
	int oh, oi, oj, ok;

	if (rx < 1)
		rx = 1;

	if (ry < 1)
		ry = 1;

	h = i = j = k = 0xFFFF;

	if (rx > ry)
	{
		ix = 0;
		iy = rx * 64;

		do
		{
			oh = h;
			oi = i;
			oj = j;
			ok = k;

			h = (ix + 32) >> 6;
			i = (iy + 32) >> 6;
			j = (h * ry) / rx;
			k = (i * ry) / rx;

			if (((h != oh) || (k != ok)) && (h < oi))
			{
				putpixel(x + h, y + k, color);
				if (h)
					putpixel(x - h, y + k, color);
				if (k)
				{
					putpixel(x + h, y - k, color);
					if (h)
						putpixel(x - h, y - k, color);
				}
			}

			if (((i != oi) || (j != oj)) && (h < i))
			{
				putpixel(x + i, y + j, color);
				if (i)
					putpixel(x - i, y + j, color);
				if (j)
				{
					putpixel(x + i, y - j, color);
					if (i)
						putpixel(x - i, y - j, color);
				}
			}

			ix = ix + iy / rx;
			iy = iy - ix / rx;
		}
		while (i > h);
	}
	else
	{
		ix = 0;
		iy = ry * 64;

		do
		{
			oh = h;
			oi = i;
			oj = j;
			ok = k;

			h = (ix + 32) >> 6;
			i = (iy + 32) >> 6;
			j = (h * rx) / ry;
			k = (i * rx) / ry;

			if (((j != oj) || (i != oi)) && (h < i))
			{
				putpixel(x + j, y + i, color);
				if (j)
					putpixel(x - j, y + i, color);
				if (i)
				{
					putpixel(x + j, y - i, color);
					if (j)
						putpixel(x - j, y - i, color);
				}
			}

			if (((k != ok) || (h != oh)) && (h < oi))
			{
				putpixel(x + k, y + h, color);
				if (k)
					putpixel(x - k, y + h, color);
				if (h)
				{
					putpixel(x + k, y - h, color);
					if (k)
						putpixel(x - k, y - h, color);
				}
			}

			ix = ix + iy / ry;
			iy = iy - ix / ry;

		}
		while (i > h);
	}
	return 1;
}

/**
	Fill an ellipse shape at position [x:y] with radius 'rx' & 'ry'and with
	current object foreground color.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::fellipse(int	x,	int	y,	int	rx, int ry)
{
	return fellipse(x,y,rx,ry,_ink);
}

/**
	Fill an ellipse shape at position [x:y] with radius 'rx' & 'ry'and with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::fellipse(int	x,	int	y,	int	rx, int ry, FGPixel color)
{
	if ((y	+ ry) < 0 || (x + rx) < 0 || (y - ry) >= h || (x	- rx) >= w || rx<1 || ry<1)
		return 0;
	int ix, iy;
	int a, b, c, d;
	int da, db, dc, dd;
	int na, nb, nc, nd;

	if (rx > ry)
	{
		dc = -1;
		dd = 0xFFFF;
		ix = 0;
		iy = rx * 64;
		na = 0;
		nb = (iy + 32) >> 6;
		nc = 0;
		nd = (nb * ry) / rx;

		do
		{
			a = na;
			b = nb;
			c = nc;
			d = nd;

			ix = ix + (iy / rx);
			iy = iy - (ix / rx);
			na = (ix + 32) >> 6;
			nb = (iy + 32) >> 6;
			nc = (na * ry) / rx;
			nd = (nb * ry) / rx;

			if ((c > dc) && (c < dd))
			{
				hline(x - b, y + c, b * 2, color);
				if (c)
					hline(x - b, y - c, b * 2, color);
				dc = c;
			}

			if ((d < dd) && (d > dc))
			{
				hline(x - a, y + d, a * 2, color);
				hline(x - a, y - d, a * 2, color);
				dd = d;
			}

		}
		while (b > a);
	}
	else
	{
		da = -1;
		db = 0xFFFF;
		ix = 0;
		iy = ry * 64;
		na = 0;
		nb = (iy + 32) >> 6;
		nc = 0;
		nd = (nb * rx) / ry;

		do
		{
			a = na;
			b = nb;
			c = nc;
			d = nd;

			ix = ix + (iy / ry);
			iy = iy - (ix / ry);
			na = (ix + 32) >> 6;
			nb = (iy + 32) >> 6;
			nc = (na * rx) / ry;
			nd = (nb * rx) / ry;

			if ((a > da) && (a < db))
			{
				hline(x - d, y + a, d * 2, color);
				if (a)
					hline(x - d, y - a, d * 2, color);
				da = a;
			}

			if ((b < db) && (b > da))
			{
				hline(x - c, y + b, c * 2, color);
				hline(x - c, y - b, c * 2, color);
				db = b;
			}

		}
		while (b > a);
	}
	return 1;
}

/**
Draws a circular arc with centre x, y and radius r, in an anticlockwise direction
starting from the angle a and ending when it reaches b. These values are specified
in radian format, with M_PI*2 equal to a full circle.
Zero is to the right of the centre point, and larger values rotate anticlockwise from there.
@param x x coordinate of the arc centre
@param y y coordinate of the arc centre
@param ang1 starting angle
@param ang2 ending angle
@param r radius in pixels
*/
int DrawBuffer::arc(int x, int y, double ang1, double ang2, int r)
{
	return arc(x,y,ang1,ang2,r,_ink);
}

/**
Draws a circular arc with centre x, y and radius r, in an anticlockwise direction
starting from the angle a and ending when it reaches b. These values are specified
in radian format, with M_PI*2 equal to a full circle.
Zero is to the right of the centre point, and larger values rotate anticlockwise from there.
@param x x coordinate of the arc centre
@param y y coordinate of the arc centre
@param ang1 starting angle
@param ang2 ending angle
@param r radius in pixels
@param color color
*/
int DrawBuffer::arc(int x, int y, double ang1, double ang2, int r, FGPixel color)
{
	int px, py;
	int ex, ey;
	int px1, px2, px3;
	int py1, py2, py3;
	int d1, d2, d3;
	int ax, ay;
	int q, qe;
	double tg_cur, tg_end;
	int done = 0;
	double rr1, rr2, rr3;
	int rr = (r * r);
	int carry_qv=0;

	rr1 = r;
	rr2 = x;
	rr3 = y;

	/* evaluate the start point and the end point */
	px = (int) (rr2 + rr1 * cos(ang1));
	py = (int) (rr3 - rr1 * sin(ang1));
	ex = (int) (rr2 + rr1 * cos(ang2));
	ey = (int) (rr3 - rr1 * sin(ang2));

	/* start quadrant */
	if (px >= x)
	{
		if (py <= y)
			q = 1;				/* quadrant 1 */
		else
			q = 4;				/* quadrant 4 */
	}
	else
	{
		if (py < y)
			q = 2;				/* quadrant 2 */
		else
			q = 3;				/* quadrant 3 */
	}

	/* end quadrant */
	if (ex >= x)
	{
		if (ey <= y)
			qe = 1;				/* quadrant 1 */
		else
			qe = 4;				/* quadrant 4 */
	}
	else
	{
		if (ey < y)
			qe = 2;				/* quadrant 2 */
		else
			qe = 3;				/* quadrant 3 */
	}

#define loc_tg(_y, _x)  (_x-x) ? (double)(_y-y)/(_x-x) : (double)(_y-y)

	tg_end = loc_tg(ey, ex);
    if (q==qe)  // su v rovnakom kvadrante
    { if (ang1 > ang2) // kreslenie cez 4 kv
      {
         carry_qv=1;
         qe+=4;
    } }

	while (!done)
	{
		putpixel(px, py);

		/* from here, we have only 3 possible direction of movement, eg.
		 * for the first quadrant:
		 *
		 *    OOOOOOOOO
		 *    OOOOOOOOO
		 *    OOOOOO21O
		 *    OOOOOO3*O
		 */

		/* evaluate the 3 possible points */
		switch (q)
		{

			case 1:
				px1 = px;
				py1 = py - 1;
				px2 = px - 1;
				py2 = py - 1;
				px3 = px - 1;
				py3 = py;

				/* 2nd quadrant check */
				if (px != x)
				{
					break;
				}
				else
				{
					/* we were in the end quadrant, changing is illegal. Exit. */
					if (qe == q)
						done = 1;
					q++;
                    if (carry_qv) {carry_qv=0;qe-=4;}
				}
				/* fall through */

			case 2:
				px1 = px - 1;
				py1 = py;
				px2 = px - 1;
				py2 = py + 1;
				px3 = px;
				py3 = py + 1;

				/* 3rd quadrant check */
				if (py != y)
				{
					break;
				}
				else
				{
					/* we were in the end quadrant, changing is illegal. Exit. */
					if (qe == q)
						done = 1;
					q++;
                    if (carry_qv) {carry_qv=0;qe-=4;}
				}
				/* fall through */

			case 3:
				px1 = px;
				py1 = py + 1;
				px2 = px + 1;
				py2 = py + 1;
				px3 = px + 1;
				py3 = py;

				/* 4th quadrant check */
				if (px != x)
				{
					break;
				}
				else
				{
					/* we were in the end quadrant, changing is illegal. Exit. */
					if (qe == q)
						done = 1;
					q++;
                    if (carry_qv) {carry_qv=0;qe-=4;}
				}
				/* fall through */

			case 4:
				px1 = px + 1;
				py1 = py;
				px2 = px + 1;
				py2 = py - 1;
				px3 = px;
				py3 = py - 1;

				/* 1st quadrant check */
				if (py == y)
				{
					/* we were in the end quadrant, changing is illegal. Exit. */
					if (qe == q)
						done = 1;

					q = 1;
                    if (carry_qv) {carry_qv=0;qe-=4;}
					px1 = px;
					py1 = py - 1;
					px2 = px - 1;
					py2 = py - 1;
					px3 = px - 1;
					py3 = py;
				}
				break;

			default:
				return 1;
		}

		/* now, we must decide which of the 3 points is the right point.
		 * We evaluate the distance from center and, then, choose the
		 * nearest point.
		 */
		ax = x - px1;
		ay = y - py1;
		rr1 = ax * ax + ay * ay;

		ax = x - px2;
		ay = y - py2;
		rr2 = ax * ax + ay * ay;

		ax = x - px3;
		ay = y - py3;
		rr3 = ax * ax + ay * ay;

		/* difference from the main radius */
		if (rr1 > rr)
			d1 = (int) (rr1 - rr);
		else
			d1 = (int) (rr - rr1);
		if (rr2 > rr)
			d2 = (int) (rr2 - rr);
		else
			d2 = (int) (rr - rr2);
		if (rr3 > rr)
			d3 = (int) (rr3 - rr);
		else
			d3 = (int) (rr - rr3);

		/* what is the minimum? */
		if (d1 <= d2)
		{
			px = px1;
			py = py1;
		}
		else if (d2 <= d3)
		{
			px = px2;
			py = py2;
		}
		else
		{
			px = px3;
			py = py3;
		}

		/* are we in the final quadrant? */
		if (qe == q)
		{
			tg_cur = loc_tg(py, px);

			/* is the arc finished? */
			switch (q)
			{

				case 1:
					/* end quadrant = 1? */
					if (tg_cur <= tg_end)
						done = 1;
					break;

				case 2:
					/* end quadrant = 2? */
					if (tg_cur <= tg_end)
						done = 1;
					break;

				case 3:
					/* end quadrant = 3? */
					if (tg_cur <= tg_end)
						done = 1;
					break;

				case 4:
					/* end quadrant = 4? */
					if (tg_cur <= tg_end)
						done = 1;
					break;
			}
		}
	}

	/* draw the last evaluated point */
	putpixel(px, py, color);
	return 1;
}

#define swap(a,b)           {a^=b; b^=a; a^=b;}
#define	MIN(a,b)	( (a) < (b) ? (a) : (b) )
#define MAX(a,b)	( (a) > (b) ? (a) : (b) )

void DrawBuffer::__line(int a1, int b1, int a2, int b2, FGPixel color)
{
	int xend, yend, dx, dy, c1, c2, step;
	register int p, x, y;

	dx = abs(a2 - a1);
	dy = abs(b2 - b1);

	if (dx > 32000)
		return;					// bulgar const

	if (dy > 32000)
		return;

	if (dx > dy)
	{							/* slope < 1 => step in x direction */
		x = MIN(a1, a2);
		xend = MAX(a1, a2);
		if (x == a1)
		{
			y = b1;
			step = ((b2 - y) < 0 ? -1 : 1);
		}
		else
		{
			y = b2;
			step = ((b1 - y) < 0 ? -1 : 1);
		}

		p = 2 * dy - dx;
		c1 = 2 * dy;
		c2 = 2 * (dy - dx);

		putpixel(x, y, color);
		while (x < xend)
		{
			x++;
			if (p < 0)
			{
				p += c1;
			}
			else
			{
				y += step;
				p += c2;
			}
			putpixel(x, y, color);
		}
	}
	else
	{							/* slope > 1 => step in y direction */
		y = MIN(b1, b2);
		yend = MAX(b1, b2);
		if (y == b1)
		{
			x = a1;
			step = ((a2 - x) < 0 ? -1 : 1);
		}
		else
		{
			x = a2;
			step = ((a1 - x) < 0 ? -1 : 1);
		}

		p = 2 * dx - dy;
		c1 = 2 * dx;
		c2 = 2 * (dx - dy);

		putpixel(x, y, color);

		while (y < yend)
		{
			y++;
			if (p < 0)
			{
				p += c1;
			}
			else
			{
				x += step;
				p += c2;
			}
			putpixel(x, y, color);
		}
	}
}

/**
	Draw a line from position [x:y] to the position [a:b] with
	current object foreground color.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::line(int x, int y, int a, int b)
{
	return line(x,y,a,b,_ink);
}

/**
	Draw a line from position [x:y] to the position [a:b] with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::line(int x, int y, int a, int b, FGPixel color)
{
	int cl=1;
	oldx = a;
	oldy = b;
	cl = ClipLine(x,y,a,b,0,0,w,h);
	if (cl)	__line(x,y,a,b,color); // null if not clipped
	return cl;
}

/**
	Draw a text with the current object font & colors from the position [x:y].

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::text(int x, int y, const char *txt)
{
	return text(x,y,txt,_ink,_paper);
}

/**
	Draw a text with the current object font, background color and foreground 'i' from the position [x:y].

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::text(int x, int y, const char *txt, FGPixel i)
{
	return text(x,y,txt,i,_paper);
}

static FGMutex TextMutex;

/**
	Draw a text with the current object font, background 'p' and foreground 'i' from the position [x:y].

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::text(int x, int y, const char *txt, FGPixel i, FGPixel p)
{
	FGPixel *dst, *dst2;
	int alloc=0;

	TextMutex.Lock();


	x+=xoff;
	y+=yoff;

	int	width =	textwidth((char *)txt);
	int _x=x, _y=y,ww=GetFontW(), bufw=width+3;
	int height = GetFontH();

	if (clip(x,y,width,height)==0 || width<2)
	{
		TextMutex.Unlock();
		return 0;
	}

	_x = x-_x;
	_y = y-_y;

	unsigned  bytes = (bufw)*height;

	if (bytes < FGTEXTBUFSZ)
		dst	= fgimagebuffer;
	else
	{
		alloc = 1;
		dst = (FGPixel *) malloc(bytes*sizeof(FGPixel));
	}

	dst2 = dst;

	if (_ppop == _GTRANSP)
	{
		p = transp_color;
	}
	set_mmx();
	while (*txt)
	{
		variable_record *fontvar = FGFontManager::IsVariableFont(_font);
		FGPixel *font = FGFontManager::GetFontImage(_font);

		if (fontvar)
		{
			ww = fontvar->w[*(unsigned char *)txt];
		}
		CharOutClip(dst,
			fontvar ? fontvar->off[*(unsigned char *)txt++] + font:
				*(unsigned char *)txt++ * (GetFontW()*GetFontH()) + font,
			bufw,
			(ww+3)&~3,
			GetFontH(), i, p);
		dst	+= ww;
	}
	reset_mmx();
	RamToRam(_x, _y, bufw, GetFontH(), x, y, w, h, width, height, dst2, image, BITBLIT_WITH_OPERATOR, _ppop);
	if (alloc==1)
		free(dst2);

	TextMutex.Unlock();


	return 1;
}

/**
	Draw a patterned line from position [a1:b1] to the position [a2:b2].

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::pline(int a1, int b1, int a2, int b2, Pattern *pat)
{
	int xend, yend, dx, dy, c1, c2, step;
	register int p, x, y;
	int cl;
	FGPixel iscolor;

	oldx = a2;
	oldy = a2;

	cl = ClipLine(a1,b1,a2,b2,0,0,w,h);

	if (cl==0) return 0;

	dx = abs(a2 - a1);
	dy = abs(b2 - b1);

	if (dx > 32000)
		return 0;			// bulgar const

	if (dy > 32000)
		return 0;

	if (pat->reset)
		pat->pos = 0;

	if (dx > dy)
	{							/* slope < 1 => step in x direction */
		x = MIN(a1, a2);
		xend = MAX(a1, a2);
		if (x == a1)
		{
			y = b1;
			step = ((b2 - y) < 0 ? -1 : 1);
		}
		else
		{
			y = b2;
			step = ((b1 - y) < 0 ? -1 : 1);
		}

		p = 2 * dy - dx;
		c1 = 2 * dy;
		c2 = 2 * (dy - dx);

		iscolor=pat->next();
		if (iscolor)
			putpixel(x, y, iscolor);
		while (x < xend)
		{
			x++;
			if (p < 0)
			{
				p += c1;
			}
			else
			{
				y += step;
				p += c2;
			}
			iscolor=pat->next();
			if (iscolor)
				putpixel(x, y, iscolor);
		}
	}
	else
	{							/* slope > 1 => step in y direction */
		y = MIN(b1, b2);
		yend = MAX(b1, b2);
		if (y == b1)
		{
			x = a1;
			step = ((a2 - x) < 0 ? -1 : 1);
		}
		else
		{
			x = a2;
			step = ((a1 - x) < 0 ? -1 : 1);
		}

		p = 2 * dx - dy;
		c1 = 2 * dx;
		c2 = 2 * (dx - dy);

		iscolor=pat->next();
		if (iscolor)
			putpixel(x, y, iscolor);

		while (y < yend)
		{
			y++;
			if (p < 0)
			{
				p += c1;
			}
			else
			{
				x += step;
				p += c2;
			}
			iscolor=pat->next();
			if (iscolor)
				putpixel(x, y, iscolor);
		}
	}
	return cl;
}

/**
	Draw a rectangle from position [x:y] with size [w:h] with
	current object foreground color.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::rect(int x, int y, int w, int h)
{
	return rect(x,y,w,h,_ink);
}

/**
	Draw a rectangle from position [x:y] with size [w:h] with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::rect(int x, int y, int w, int h, FGPixel color)
{
	hline(x, y,	w,color);
	hline(x, y + h - 1,	w,color);
	vline(x, y,	h,color);
	vline(x	+ w	- 1, y,	h,color);
	return 1;
}

/**
	Fill a rectangle from position [x:y] with size [w:h] with
	current object foreground color.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::box(int x, int y, int ww, int hh)
{
	return box(x,y,ww,hh,_ink);
}

/**
	Fill a rectangle from position [x:y] with size [w:h] with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::box(int x, int y, int ww, int hh, FGPixel i)
{
	x += xoff;
	y += yoff;

	if (clip(x, y, ww, hh) == 0) return 0;

	L1Box(CalcAddr(x,y), ww, hh, i, w, _ppop);
	return 1;
}

/**
	Fill convex polygon with 'n' vertices. The each vertice is
	a pair of two integers X and Y. Parameter 'pt' is an array
	of these pairs. FUnction return the Rectangle of the polygon
	boundary. The used color is current object's foreground color.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
FGRect DrawBuffer::fpolygon(int n, int pt[][2])
{
	return fpolygon(n,pt,_ink);
}

/**
	Fill convex polygon with 'n' vertices. The each vertice is
	a pair of two integers X and Y. Parameter 'pt' is an array
	of these pairs. FUnction return the Rectangle of the polygon
	boundary.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
FGRect DrawBuffer::fpolygon(int n, int pt[][2], FGPixel color)
{
	FGRect rect;
	edge *edges=edge_array, *ep;
	scan *scans=scan_array, *sp, *points, *segments;
	int	xmin, xmax,	ymin, ymax;
	int	ypos, nedges, mem = 0;

	if ((n > 1)	&&
		(pt[0][0] == pt[n -	1][0]) &&
		(pt[0][1] == pt[n -	1][1]))
	{
		n--;
	}
	if (n <	1)
	{
		return rect;
	}
	if (n>128)
	{
		edges =	(edge *) malloc(sizeof(edge) * (n +	2));
		scans =	(scan *) malloc(sizeof(scan) * (n +	8));
		mem = 1;
	}
	if (edges && scans)
	{
		int	prevx =	xmin = xmax	= pt[0][0];
		int	prevy =	ymin = ymax	= pt[0][1];

		nedges = 0;
		ep = edges;
		while (--n >= 0)
		{
			if (pt[n][1] >=	prevy)
			{
				ep->e.x	= prevx;
				ep->e.y	= prevy;
				ep->e.xlast	= prevx	= pt[n][0];
				ep->e.ylast	= prevy	= pt[n][1];
			}
			else
			{
				ep->e.xlast	= prevx;
				ep->e.ylast	= prevy;
				ep->e.x	= prevx	= pt[n][0];
				ep->e.y	= prevy	= pt[n][1];
			}
			if ((ep->e.y > (hwrk+yoff)) || (ep->e.ylast <	(0)))
				continue;
			{
				if (ep->e.y	< 0)
				{
					ep->e.x	+= (((int) (((long)	(((int)	((ep->e.xlast -	ep->e.x)) << 1)) * (long) ((0	- ep->e.y))) / (long) ((ep->e.ylast	- ep->e.y))) + (((int) ((ep->e.xlast - ep->e.x)) ^ (int) ((0 - ep->e.y)) ^ (int) ((ep->e.ylast - ep->e.y))) >> ((sizeof(int) * 8)	- 1)) +	1) >> 1);

					ep->e.y	= 0;;
				}
			};
			if (ymin > ep->e.y)
				ymin = ep->e.y;
			if (ymax < ep->e.ylast)
				ymax = ep->e.ylast;
			if (xmin > ep->e.x)
				xmin = ep->e.x;
			if (xmax < ep->e.xlast)
				xmax = ep->e.xlast;
			{
				(&ep->e)->dy = (&ep->e)->ylast - (&ep->e)->y;
				(&ep->e)->dx = (&ep->e)->xlast - (&ep->e)->x;
				if ((&ep->e)->dx < 0)
				{
					(&ep->e)->xstep	= (-1);
					(&ep->e)->dx = (-(&ep->e)->dx);
				}
				else
				{
					(&ep->e)->xstep	= 1;
				}
				if ((&ep->e)->dx > (&ep->e)->dy)
				{
					(&ep->e)->xmajor = 1;
					(&ep->e)->error	= (&ep->e)->dx >> 1;
				}
				else
				{
					(&ep->e)->xmajor = 0;
					(&ep->e)->error	= ((&ep->e)->dy	- ((1 -	(&ep->e)->xstep) >>	1))	>> 1;
				}
			};
			ep->status = inactive;
			nedges++;
			ep++;
		}
		rect.x = xmin;
		rect.y = ymin;
		rect.w = xmax-xmin;
		rect.h = ymax-ymin;
		if ((nedges	> 0) &&	(xmin <= wwrk+xoff)	&& (xmax >=	0))
		{
			if (xmin < 0)
				xmin = 0;
			if (ymin < 0)
				ymin = 0;
			if (xmax > (wwrk+xoff))
				xmax = (wwrk+xoff);
			if (ymax > (hwrk+yoff))
				ymax = (hwrk+yoff);

			for	(ypos =	ymin; ypos <= ymax;	ypos++)
			{
				sp = scans;
				points = 0;
				segments = 0;
				for	(n = nedges, ep	= edges; --n >=	0; ep++)
				{
					switch (ep->status)
					{
						case inactive:
							if (ep->e.y	!= ypos)
								break;
							if (ep->e.dy ==	0)
							{
								ep->status = passed;
								xmin = ep->e.x;
								xmax = ep->e.xlast;
								{
									if ((int) (xmin) > (int) (xmax))
									{
										int	_swap_tmpval_ =	(xmin);

										(xmin) = (xmax);
										(xmax) = _swap_tmpval_;
									}
								};
								{
									scan *prev = 0;
									scan *work = segments;
									int	overlap	= 0;

									while (work	!= 0)
									{
										if ((work->x1 <= xmax) && (xmin	<= work->x2))
										{
											overlap	= 1;
											if (xmin < work->x1)
												work->x1 = xmin;
											if (xmax > work->x2)
											{
												prev = work;
												while ((work = work->next) != 0)
												{
													if (work->x1 > xmax)
														break;
													if (work->x2 > xmax)
														xmax = work->x2;
												}
												prev->x2 = xmax;
												prev->next = work;
											}
											break;
										}
										if (work->x1 > xmax)
											break;
										prev = work;
										work = work->next;
									}
									if (!overlap)
									{
										sp->x1 = xmin;
										sp->x2 = xmax;
										sp->next = work;
										if (prev)
											prev->next = sp;
										else
											segments = sp;
									}
								};
								sp++;
								break;
							}
							ep->status = active;
						case active:
							xmin = xmax	= ep->e.x;
							if (ep->e.ylast	== ypos)
							{
								ep->status = passed;
								xmax = ep->e.xlast;
								{
									if ((int) (xmin) > (int) (xmax))
									{
										int	_swap_tmpval_ =	(xmin);

										(xmin) = (xmax);
										(xmax) = _swap_tmpval_;
									}
								};
								{
									scan *prev = 0;
									scan *work = points;

									while (work	!= 0)
									{
										if (work->x1 > xmin)
											break;
										prev = work;
										work = work->next;
									}
									sp->x1 = xmin;
									sp->x2 = xmax;
									sp->next = work;
									if (prev)
										prev->next = sp;
									else
										points = sp;
								};
								sp++;
							}
							else if	(ep->e.xmajor)
							{
								for	(;;)
								{
									(&ep->e)->x	+= (&ep->e)->xstep;
									if (((&ep->e)->error -=	(&ep->e)->dy) <	0)
									{
										(&ep->e)->error	+= (&ep->e)->dx;
										break;
									}
								};
								xmax = ep->e.x - ep->e.xstep;
								{
									if ((int) (xmin) > (int) (xmax))
									{
										int	_swap_tmpval_ =	(xmin);

										(xmin) = (xmax);
										(xmax) = _swap_tmpval_;
									}
								};
							}
							else
							{
								{
									if (((&ep->e)->error -=	(&ep->e)->dx) <	0)
									{
										(&ep->e)->x	+= (&ep->e)->xstep;
										(&ep->e)->error	+= (&ep->e)->dy;
									}
								};
							}
							{
								scan *prev = 0;
								scan *work = points;

								while (work	!= 0)
								{
									if (work->x1 > xmin)
										break;
									prev = work;
									work = work->next;
								}
								sp->x1 = xmin;
								sp->x2 = xmax;
								sp->next = work;
								if (prev)
									prev->next = sp;
								else
									points = sp;
							};
							sp++;
							break;
						default:
							break;
					}
				}
				while (points != 0)
				{
					scan *nextpt = points->next;

					if (!nextpt)
						break;
					xmin = points->x1;
					xmax = nextpt->x2;
					points = nextpt->next;
					{
						scan *prev = 0;
						scan *work = segments;
						int	overlap	= 0;

						while (work	!= 0)
						{
							if ((work->x1 <= xmax) && (xmin	<= work->x2))
							{
								overlap	= 1;
								if (xmin < work->x1)
									work->x1 = xmin;
								if (xmax > work->x2)
								{
									prev = work;
									while ((work = work->next) != 0)
									{
										if (work->x1 > xmax)
											break;
										if (work->x2 > xmax)
											xmax = work->x2;
									}
									prev->x2 = xmax;
									prev->next = work;
								}
								break;
							}
							if (work->x1 > xmax)
								break;
							prev = work;
							work = work->next;
						}
						if (!overlap)
						{
							nextpt->x1 = xmin;
							nextpt->x2 = xmax;
							nextpt->next = work;
							if (prev)
								prev->next = nextpt;
							else
								segments = nextpt;
						}
					};
				}
				while (segments	!= 0)
				{
					xmin = segments->x1;
					xmax = segments->x2;
					segments = segments->next;
					{
						if (xmin > wwrk+xoff)
						{
							continue;
						}
						if (xmax < 0)
						{
							continue;
						}
						if (xmin < 0)
						{
							xmin = 0;;
						}
						if (xmax > wwrk+xoff)
						{
							xmax = wwrk+xoff;
						}
					}
					hline(xmin, ypos, xmax - xmin +	1, color);
				}
			}
		}
	}
	if (mem)
	{
		free(edges);
		free(scans);
	}
	return rect;
}

/**
	Draw convex polygon with 'n' vertices. The each vertice is
	a pair of two integers X and Y. Parameter 'pt' is an array
	of these pairs. The used color is current object's foreground color.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
void DrawBuffer::polygon(int n, int pt[][2])
{
	polygon(n,pt,_ink);
}

/**
	Draw convex polygon with 'n' vertices. The each vertice is
	a pair of two integers X and Y. Parameter 'pt' is an array
	of these pairs.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
void DrawBuffer::polygon(int n, int pt[][2], FGPixel color)
{
	int doClose=0;
	if (n <= 1)	return;
	if (n == 2) doClose	= 1;
	moveto(pt[0][0], pt[0][1]);
	for	(int i = 1;	i <	n; i++)
	{
		lineto(pt[i][0], pt[i][1], color);
	}
	if (doClose || (pt[0][0]!=pt[n-1][0] && pt[0][1]!=pt[n-1][1]))
		lineto(pt[0][0], pt[0][1], color);
}

/**
	Draws a bezier spline using the four control points specified in the points array.
	The used color is current object's foreground color.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
void DrawBuffer::spline(int points[8])
{
	spline(points, _ink);
}

/**
	Draws a bezier spline using the four control points specified in the points array.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
void DrawBuffer::spline(int points[8], FGPixel color)
{
	#define NPTS   64

	int xpts[NPTS], ypts[NPTS];
	int i;

	calc_spline(points, NPTS, xpts, ypts);
	moveto(xpts[0], ypts[0]);
	for (i=1; i<NPTS; i++)
	{
		lineto(xpts[i], ypts[i], color);
	}
}

/**
	Draw a line from the latest PEN position to the position [x:y]
	The used color is current object's foreground color.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::lineto(int x, int y)
{
	return lineto(x,y,_ink);
}

/**
	Draw a line from the latest PEN position to the position [x:y]

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::lineto(int x, int y, FGPixel color)
{
	int rc = line(oldx, oldy, x, y, color);
	if (_ppop == _GXOR)
		putpixel(x, y, color);
	return rc;
}

/**
	Draw a text via standard "C" behaviour with the current object font & colors from the position [x:y].

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int DrawBuffer::printf(int xx, int yy, const char *format,...)
{
	char s[256];
	va_list arglist;

	va_start(arglist, format);

#ifndef _MSC_VER
	vsnprintf(s, 255, format, arglist);
#else
	vsprintf(s, format, arglist);
#endif

	va_end(arglist);
	int len = strlen(s);
	if (s[len-1] == '\n')
	{
		s[--len] = 0;
	}
	text(xx,yy,s);
	return len;
}

void DrawBuffer::Resize(int dx, int dy)
{
	// test to new new size
	if (w + dx<1 || h + dy<1)
		return;
	if (dx==0 && dy==0 && image)
		return;
	lock();
	if (image) free(image);
	w += dx;
	h += dy;
	if (w*h)
		assert((image = (FGPixel *) malloc(areasize(w, h)))!=0);
	wwrk += dx;
	hwrk += dy;
	unlock();
}

/**
	Change the name of the object.
*/
void DrawBuffer::SetName(char *s)
{
	lock();
	if (name != empty_string) delete [] name;
	name = new char[strlen(s)+1];
	strcpy(name,s);
	unlock();
}

/**
 * Resize the image with ratio 'rx' and 'ry'.
 */
void DrawBuffer::stretch(float rx, float ry)
{
	stretch((int)(w*rx), (int)(h*ry));
}

/**
 * Resize the image to the new size.
 */
void DrawBuffer::stretch(int _w, int _h)
{
#ifndef INDEX_COLORS
		int xpos = 0, ypos = 0;
		unsigned pix;
		float r, g, b;
		int r1, g1, b1;
		int r2, g2, b2;
		int r3, g3, b3;
		int r4, g4, b4;
		int r0, g0, b0;
		float ratio_x;
		float ratio_y;
		float x1, y1;
		float xerr, yerr;
		float xfloat, yfloat;

		if (_w == w && _h == h)
			return;
		if (_w <1 || _h < 1)
			return;

		FGPixel *dest = (FGPixel *)malloc(areasize(_w, _h)), *dest_pom=dest;

		ratio_x = (float)w / (float)_w;
		ratio_y = (float)h / (float)_h;

		for (ypos = 0; ypos < _h; ypos++)
				for (xpos = 0; xpos <  _w; xpos++) {
						xfloat = (xpos) * ratio_x;
						yfloat = (ypos) * ratio_y;

						x1 = (int)xfloat;
						y1 = (int)yfloat;

						xerr = 1.0 - (xfloat - (float)x1);
						yerr = 1.0 - (yfloat - (float)y1);
#define _getpixel(x,y,r,g,b) { pix = getpixel((int)x, (int)y); r=kRedComponent(pix); g=kGreenComponent(pix); b=kBlueComponent(pix); }
						_getpixel(x1, y1, r1, g1, b1);
						_getpixel(x1 + 1, y1, r2, g2, b2);
						_getpixel(x1 + 1, y1 + 1, r3, g3, b3);
						_getpixel(x1, y1 + 1, r4, g4, b4);
						r = (float)r1 * xerr +
							(float)r2 * (1.0 - xerr) +
							(float)r3 * (1.0 - xerr) +
							(float)r4 * xerr;

						r += (float)r1 * yerr +
							 (float)r2 * yerr +
							 (float)r3 * (1.0 - yerr) +
							 (float)r4 * (1.0 - yerr);

						r0 = int(r * 0.25);

						g = (float)g1 * xerr +
							(float)g2 * (1.0 - xerr) +
							(float)g3 * (1.0 - xerr) +
							(float)g4 * xerr;

						g += (float)g1 * yerr +
							 (float)g2 * yerr +
							 (float)g3 * (1.0 - yerr) +
							 (float)g4 * (1.0 - yerr);

						g0 = int(g * 0.25);

						b = (float)b1 * xerr +
							(float)b2 * (1.0 - xerr) +
							(float)b3 * (1.0 - xerr) +
							(float)b4 * xerr;

						b += (float)b1 * yerr +
							 (float)b2 * yerr +
							 (float)b3 * (1.0 - yerr) +
							 (float)b4 * (1.0 - yerr);

						b0 = int(b * 0.25);

#ifndef BPP32
							*dest_pom++ = (((r0<<8)&0xf800) | ((g0<<3)&0x7e0) | ((b0>>3)&0x1f));
#else
							*dest_pom++ = ((r0<<16) | (g0<<8) | b0);
#endif
		}
	Resize(_w-w, _h-h);
	free(image);
	image = dest;
#endif
}

/**
	Fill a triangle with these three vertices.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
	@param x1 x position of first vertice
	@param y1 y position of first vertice
	@param x2 x position of second vertice
	@param y2 y position of second vertice
	@param x3 x position of third vertice
	@param y3 y position of third vertice
*/
FGRect DrawBuffer::ftriangle(int x1, int y1, int x2, int y2, int x3, int y3)
{
	int point[3][2];
	point[0][0] = x1;
	point[0][1] = y1;
	point[1][0] = x2;
	point[1][1] = y2;
	point[2][0] = x3;
	point[2][1] = y3;
	return fpolygon(3, point);
}

/**
	Fill a triangle with these three vertices.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function returns false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
	@param x1 x position of first vertice
	@param y1 y position of first vertice
	@param x2 x position of second vertice
	@param y2 y position of second vertice
	@param x3 x position of third vertice
	@param y3 y position of third vertice
	@param Ink foreground color
*/
FGRect DrawBuffer::ftriangle(int x1, int y1, int x2, int y2, int x3, int y3, FGPixel Ink)
{
	int point[3][2];
	point[0][0] = x1;
	point[0][1] = y1;
	point[1][0] = x2;
	point[1][1] = y2;
	point[2][0] = x3;
	point[2][1] = y3;
	return fpolygon(3, point, Ink);
}

/*
 *  geometry is 0 .. w       in X
 *				0 .. h       in Y
 *	clipping is 0 .. wwrk in X
 *				0 .. hwrk in Y
 *	offset is xoff & yoff 0 .. w-1
 *
 *	NOTE: you must applicate offset first before clipping!
 */


