/*
    OpenGUI - Drawing & Windowing library

    Copyright (C) 1996,2003  Marian Krivos

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	nezmar@atlas.sk

	fastgl.h   - declarations of many objects for FastGL
*/

#ifndef	__WINDOW_H
#define	__WINDOW_H

/* HJH modification: protect compiler options for structure alignment and enum
 * size if the compiler is Borland C++ */
#include "base.h"
#pragma interface
#ifdef __BORLANDC__
#pragma option push -b
#pragma option push -a4
#endif

#define	BNORMAL		7
#define	BPUSHED		3
#define	BDISABLED	5
#define	BUNVISIBLE	6

#if	1
#define	FGFONT_TITLE	FONTSYS
#define	FGFONT_MENU		FONTSYSLIGHT
#define	FGFONT_BUTTON	FONTSYSLIGHT
#else
#define	FGFONT_TITLE	FONT0816
#define	FGFONT_MENU		FONT0816
#define	FGFONT_BUTTON	FONT0816
#endif

/**
	@addtogroup Defines
	@{
*/
//! predefined Window foreground color
#define	IM		 	CScheme->window_fore // CBLACK standard menu fore & background colors
//! predefined Window background color
#define	PM			CScheme->window_back

//! The Window has frame.
#define	WFRAMED		0x0001
//! The Window has caption.
#define	WTITLED		0x0002
//! The Window will be shown as modal.
#define	WMODAL		0x0004
//! The Window sends EVENTNOTIFY to its handler when it is destroyed.
#define	WNOTIFY		0x0008
//! internal
#define	WHIDEN		0x0010
//! The Window is unable to move/change its position.
#define	WUNMOVED	0x0020
//! The Window has a line menu.
#define	WMENU		0x0040
//! The Window is without pictograms (to minimize and to hide it).
#define	WNOPICTO	0x0080
//! internal
#define	WBITMAP		0x0100
//! The Window emit CLICKLEFTEVENT and CLICKRIGHTEVENT when you clicks with mouse on it.
#define	WCLICKABLE	0x0200
//! internal
#define	WDIRTY		0x0400
//! internal
#define	WTEMPHIDEN	0x0800
//! internal
#define	WLOCKED		0x1000
//! internal
#define	WTRIGGER	0x2000
//! The Window don't get focus on create.
#define	WLASTFOCUS	0x4000
//! The Window is able to change its size.
#define	WSIZEABLE	0x8000
//! The Window sets its size and position from WindowDatabase if any on create time.
#define	WUSELAST	0x00010000
//! Show the minimize pictogram in the Window title.
#define	WMINIMIZE	0x00020000
//! The Window has a statusbar in the bottom.
#define	WSTATUSBAR	0x00040000
//! internal
#define	WEXIST		0x00080000
//! internal
#define	WGLCONTEXT	0x00100000
/**
The Window uses notebooks. Use this flag to create multiple page dialog
or tabbed notebook. It displays multiple overlapping pages that are TabPage
objects. The user selects selects a page by clicking the page's tab that appears
at the top of control.
@see Window::AddTabPage()
*/
#define	WNOTEBOOK	0x00200000
//! Show the Window in the center of screen.
#define	WCENTRED	0x00800000
//! Close the Window when ESCAPE key is pressed.
#define	WESCAPE		0x01000000
//! Alligne the Window to 32 bit sizes
#define	WALIGNED	0x02000000
//! Enable fast move of the Window, i.e. don't redraw opaque.
#define	WFASTMOVE	0x04000000
//! internal
#define	WNOICON		0x10000000
//! internal
#define	WMENUWINDOWTYPE	   0x80000000
//! internal
#define	WICONWINDOWTYPE	   0x40000000
//! standard predefined values.
#define	WSTANDARD	(WCLICKABLE|WTITLED|WFRAMED)

//! Enable to remember Window's positions and sizes.
#define	APP_WINDOWDATABASE		1
//! Enable Configuration file in form EXEFILENAME.rc
#define	APP_CFG					2
//! Enable The ALT+X hotkey to terminates your application.
#define	APP_ENABLEALTX			4
//! Enable the rectangle selections by mouse.
#define	APP_MAGNIFIER			8
//! Enable and initialize the root Window, i.e. background.
#define	APP_ROOTWINDOW			16
//! Enable visual efects.
#define	APP_ENABLEEFECTS		0x20
//! Enable all the above.
#define APP_ALL					0x1F
//! Use local application directory for Config files instead of HOMEDIR.
#define	APP_LOCALDIR			0x80000000

/**	@} */

/**
	The current Operating System code.
	@ingroup Enums
*/
enum ENUM_OS { OSTYPE_LINUX=1, OSTYPE_MSDOS, OSTYPE_WIN32, OSTYPE_QNX };

/**
	Return codes for App::ShowModal()
	@ingroup Enums
*/
enum MODAL_RETURN
{	mrNone   = 	 0,
	mrOk	 = 	-1,
	mrCancel =	-2,
	mrYes	 =	-1,
	mrNo	 =	-2,
	mrAll	 =	-3,
	mrIgnore =	-4,
	mrClose	 =	-5,
	mrQuit	 =	-6
};

class BaseGui;
class FileDialog;
class ColorDialog;
class Control;
class Window;
class MenuWindow;
class Bitmap;
class EditBox;
class ButtonGroup;
class FGListBox;
class PushButton;
class PointButton;
class CheckButton;
class BaseMenu;
class PushButton;
class Slider;
class FGTabPage;

/**
	Show the error dialog with defined string.
	@ingroup Misc
	@param s a string that will be shown
	@param flag 0 for warning or for fatal error (will exit after presskey). If you want MODAL dialog, you must add to the flag the value 2.
*/
extern "C" void	IError(char	*s,	int	flag);

/**
Provides simply and easy theme like object.
The library creates & initializes an instance of this structure on the startup
and uses it for a default settings for Controls and Windows that are created then.
*/
struct ColorScheme
{
	/** a default Window background color */
	int		window_back;
	/** a default Window foreground color */
	int		window_fore;
	/** an active Window title color */
	int		active_title;
	/** an inactive Window title color */
	int		inactive_title;
	/** a default Window shaded frame color 1 */
	int		wnd_bord1;
	/** a default Window shaded frame color 2 */
	int		wnd_bord2;
	/** a default Window shaded frame color 3 */
	int		wnd_bord3;
	/** a default Window statusbar color */
	int		statusbar;
	int		menu_back;
	int		menu_fore;
	int		menu_back_active;
	int		menu_fore_active;
	int		button_fore;
	int		button_back;
	int		button_fore_pushed;
	int		button_back_pushed;
	int		button_bord1;
	int		button_bord2;
	int		button_bord3;
	int		edit_back;
	int		edit_fore;
	int		edit_bord1;
	int		edit_bord2;
	int		slider;
	int		menuwindow_back;
	int		menuwindow_fore;
	int		pdmenu_back_active;
	int		pdmenu_fore_active;
	int		pdmenu_gray;
	int		edit_disable;
	int		notebook_active;
	int		slider_disable;
};

// subbitmap

extern ColorScheme * CScheme;

/**
    The TIMER procedure callback type.
	@ingroup Types
*/
typedef unsigned int (*TimerCall)(unsigned int);

#define TIMER_RESOLUTION	10	/* Experimentally determined */

/**
	Some timing class.
*/
class FGTimer
{
	protected:
		friend void  			HandleAlarm(int sig);
		int						state;
		static void				StartTicks(void);
	public:
		FGTimer(unsigned int val, TimerCall cb)
		{
			state = SetTimer(val, cb);
		}
		FGTimer(TimerCall cb)
		{
			state = SetTimer(0, cb);
		}
		~FGTimer()
		{
			TimerQuit();
		}
		static int			SetTimer(unsigned int val, TimerCall cb);
		static unsigned int GetTicks(void);
		static void			Delay(unsigned int);
		static int			TimerInit(void);
		static void			TimerQuit(void);
};

/**
	Open a configuration file with name (without the extension ".rc").
	If any not exists, system creates a new empty into the RAM.
	If you use switch APP_CFG, system automatically opens in
	the App constructor. Default config file is appname.rc.
	The pointer at this config saves to global variable
	cApp->cfg and its destructed (and saved) in App destructor.
	The maximal number of items into the config file is up to 512.
	The number of at once opens config files is unlimited. The maximal
	size of name of variable is up to 32 chars. The maximal size
	of saved text string or byte array is up to 127 chars + '\\0'.
*/
class Config
{
  class Node
  {
	enum itemtype
	{
	_DOUBLE='d',
	_INT='i',
	_STRING='s',
	_COMMENT='c',
	_HEXA='h'
	};

	friend class Config;
	int type;
	char name[36];
	int	i;
	char  s[132];
	double d;
	int len;
	void ModifyNode(char * comment)	// Modify comment
	{
		type = _COMMENT;
		*name =	0;
		strncpy(s, comment, 128);
	}
	void ModifyNode(char * meno, char *	val) // Modify string
	{
		type = _STRING;
		strncpy(name, meno, 32);
		strncpy(s, val, 128);
	}
	void ModifyNode(char * meno, int val) // Modify integer
	{
		type = _INT;
		strncpy(name, meno, 32);
		i	 = val;
	}
	void ModifyNode(char * meno, double	val) // Modify double
	{
		type = _DOUBLE;
		strncpy(name, meno, 32);
		d	 = val;
	}
	void ModifyNode(char * meno, char *	val, int kolko) // Modify string
	{
		type = _HEXA;
		strncpy(name, meno, 32);
		len = kolko>128? 128 : kolko;
		memcpy(s, val, len);
	}
  };

		enum { MAX_CONF_ITEMS=512 };
		Node 	pole[MAX_CONF_ITEMS];
		char 	name[128];
		FILE 	*f;
		int		entries;
		int		change;
		int		writable;

		int	 find(char *	s);
		void CreateEmpty(void);
		void ReadConfig(void);

	protected:
	public:
		void 	WriteComment(char * s);
		void	Sync(void);
		//! deprecated
		int		ReadInt(char	* s);
		int		ReadInt(char	* s, int &p);
		int		ReadString(char * s,	char *p, int max);
		//! deprecated
		char	const * ReadString(char * s);
		int		ReadString(char * s,	char **p);
		//! deprecated
		double	ReadDouble(char * s);
		int		ReadDouble(char * s, double &p);
		int		ReadHex(char * s, char *, int);

		void	WriteInt(char * s, int);
		void	WriteString(char	* s, char *);
		void	WriteDouble(char	* s, double);
		void	WriteHex(char * s, char *, int);

		Config(char	*,int write=1);
		~Config();
};

/**
	System event object. The events are presents in many types. By example
	event is key pressing, mouse move, click on the mouse button, window move etc.
*/
class GuiEvent
{
		friend class BaseGui;
		friend class App;
		int		type;
		int		key;
		int		buttons, x,	y, w, h;
		static	char *eventName[];
	protected:
	public:
    	/**
		This is a public member of GuiEvent class and it is initialised for
        window procedures only. It contains the pointer at the target window.
		It is useful when one procedure is assigned to more windows.
		Using of this member is in general better, than global pointers.
		*/
		Window	*wnd;
		/**
		This is a public member of GuiEvent class and it is assigned for ACCELEVENT only.
		It contains the type of controls. Normally it is unnecessary. The types are
		enum Type {WINDOW, CHECKBUTTON, e.g.
		@deprecated - will be removed in OpenGUI 4.1.0
		*/
		int		guiType;
		/**
		This is a public member of GuiEvent class and it is initialised for
		window procedures only and ACCELEVENT only. It contains the pointer
		at the Controls that emit this event.
		*/
		Control *accel;
		/**
		Create the event of appropriate type. In the most cases, you will parse
		system events only. Creating of the events by the hand and sending them
		to other windows is less frequently. The response to the system events
		is adjusted into separate window procedures, each per window. This procedure
		is fact a one big switch statement with many cases.
		*/
		GuiEvent(int T=NOEVENT,	int	K=0, int X=0, int Y=0, int W=0, int H=0)
			: type(T), key(K), x(X), y(Y), w(W), h(H), wnd(0), guiType(0), accel(0)
		{
		}
		//! Returns the value of key member of class GuiEvent. As for event KEYEVENT, it return a key code.
		int		GetKey(void)  const { return key;	}
		//! Return a text string, which identified the name of events, or string "NONSENSE".
		char	*GetName(void)  const { if (type<0 || type>=LASTEVENT) return eventName[LASTEVENT]; return eventName[type]; }
		//! Returns type of the event.
		int		GetType(void)  const { return	type; }
		/**
		Returns an OR-ed value of the all mouse's buttons that are pushed.
		There are these symbolic values for buttons:
			-	FG_BUTTON_LEFT
			-	FG_BUTTON_RIGHT
			-	FG_BUTTON_MIDDLE
		*/
		int		GetButtons(void)  const {	return buttons;	}
		//! Returns x co-ordination for events that initialise this member (MOVEEVENT, CLICKLEFTEVENT).
		int		GetX(void)  const { return	x; }
		//! Returns y co-ordination for events that initialise this member (MOVEEVENT, CLICKLEFTEVENT).
		int		GetY(void)  const { return	y; }
		//! Returns a width.
		int		GetW(void)  const { return	w; }
		//! Returns a width.
		int		GetH(void)  const { return	h; }
};

/**
	@addtogroup Types
	@{
*/
/**
The type of an App::SetTimerProc procedure callback.
*/
typedef	void (*TimerProc)(int);
/**
The type of a Window handler. This procedure is called when the Window EVENT (of type GuiEvent) is occured.
*/
typedef	void (*GuiHwnd)(GuiEvent *);
/**
The type of an Application handler. This procedure is called when system EVENT (of type GuiEvent) is occured.
*/
typedef	int	(*MainHwnd)(GuiEvent *);
/**
The pointer to a Control type.
*/
typedef	Control * CallBack;
/**
The prototype for a LISTBOX container callback procedure.
*/
typedef void (*FGListBoxCallBack)(int flag, Window *wnd, int x, int y, void *data, int index);
/**
The prototype for a FileDialog container callback procedure.
*/
typedef void (*FileDialogCallBack)(char * filename, FileDialog *);
/**
The prototype for a ColorDialog container callback procedure.
*/
typedef void (*ColorDialogCallBack)(FGPixel color);
/**
The prototype for a ColorDialog container callback procedure.
*/
typedef void (*FontDialogCallBack)(int f);

/** @} Types */

/**
Internal.
*/
typedef	void (*ControlCall)(Control *);


//!	base objects for all graphics objects
/**
	This class is base parent of all graphics objects into OpenGUI library
	and encapsulates all basic methods. These methods are used in Window,
	EditBox, PointButton and other GUI classes.
	It implements a linked list of objects also. This is used to release memory
	when Window is destroyed.
*/
class BaseGui : public DrawBuffer
{
	friend class Window;
	friend class Control;
	protected:
		enum bs	{ EBOX_LOW=1, EBOX_UPR, BDOWN, BUP, WHIDE, WVISIBLE, WDEACTIVE, WACTIVE };
		BaseGui	*last, * next;
		int		 x,y;
		FGMutex	ListMutex;
		FGPixel	ink, paper;

		BaseGui(const BaseGui &)
		{
			::printf("No COPY Constructor defined!\n");
			exit(-1);
		}
		BaseGui	* 	OverSprite(BaseGui * od, int	x, int y, int w, int h);
		BaseGui	* 	FindId(int	idw);
		BaseGui	* 	BaseGuiFind(GuiEvent *e);
		int		  	IsVisible(void)  const	{ return !(status&(WHIDEN|WTEMPHIDEN));	}
		void	  	Resize(int dx, int dy);
		void		SetLast(BaseGui *l) { last	= l; }
		void		SetNext(BaseGui *l) { next	= l; }
		BaseGui	* 	GetNext(void)  const {	return next; }
		BaseGui	* 	GetLast(void)  const {	return last; }
		virtual		void MoveIn(void) {}
		virtual		void MoveOut(void) {}
		BaseGui(int	xs,	int	ys,	int	ws,	int	hs,	char *nm, Type typ, FGPixel i, FGPixel p, long flag);
		virtual		~BaseGui();

	public:
		/**
		Set the object's color as default for drawing.
		*/
		void        SetColors(void)
		{
			set_colors(ink, paper);
		}
		/**
		Returns the 'x' coordinates of the object.
		*/
		int		  	GetX(void)  const { return	x; }
		/**
		Returns the 'y' coordinates of the object.
		*/
		int		  	GetY(void)  const { return	y; }
		/**
		Callback on focus get.
		*/
		virtual void OnFocus(void) { }
		/**
		Callback on focus lost.
		*/
		virtual void OnLostFocus(void) { }
};

#ifndef __GNUC__
#pragma	pack(1)
#endif

//!	Used for an image file manipulating (load/save of various formats is implemented).
class Bitmap : public DrawBuffer
{
	struct RGB3
	{
		char bBlue;				 /* Blue component of the color definition */
		char bGreen;			 /* Green component of the color definition*/
		char bRed;				 /* Red component of the color definition  */
	};
	struct RGB4
	{
		char bBlue;				 /* Blue component of the color definition */
		char bGreen;			 /* Green component of the color definition*/
		char bRed;				 /* Red component of the color definition  */
		char fcOptions;			 /* Reserved, must be zero                 */
	};
	struct bitmap_file_header
	{
		unsigned short	usType;
		unsigned short	cbSizeLow, cbSizeHigh;
		short			reserve1, reserve2;
	};
	struct bitmap_info
	{
		unsigned short	imageOffset;
		short			reserve;
	};
	struct bitmap_info_header1
	{
	  unsigned long	 size;
	  unsigned short cx;			   /* Bit-map width in pels                  */
	  unsigned short cy;			   /* Bit-map height in pels                 */
	  unsigned short cPlanes;		   /* Number of bit planes                   */
	  unsigned short cBitCount;		   /* Number of bits per pel within a plane  */
	  unsigned long	 ulCompression;	   /* Compression scheme used to store the bitmap */
	  unsigned long	 cbImage;		   /* Length of bit-map storage data in chars*/
	  unsigned long	 cxResolution;	   /* x resolution of target device          */
	  unsigned long	 cyResolution;	   /* y resolution of target device          */
	  unsigned long	 cclrUsed;		   /* Number of color indices used           */
	  unsigned long	 cclrImportant;	   /* Number of important color indices      */
	};
	struct bitmap_info_header2
	{
	  unsigned long	 size;
	  unsigned long	 cx;			   /* Bit-map width in pels                  */
	  unsigned long	 cy;			   /* Bit-map height in pels                 */
	  unsigned short cPlanes;		   /* Number of bit planes                   */
	  unsigned short cBitCount;		   /* Number of bits per pel within a plane  */
	  unsigned long	 ulCompression;	   /* Compression scheme used to store the bitmap */
	  unsigned long	 cbImage;		   /* Length of bit-map storage data in chars*/
	  unsigned long	 cxResolution;	   /* x resolution of target device          */
	  unsigned long	 cyResolution;	   /* y resolution of target device          */
	  unsigned long	 cclrUsed;		   /* Number of color indices used           */
	  unsigned long	 cclrImportant;	   /* Number of important color indices      */
	};
	void init(void);
		bitmap_file_header	bmfh;
		bitmap_info			bmi;
		union {
			bitmap_info_header1	bmih1;
			bitmap_info_header2	bmih2;
		} bmih;
		union {
			RGB3 rgb3[256];
			RGB4 rgb4[256];
		} rgb;

		long		*cUsedTable;
		char		*original_name;
		void		turn_bitmap(int cnt, FGPixel *from, FGPixel *to);
		void		DoPalette(int);
		int			pcx(char *n, char *gifp, int& one, unsigned int& bits, unsigned int& c);
		int			gif(char *n, char *gifp, int& one, unsigned int& bits, unsigned int& c);
		int			jpg(char *n, char *gifp, int& one, unsigned int& bits, unsigned int& c);
		int			png(char *n, char *gifp, int& one, unsigned int& bits, unsigned int& c);
		int			tga(char *n, char *gifp, int& one, unsigned int& bits, unsigned int& c);
	protected:
	public:
		/**
			Get original file name. Used for file operations.
		*/
		char *GetOriginalName(void) { return original_name; }
		/**
			Set original file name. Used for file operations.
		*/
		void SetOriginalName(char *s) { if (original_name) free(original_name); original_name = strdup(s); }
		/**
		This creates "in memory Bitmap" object from the bitmap file with name.
		The file must be one from supported and enabled image formats.
		If an operation is OK, then methods data member type is != BMP_NONE,
		else file doesn't exist or isn't a valid image file.
		*/
		Bitmap(char	*name);
		//! deprecated
		Bitmap(void	*memimage);
		Bitmap(DrawBuffer *);
		Bitmap(int,int,FGPixel=CBLACK);
		virtual ~Bitmap();
		int		BitmapSave(char *);
};
#ifndef __GNUC__
#pragma	pack()
#endif

/**
When you write the code to manage some menus, sometimes you need program
the group of buttons, that would worked as welded. When you choose one,
previous goes to the inactive state - this is called a ButtonGroup -
and make possible to you easy aggregate items like CheckButtons,
PointButtons, or PushButton (the word or is at a right place,
because you can't mix these types together!).
The default max. used items at once are 32.
*/
class ButtonGroup
{
		friend class Control;
		void RefreshGroup(Control *c);
		short	count;
		short	size;
		short	type; // 1-check,point; 2-push; 0 undefined
		Control * 	curr;
		Control **	array;
	protected:
	public:
		/**
		Creates an empty group of the Control. Default is 32 items there. You can pass your own value.
		*/
		ButtonGroup(int maximum=32)
		{
			size = maximum;
			count = type = 0;
			curr = 0;
			array = new Control *[size];
		}
		~ButtonGroup() { delete [] array; }
		/**
        Add Controls object to the ButtonGroup. The argument active is
		not mandatory (default is false) and if you set this to true,
		then object will have the active state at the start.
		If you mix no acceptable items together, or if you add other
		than allowed types, an error dialog will be showed.
        */
		void AddToGroup(Control *p, int activ = 0);
		/**
			Reset the group. You can add controls now again. 
		*/
		void Clear(void) { type = count = 0; curr = 0; }
};

//! The predecessor for the all widgets.
/**
This family of graphics controls contains the class PushButton, CheckButton,
EditBox and Base menu. It is child objects of class Controls, which is
a child object of base class BaseGui. There is a more information.
NOTE: All of the Control class objects and its parents are on the create
time automatically attached with its Window. When a Window is destroyed,
Controls are destroyed either automatically. Don't delete it explicitly!
@code
void call_back_procedure(CallBack cb)
{
	... // your code
}
@endcode
*/
class Control :	public BaseGui
{
		friend class Window;
		friend class ButtonGroup;
		friend class App;

		ControlCall	GetHandler(void)  const { return fnc; }

		void		RegisterToGroup(ButtonGroup *g) { grp = g;}
		void		RefreshGroup(void);
		void		RemoveFromGroup(void) { grp = 0; }
		static 		Control * ButtonFind(GuiEvent *e,int	n);

	protected:
		int			local_id;
		int			key;		   	// hotkey
		ControlCall	fnc;			// handler
		Window		*owner;			// owner of this
		ButtonGroup *grp;			// if is group members

		Control(int	xs,	int	ys,	int	ws,	int	hs,	char *nm, Type typ, int key, int flag, Window *, ControlCall f=0, FGPixel p=CGRAY1, FGPixel i=CDARK);
		virtual		~Control();

		int	   		GetXr(void);
		int	   		GetYr(void);
		void		Underscore(int xo, int yo, int c);
		inline void frame(void);
		void 		update_owner(void);
		virtual		void ClickDown(void) {}
	public:
		/**
		Redraws whole the object.
		*/
		virtual		void draw(void)=0;
		/**
		Change the name of the object and redraws one.
		*/
		virtual		void SetName(char *s) {	BaseGui::SetName(s); draw(); }
		/**
		Callback on the Control activation by user by click or hotkey pressing.
		You can override this in child object.
		*/
		virtual		void OnActivate(void) { RunSignal(); }
		/**
		Change the 'hotkey' of the object.
		*/
		void		SetKey(int k) {	if (k>='a' && k<='z') k	= toupper(k); key=k; draw(); }
		/**
			Returns current 'hotkey'.
		*/
		int			GetKey(void) const { return key; }
		/**
		Changes the state of the object (draws it as normal and allows activate one).
		*/
		void		Enable(void);
		/**
		Changes the state of the object (draws it as grayed and don't allows activate one).
		*/
		void		Disable(void);
		/**
		Change the state of the object to state "active" (as you as click
		with mouse at one). It is a good choice if you want input
		to the EditBox immediately without clicking at one.
		*/
		virtual		void ClickUp(int a);  // if TRUE, so call handler
		/**
		With this procedure you can change the state of the two-state (ON/OFF) switches.
		*/
		void		SetTrigger(int a) {	if (a) status |= WTRIGGER;	else	status &= (~WTRIGGER);	draw();}
		/**
		Returns the state of the Control (TRUE or FALSE).
		*/
		int	   		GetTrigger(void) const { return status&WTRIGGER?1:0;	}
		/**
		Returns the parent Widget, i.e. window that contains this Control.
		*/
		Window		*GetOwner(void) const { return owner;	}
		/**
		Each Control in the Window has local ID. You can use this ID with handling ACCELEVENT by example.
		The all controls have local id. This value goes from 0 to .., for each
		window and its control items. For example: when window contains 5 buttons,
		then these buttons have local id values from 0 to 4,according to the order
		of its creating. This is useful for fast and easy testing which button
		(when you use one call-back procedure for more buttons i.e.).
		*/
		int			GetLocalId() const { return local_id; }
		/**
		Set new foreground color and redraws object.
		*/
		void		SetInk(FGPixel i) { _ink = ink = i; draw(); }
		/**
		Set new background color and redraws object.
		*/
		void		SetPaper(FGPixel p) { _paper = paper = p; draw(); }
		/**
		Set a new handler for object.
		*/
		void		SetHandler(ControlCall w) {	fnc=w; }
		/**
		You can use this static method as predefined callback when you can
		close the parent Window.
		@code
			my_window->AddBaseMenu("Close the window", 'C', Control::Close);
		@endcode
		*/
		static		void Close(CallBack cb);
		/**
		You can use this static method as predefined callback when you can
		close the parent Window.
		@code
			my_window->AddBaseMenu("Quit the Application", 'C', Quit::Close);
		@endcode
		*/
		static		void Quit(CallBack cb);
};

/**
	Use Label to add text that user can't edit to a Window. This text can be
	used to label another Control object. To add an object to a Window that displays text that
	a user can edit use EditBox.

	This button can be grouped into the set of mutually exclussive options
	to the user - that is, only one radio button in a set can be selected
	at a time. When the user selects a button, the previously selected
	button becomes unselected. Buttons are frequently grouped in
	a button group box (ButtonGroup). Add the group box to the Window first,
	then put the buttons into the group box.
	For example, two buttons on a Window can be checked at the same time only if they are
	contained in separate containers, such as two different group boxes.

	Use Window::AddLabel() instead.
*/
class Label : public Control {
	protected:
		virtual	void draw(void);
		void ClickUp(int a);
		unsigned short int is_transparent:1, high_light:1;
	public:
		Label(int xs, int ys, char	*nm, int key, Window *w, ControlCall f, int i, int p);
		void SetTransparent(void) { is_transparent = 1; draw(); }
		void MoveIn(void)
		{
			high_light = 1;
			draw();
		}
		void MoveOut(void)
		{
			high_light = 0;
			draw();
		}
};

/**
	Use PushButton to put a standard push button on a Window. PushButton
	introduces several properties to control its behaviour in a dialog box settings.
	Users choose button controls to initiate actions.

	@image html pushbutt.png

	This button can be grouped into the set of mutually exclussive options
	to the user - that is, only one radio button in a set can be selected
	at a time. When the user selects a button, the previously selected
	button becomes unselected. Buttons are frequently grouped in
	a button group box (ButtonGroup). Add the group box to the Window first,
	then put the buttons into the group box.
	For example, two buttons on a Window can be checked at the same time only if they are
	contained in separate containers, such as two different group boxes.

	To add this Control use Window::AddPushButton()
*/
class PushButton : public Control {
		DrawBuffer	*icon;
	protected:
			virtual	void draw(void);
	public:
		PushButton(int xs, int ys, int ws, int hs, char	*nm, int key, int flag,	Window *w, ControlCall f=0)
			: Control(xs, ys, ws, hs, nm, PUSHBUTTON, key, flag, w,	f)
			{
				icon = 0;
				draw();
			}
		PushButton(int xs, int ys, int key,	int	flag, Window *w, DrawBuffer	*bm, ControlCall f=0)
			: Control(xs, ys, bm->GetW() + 4, bm->GetH() + 4, "", PUSHBUTTON, key, flag|WBITMAP, w, f)
			{
				icon = bm;
				draw();
			}
		void ClickDown(void);
		void ClickUp(int);
		void MoveIn(void);
		void MoveOut(void);
		void Push(void);
		void Release(void);
};

/**
	CheckButton represents a check box that can be on (checked)
	or off (unchecked). It differents from PointButton only visually.

	@image html check.png

	This button can be grouped into the set of mutually exclussive options
	to the user - that is, only one radio button in a set can be selected
	at a time. When the user selects a button, the previously selected
	button becomes unselected. Buttons are frequently grouped in
	a button group box (ButtonGroup). Add the group box to the Window first,
	then put the buttons into the group box.
	For example, two buttons on a Window can be checked at the same time only if they are
	contained in separate containers, such as two different group boxes.

	To add this Control use Window::AddCheckButton() .
*/
class CheckButton :	public Control {
	protected:
		virtual	void draw(void);
		int		mask;
		int		* variable;
		void MoveIn(void);
		void MoveOut(void);
	public:
		CheckButton(int m, int	xs,	int	ys,	char *nm, int key, int flag, Window	*w,	int	fg,	int	bg,	ControlCall f=0, int *var=0)
			: Control(xs, ys, strlen(nm)?strlen(nm)*8+20:16, 16, nm, CHECKBUTTON,	key, flag, w, f,bg,fg)
			{
				variable = var;
				mask = m;
				if (variable) SetTrigger(*variable&mask); else draw();
			}
		void ChangeItem(int *v) { variable = v; SetTrigger(*v&mask); draw(); }
};

/**
	PointButton represents a radio button that can be on (checked)
	or off (unchecked). It differents from CheckButton only visually.

	@image html point.png

	This button can be grouped into the set of mutually exclussive options
	to the user - that is, only one radio button in a set can be selected
	at a time. When the user selects a button, the previously selected
	button becomes unselected. Buttons are frequently grouped in
	a button group box (ButtonGroup). Add the group box to the Window first,
	then put the buttons into the group box.
	For example, two buttons on a Window can be checked at the same time only if they are
	contained in separate containers, such as two different group boxes.

	To add this Control use Window::AddCheckButton() .
*/
class PointButton :	public Control {
	protected:
		virtual	void draw(void);
		int		mask;
		int		* variable;
		void MoveIn(void);
		void MoveOut(void);
	public:
		PointButton(int m, int	xs,	int	ys,	char *nm, int key, int flag, Window	*w,	int	fg,	int	bg,	ControlCall f=0, int *var=0)
			: Control(xs, ys, strlen(nm)?strlen(nm)*8+20:16, 16, nm, POINTBUTTON,	key, flag, w, f,bg,fg)
			{
				variable = var;
				mask = m;
				if (variable) SetTrigger(*variable&mask); else draw();
			}
		void ChangeItem(int *v) { variable = v; SetTrigger(*v&mask); draw(); }
};

/**
	The two state (like CheckButton) PushButton.
*/
class TwoStateButton : public PointButton
{
	protected:
		virtual void draw(void);
	public:
		TwoStateButton(int xs, int ys, int ws, int hs, char *nm, int key, Window *wnd, int *var=0, ControlCall f=0) :
			PointButton(1, xs,ys,"", key, BNORMAL, wnd, 0, 15, f, var)
		{
			Resize(ws-w, hs-h);
			SetName(nm); // ma v sebe draw
		}
};

/**
You can use this object to get any text or number from user.
When you activate such object, you can use keyboard to edit
the text (HOME, DEL, BACKSPACE, ARROWS and ESC work properly).
The input is passed to your program when the user hits
the 'ENTER' key. The ESC key pressing terminates the input without change.
You can type input line up to 127 characters long. The special shortcuts
CTRL+INS and SHIFT+INS are used to standard COPY&PASTE mechanism.

@image html editbox.png

*/
class EditBox :	public Control {
		friend class FGUpDown;
		friend class Window;
		void init(void);
		enum dtype { EDIT_INT, EDIT_STRING, EDIT_DOUBLE};
		void	draw(int);
		void UpDown(double step)
		{
			if (data_type==EDIT_INT)
			{
				 *(int	*)ptr += int(step);
			}
			else if	(data_type==EDIT_DOUBLE)
			{
				 *(double *)ptr += step;
			}
			TestRange();
		}
		void TestRange(void)
		{
			if (check_range==0) return;
			if (data_type==EDIT_INT)
			{
				int	dato = *(int *)ptr;
				if (dato<min) *(int	*)ptr =	min;
				else if	(dato>max) *(int *)ptr = max;
			}
			else if	(data_type==EDIT_DOUBLE)
			{
				double dato	= *(double *)ptr;
				if (dato<mind) *(double	*)ptr =	mind;
				else if	(dato>maxd)	*(double *)ptr = maxd;
			}
		}
		int	inputproc(unsigned);
		void input(void);
		void FlushInput(void)
		{
			if (isinput)				   // if input, force ESC
			{
				inputproc(CR);
			}
		}
	protected:
		int		w1,w2,min,max;
		dtype   data_type;
		unsigned first:1,passwd:1,check_range:1,hex:1,caps:2,isinput:1;
		double	mind,maxd;
		void  *	ptr;
		char	buf[128];
		static	char clip[128];
		static	int is_clip;
		int		pos, maxpos, iline, icol, offset, size;

		virtual	void draw(void);
		void MoveIn(void);
		void MoveOut(void);
	public:
		virtual ~EditBox();
		/**
		Switch the inputbox into input mode (shows cursor and redirect all Window's keypress into one).
		*/
		void ClickUp(int state=1);
		/**
		Sets password mode. All typed characters will be shown as '*'.
		*/
		void PasswdMode(int m) { passwd = m; draw(); }
		/**
		The input will parsed as a hexadecimal number (without '0x' prefix).
		*/
		void HexMode(int m) { hex = m; draw(); }
		/**
		All letters will be converted to the capitals.
		*/
		void CapsMode(int m);
		/**
		Change the data that are edited by editbox.
		You must call this methods to apropriate EditBox type - this is for an integer type of the object.
		*/
		void ChangeItem(int	*p)
		{
			assert(data_type==EDIT_INT);
			ptr	= p;
			draw();
		}
		/**
		Change the data that are edited by editbox.
		You must call this methods to apropriate EditBox type - this is for an string type of the object.
		*/
		void ChangeItem(char *p)
		{
			assert(data_type==EDIT_STRING);
			ptr= p;
			draw();
		}
		/**
		Change the data that are edited by editbox.
		You must call this methods to apropriate EditBox type - this is for an double type of the object.
		*/
		void ChangeItem(double *p)
		{
			assert(data_type==EDIT_DOUBLE);
			ptr	= p;
			draw();
		}
		/**
		Sets a new size of the input string. Default is calculated from the objects width. You can set values from 1 to 127.
		*/
		void SetSize(int ns)
		{
			pos = offset = 0;
			if(ns>0 && ns<128)
			{
				size = ns;
				draw();
			}
		}
		/**
		Override the disable methods.
		*/
		void Disable(void);
		/**
		Create an editbox for an integer number input. Don't call this directly. Rather use Window::AddEditBox() methods instead.
		*/
		EditBox(int xs,	int ys,	int ws1, int ws2, char *nm, int	key, Window *w,	int *pt, int ink, int paper, ControlCall f, int mn, int mx, int check);
		/**
		Create an editbox for an text input. Don't call this directly. Rather use Window::AddEditBox() methods instead.
		*/
		EditBox(int sz, int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, Window	*w,	char *pt, int ink, int paper, ControlCall f);
		/**
		Create an editbox for an double precission number input. Don't call this directly. Rather use Window::AddEditBox() methods instead.
		*/
		EditBox(int xs,	int ys,	int ws1, int ws2, char *nm, int	key, Window *w,	double *pt, int	ink, int paper,	ControlCall f, double mn, double mx, int check);
};

/**
When your application performs a time-consuming operation, you can use a progress bar
to show how much of the task is completed. A progress bar displays a dotted line
that grows from left to right.

@image html progress.png

*/
class ProgressBar : public Control
{
	protected:
		int		steps, value;
		int		sirka;
	public:
		void	draw(void);
		void	setProgress(int a) { if (a>=0 && a<=steps) value = a; draw(); }
		int		progress(void) const { return value; }
		ProgressBar(Window *w, int xx, int yy, int ww, int hh, int s);
};

/**
	Use BaseMenu to specify the apperance and behaviour of an item
	in a menu. Each Window can contain multiple menu items.
*/
class BaseMenu : public	Control	{
			virtual	void draw(void);
	public:
		void MoveIn(void);
		void MoveOut(void);
		virtual	void ClickUp(int a);  // if TRUE, so call handler
		BaseMenu(char *nm, int key, Window *w, ControlCall f, int font);
};

/**
	Use BaseMenuItem to specify the apperance and behaviour of an item
	in a menu. Each MenuWindow can contain multiple menu items.
*/
class BaseMenuItem : public	Control	{
			virtual	void draw(void);
	public:
		void MoveIn(void);
		void MoveOut(void);
		virtual	void ClickUp(int a);  // if TRUE, so call handler
		BaseMenuItem(char *nm, int key, MenuWindow *w, ControlCall f, int font);
};

/**
The parents for all Sliders.
*/
class Slider  : public Control
{
	protected:
		virtual void draw(void) {}
		int		minv,maxv,*val,steps,smer; // 0 je vodorovny, 1 zvysly
		void 	frame(int,int,int,int,int f=0);
		virtual void MoveIn(void);
		virtual void MoveOut(void);
		virtual void ClickDown(void);
	public:
		void	modify(int);
		void    redraw(void)
		{
			ControlCall tmp = fnc;
			fnc = 0;
			draw();
			fnc = tmp;
		}
		Slider(Window *win, int xs, int ys, int ws, int hs, int minimal, int maximal, int s, int *v, ControlCall f) :
			Control(xs, ys, ws, hs, 0, SLIDEBAR, 0, BNORMAL, win, f), minv(0), maxv(0)
		{
			if (minimal>maximal)
			{
				minv = maximal;
				maxv = minimal;
			}
			else
			{
				minv = minimal;
				maxv = maximal;
			}
			if (*v<minv)
				*v=minv;
			if (*v>maxv)
				*v=maxv;
			val  = v;
			steps= s;
			smer = h>w;
		}
};

/**
	A horizontal Slider.

	@image html sliderh.png

*/
class SlideBarH : public Slider
{
// 3 - 16 - val - 3 - 38
	public:
		SlideBarH(int x, int y, int min, int max, int step, int *val, Window *win, ControlCall f=0):
			Slider(win,x,y,(max-min)/step+16+12+38,16,min,max,step,val,f)
		{
			draw();
		}
		virtual void draw(void);
};

/**
	A vertical Slider.

	@image html sliderv.png

*/
class SlideBarV : public Slider
{
// 3 - 16 - val - 3 - 38
	public:
		SlideBarV(int x, int y, int min, int max, int step, int *val, Window *win, ControlCall f=0):
			Slider(win,x,y,16,(max-min)/step+16+12+38,min,max,step,val,f)
		{
			draw();
		}
		virtual void draw(void);
};

/**
	It is a rectangle space at the screen that is intended to the communication
	with the user. Its look & feel is derived from the flags assigned
	on the create time. The maximal size of the window is limited
	by current video mode resolutions only. It's have own fore
    and back color, title (or no). You can use many windows
    at once (it is limited with free memory only), but only and just
	only one may be in the active state. You can draw your windows everywhere,
    but if you go out of bounds of the screen, then graphics system
    adjust the size and position at the right values. You can set
	that windows to remember its position and size. The each window is active
	immediate after it is created. To the active window goes user's input,
    key pressing, mouse events and other system events. You can serve
	these events by defining your own Window Handler. It is a single
	procedure with type GuiHwnd and it is described below. Because using
	Window methods to a not existing object, library has implemented
	safe mechanism (it is not mandatory but it is a good practise).

	@image html wnd.png

*/
class Window : public BaseGui
{
		friend	class Control;
		friend	class EditBox;
		friend  class BaseMenu;
		friend  class App;
		
		Window(Window **, FGPixel);	// root window
		Control		*oldidm;
		int			curr_ctrl;			// default button
		int			c_menupos;
		unsigned 	first_draw:1, changed:1, nControls:16; // pocet buttonov priradenych oknu
		int			xpos, ypos;
		void 		(*handler)(GuiEvent	*);
		// for locking dynamic (WindowMove(), WindowLine() ... )
		FGMutex 	StateMutex;
		void		*tab_page;	// std::list<FGTabPage>
		int			optimize;
		FGPixel		statuscolor;
		FGPixel		optimize_color;
		EditBox 	*ininput;
		Control 	*call_handler; // set if need call handler (in translateevent)
		int		line_x1, line_y1, line_x2, line_y2;
		const   char * current_tab_page;
		int			icon_x, icon_y;
		Window	 	**itself;
		Control		*Buttony;
		Window		*iconized;
		void	(*CallWindowHook)(int x, int y, int& nx, int& ny, int);
		static	Window *current, *first;  // last in list of all windows
		static	int	nWnd;

		Window *OdkryteOkno(Window *pokial, int x, int y, int w, int h);
		virtual int  iconize(int xs, int ys);
		int		Rozporcuj(int X1, int Y1, int X2, int Y2, int *in, int *out);
		void	intersect(int, int, int, int);
		static  int	_over(int, int, int, int, int xx, int yy, int ww, int hh);
		void	Relocate(int xx, int yy);
		static 	void DestroyWindow(void);
		static	void nextwnd(void);
		EditBox *FindNextEditBox(void);
		int	 	TitleFind(GuiEvent	*e);
		void 	WindowUpdate(int dx, int dy, int all=0);
		BaseGui	*GetLastButton(void); // last in list

		void	ResetActive(void)
		{
			call_handler = 0;
		}
		Control * IsActive(void)
		{
			return call_handler;
		}
		int		IsInInput(void) { return !!ininput; }
		void 	SetInInput(EditBox *e)
		{
			ininput = e;
		}
		void	WindowInputChar(unsigned c)
		{
			ininput->inputproc(c);
		}
		void	SetDefaultControl(int _id)
		{
			curr_ctrl = _id;
		}
		void	WindowUpdateBlock(int x, int y, int w, int h);
		void		SetNextTabPage(void);
		void		DrawTabPages(void);
		int			ClickTabPage(int, int);
		void		RelocateTabPage(int xx, int yy);
		void 		puts(char *s) { while(*s) wputc(*s++);	}
		void		wputc(int);
		void		move(int x=0,	int	y=0) { wputc(-1); xpos=x; ypos=y; }
		//! internal
		int		GetXM(int);
		//! internal
		int		GetYM(void);
	protected:
		//! @deprecated
		void	WindowHide(void);
		//! @deprecated
		void	WindowShow(void);
		//! @deprecated
		void	WindowRecopy(void);

		DrawBuffer 	*icon;

		void	RepaintBlock(int x,	int	y, int w, int h);
		void	init(void);
		void	default_clipping(void);
		long	WindowStatus(long status);
		virtual void WindowPiktograms(int x, int y);
		virtual void WindowDrawTitle(int color); // old virtual
		virtual	void draw(void);
	public:
		//! internal
		void	InstallWindowHook(void(*hook)(int, int, int&, int&, int))
		{
			CallWindowHook = hook;
		}
		//! internal
		void	SetActive(Control *c);
		//! internal
		void	RunHandler(void);
		//! @deprecated
		static	void HideAll(void);
		//! @deprecated
		static	void ShowAll(void);
		static 	void WindowVirtualRoot(int, int, int, int);
		//! Returns current focussed Window if any.
		static	Window * GetCurrent(void)	{ return current; }
		static 	Window *WindowFind(GuiEvent	*e);
		//! Finds a Window by ID.
		static  Window * WindowFind(int idw)
		{
			return (Window *) (first->FindId(idw));
		}
		void 	BroadcastMessage(GuiEvent *e);
		//! Lock the Window to access from other threads.
		void	StateLock(void)
		{
			StateMutex.Lock();
		}
		//! Unlock the Window to access from other threads.
		void	StateUnlock(void)
		{
			StateMutex.Unlock();
		}
		// the window notify interface
		//! Marks the Window state as 'CHANGED'. Works together with notifycation on window close.
		void	SetChange(void) { changed = 1; }
		//! Marks the Window state as 'NOT CHANGED'. Works together with notifycation on window close.
		void	ResetChange(void) { changed = 0; }
		bool	ShowNotify(void);

		//! Sets the icon position.
		void	SetIconPosition(int x, int y) { icon_x = x; icon_y = y; }
		//! Returns true/false if the Window is iconized.
		int		IsIconized(void) const { return !!(int)iconized; }
		//! Iconize the Window at its icon position or if not available, at the mouse position.
		int 	WindowIconize(void);
		//! Iconize the Window at the exact position.
		int 	WindowIconize(int, int);
		//! Attaches the image as icon to the Window.
		void 	WindowAttachIcon(DrawBuffer *ico) { icon = ico; }

		//! Returns the pointer to the first Control in the Window.
		Control	* GetCurrentControls(void) { return	Buttony; }

		void	WindowRepaint(int	xr,	int	yr,	int	w, int h);
		void	WindowRepaintUser(int	xr,	int	yr,	int	w, int h);

		Window(Window **, int xs, int ys, int ws, int hs, char *nm,	GuiHwnd=0, int i=IM, int p=PM, int flag=WTITLED|WFRAMED|WCLICKABLE);
		virtual ~Window();
		bool	WindowClose(void);
		/**
			Flushes the input line if one is currently edited.
			@return true if any input flushed or false if there is nothing to do.
		*/
		bool	WindowFlushInput(void)
		{
			if (ininput)
			{
				WindowInputChar(CR);
				ininput = 0;
				return true;
			}
			return false;
		}

		PushButton	*AddPushButton(int xs, int ys, int ws, int hs, char *nm,	int	key=0, ControlCall f=0, int flag=BNORMAL);
		PushButton	*AddPushButton(int xs, int ys, int key, Bitmap *bm, ControlCall f=0, int flag=BNORMAL);
		SlideBarH	*AddSlideBarH(int xs, int ys, int min, int max, int step, int *val, ControlCall f=0);
		SlideBarV	*AddSlideBarV(int xs, int ys, int min, int max, int step, int *val, ControlCall f=0);
		CheckButton	*AddCheckButton(int	xs,	int	ys,	char *nm, int key=0, int * variable=0, ControlCall=0, int flag=BNORMAL);
		PointButton	*AddPointButton(int	xs,	int	ys,	char *nm, int key=0, int * variable=0, ControlCall=0, int flag=BNORMAL);
		CheckButton	*AddCheckButtonMask(int mask, int	xs,	int	ys,	char *nm, int key=0, int * variable=0, ControlCall=0, int flag=BNORMAL);
		PointButton	*AddPointButtonMask(int mask, int	xs,	int	ys,	char *nm, int key=0, int * variable=0, ControlCall=0, int flag=BNORMAL);
		TwoStateButton *AddTwoStateButton(int	xs,	int	ys,	int w, int h, char *nm, int key=0, int * variable=0, ControlCall=0);
		EditBox		*AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, int *p, ControlCall f, int min, int max);
		EditBox		*AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, int *p, ControlCall f=0);
		EditBox		*AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f=0);
		EditBox		*AddEditBox(int sz, int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f=0);
		EditBox		*AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, double *p, ControlCall f, double min, double max);
		EditBox		*AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, char *nm,	int	key, double *p, ControlCall f=0);
		BaseMenu	*AddBaseMenu(char *nm,	int	key=0, ControlCall f=0);
		ProgressBar	*AddProgressBar(int xs, int ys, int ws, int hs, int size);
		FGListBox 	*AddListBox(int xs, int ys, int w, int h, int dropdown, FGListBoxCallBack drawone);
		Label		*AddLabel(int x, int y, char *s, int key=0, ControlCall f=0, int i=-1, int p=-1);

		//! @deprecated for compatibility only!!
		FGListBox  	*AddListBox(int xs, int ys, int wpol, int hpol, int w, int h, int *cnt, int *pos, FGListBoxCallBack drawone, void *data=0);
		//! @deprecated for compatibility only!!
		FGListBox 	*AddListBox(int xs, int ys, int wpol, int hpol, int w, int h, int cnt, FGListBoxCallBack drawone, void *data);

		void 		WindowText(int	x, int y, char *s);
		void 		WindowText(int	x, int y, char *s, int color);
		void 		WindowText(int	x, int y, char *s, int color, int bk);
		void 		WindowText(int x, int y, char *s, FontProperty * f, int color, int bk);
		void 		WindowTextUTF8(int x, int y, char *s, FontProperty * f, int color, int bk);
		void 		WindowTextUnicode(int x, int y, const unsigned short *s, FontProperty * f, int color, int bk);
		void 		WindowPanel(int, int, int, int, char *s=0, int ink=CWHITED, int ink2=CDARK);
		void 		WindowBox(int x, int y, int a,	int	b);
		void 		WindowBox(int x, int y, int a,	int	b, int c);
		void 		WindowPixel(int x,	int	y, int c);
		void 		WindowPixel(int x,	int	y);
		void 		WindowRect(int	x, int y, int a, int b);
		void 		WindowRect(int	x, int y, int a, int b,	int	c);
		void 		WindowPatternRect(int	x, int y, int a, int b, Pattern *);
		void 		WindowLine(int	x, int y, int a, int b);
		void 		WindowLine(int	x, int y, int a, int b,	int	c);
		void 		WindowPatternLine(int	x, int y, int a, int b, Pattern *p=&PatternDot);
		void 		WindowDrawCircle(int x, int y,	int	r, int c);
		void 		WindowDrawCircle(int x, int y,	int	r);
		void 		WindowFillCircle(int x, int y,	int	r, int c);
		void 		WindowFillCircle(int x, int y,	int	r);
		void 		WindowDrawEllipse(int x, int y,	int	rx, int ry, int c);
		void 		WindowDrawEllipse(int x, int y,	int	rx, int ry);
		void 		WindowFillEllipse(int x, int y,	int	rx, int ry, int c);
		void 		WindowFillEllipse(int x, int y,	int	rx, int ry);
		void 		WindowDrawArc(int x, int y,	double ang1, double ang2, int r, int color);
		void 		WindowDrawArc(int x, int y,	double ang1, double ang2, int r);
		FGRect 		WindowFillPolygon(int,	int[][2], int col=-1);
		FGRect 		WindowFillTriangle(int x1, int y1, int x2, int y2, int x3, int y3, int col=-1);
		void 		WindowDrawPolygon(int,	int[][2], int col=-1);
		void 		WindowSpline(int points[8]);
		void 		WindowScrollDown(int xx, int yy, int ww, int hh ,int about);
		void 		WindowScrollUp(int xx, int yy, int ww, int hh ,int about);
		void		WindowScrollLeft(int xx, int yy, int ww, int hh ,int about);
		void		WindowScrollRight(int xx, int yy, int ww, int hh ,int about);
		void 		WindowStatusBar(int, char *s, int c=CBLACK);

		int	 		printf(const char * , ... );
		int	 		printf(int,int,const char *, ... );

		//! The helper for event sending to the Window.
		void 		SendEvent(int event, int key=0, int x=0, int y=0)
		{
			GuiEvent e(event,key,x,y);
			SendToWindow(&e);
		}
		//! The helper for event sending to the Window.
		void 		SendEvent(int event, int x, int y, int ww, int hh)
		{
			GuiEvent e(event,0,x,y,ww,hh);
			SendToWindow(&e);
		}
		//! Send the event to the Window handler.
		void 		SendToWindow(GuiEvent *p);

		/**
			Lock the Window for the immediate update to the screen. This lock disable
			the Window redrawing on the screen until WindowUnlock() is called.
			Useful when you draw many small objects at once and you want optimize for speed.
			@note this is not related to threading
		*/
		void 		WindowLock(void) {	status|=WLOCKED; }
		/**
			Unlock (and update on the screen if needed) the Locked ( WindowLock() ) Window.
			@note this is not related to threading
		*/
		void 		WindowUnLock(void);
		void 		WindowResize(int, int);
		void 		WindowMove(int,int);
		void 		WindowFocus(void);
		void 		WindowShape(int, int, int, int);
		void 		WindowPutBitmap(int x,	int	y, int xs, int ys,	int	w, int h, DrawBuffer *p);
		#define		WindowCopyFrom WindowPutBitmap
		void 		WindowGetBitmap(int x,	int	y,	int	xs,	int	ys,	int	w, int h, DrawBuffer *p);
		MODAL_RETURN ShowModal(void);
		bool 		WindowSetWorkRect(int,int,int,int);
		//! Removes all the controls from window (buttons, listoboxes, editboxes etc.).
		void 		RemoveControls(void);
		//! Disables all the controls in the Window at once.
		void		DisableControls(void);
		//! Enables all the controls in the Window at once.
		void		EnableControls(void);
		//! Sets the Window foreground color for layout and default foreground color for drawing.
		void		SetInk(FGPixel i)
		{
			_ink = ink = i;
		}
		//! Sets the Window background color for layout and default background color for drawing.
		void		SetPaper(FGPixel p)
		{
			_paper = paper = p;
		}
		//! Sets the new caption for the Window.
		void 		SetName(char *s)
		{
			BaseGui::SetName(s);
			if (this==current) WindowStatus(WACTIVE);
			else WindowStatus(WDEACTIVE);
		}
		/**
		Adds a TabPage to the Window. This Tab will contens all controls added to the Window before this.
		From now, you can add controls to the next Tab.
		@param name caption / id string
		@note that the call to this function creates a tab with controls that
		are ALREADY in the window but not with controls that you will add after this!
		@see SetTabPage() GetCurrentPage()
		*/
		void		AddTabPage(const char *name);
		/**
		Returns current tab page name or null.
		@see SetTabPage()
		*/
		const char *GetCurrentTabPage(void)
		{
			return current_tab_page;
		}
		/**
		Sets the new current tab page.
		@param name caption / id string
		@see GetCurrentPage()
		*/
		FGTabPage * SetTabPage(const char *name);
		//! Removes all tab pages from the window.
		void		DeleteTabPages(void);

		// the virtual callback interface
		virtual void OnKeyPress(int) { }
		virtual void OnMouseMove(int , int) { }
		virtual void OnClick(int, int) { }
		virtual void OnContextPopup(int, int) { }
		virtual void OnPaint(void) { }
		virtual void OnStartDrag(int , int, int) {}
		virtual void OnEndDrag(int ,int, int, int, int) {}
		virtual void OnMove(int , int) { }
		virtual void OnResize(int , int) { }
		virtual void OnIconize(void) { }
		virtual void OnAccelerator(Control *) { }
};

/**
	The instance of this class encapsulates the application.
	You have to declare this construct at the start of your
	program. See the "Hello World" example.
*/
class App : public FGConnector
{
		friend class Window;
		int mouseflag;
		static int mx, my;
		static  int 	fulldrag;
		int mb, oldx, oldy, oldb, drag, action;
		int XX, YY;
		static int __fg_fps, __fg_fps_total;

		int mousefirst, resize, tmp_event,
			resizex, resizey,
			dragx, dragy,
			clickx,	clicky,	clickw,	clickh,
			vyrez;

		Control *idcb;
		Control *idc2;
		Window *drag_okno;
		static int repeat_status, repeat_delay1, repeat_delay2;
		GuiEvent *hold_event;

		enum { MAX_EVENT_QUEQUE=32	}; // numbers of events to app per one user event

		void blue_rect(int x, int y, int w, int h);
		void blue_rect2(int x, int y, int w, int h);
		void AutoRepeatStart(Control *);
		void AutoRepeatDo(void);
		void AutoRepeatEnd(void);
		void DoubleClick(int& event);

		static	GuiEvent queque[MAX_EVENT_QUEQUE];
		static	void (*DelayProc)(void);
		static	void (*OnEverySecond)(int);
		static	int	quequeIndex;
		static  Window *Root;
		static	MainHwnd appHandler;
		static  unsigned long ticks;
		static  int ctrl_break;
		static  int startXdrag;
		static  int startYdrag;
		static  int endXdrag;
		static  int endYdrag;
		static void (*DragShape)(int , int ,int	, int );
		static FGPixel __p4[8];
		static Pattern __patt;
		void CallDelayProc(void)
		{
			if (DelayProc)
			{
				DelayProc();
				return;
			}
			else if (OnIdle() == true)
			{
				return ;
			}
			// wait 10 msec if nothing to do
			delay(10);
		}
		void GetUserEvent(GuiEvent& e, int);
		void TranslateUserEvent(GuiEvent& e);
		void Timer(void);
		int isgmkey(int	k);
	protected:
	public:
		int  RemoveMousePointer(void);
		void ShowMousePointer(void);
		int get_key(void);
		static  FGMouseCursor *__fg_cursor;
		//! Internal flags.
		static	int		flags;
		//! Videomode number that is used for application.
		static	int		video;
		//! The color of application background.
		static	int		background;
		//! Standard C 'argc' on startup.
		static	int		Argc;
		//! Standard C 'argv' on startup.
		static	char	**Argv;
		//! Name of the application exefile.
		static	char	* name;
		//! Home directory (e.g. '/home/user' or '/root' ).
		static	char	*homedir;
		//! The application's starts directory.
		static  char	currdir[128];
		//! true/false if TrueType fonts are supported.
		static	int 	ttf_support;
		/**
		Set the procedure that will be call many times per second when application will be idle.
		When the system is waiting to user input, it waits into inner loop.
		Because this time, I add to the library little hack. You can define
        some very little bit of code, procedure that will be called
		many times per second. Call with parameter 0 to switch this feature off.
		@param fnc your callback procedure of type void (*fnc)(void).
		*/
		void SetDelayProc(void (*fnc)(void))
		{
			DelayProc =	fnc;
		}
		void UpdateMousePointer(void);
		void SaveScreen(void);
		/**
		By default, after <CTRL+C> hitting application normally shutdowns.
		You can switch this behaviour on/off by single calling this method.
		*/
		static void DisableCtrlBreak(void)
		{
			ctrl_break ^= 1;
		}
        /** To change window drawing when WindowMove event is performed.
			@param a true for opaque, false for window frame drawing only (fast).
		*/
		static void SetWindowMoveStyle(int a)
		{
			fulldrag = a;
		}
		App(int m, int& argc, char **& argv, int bck, int appFlags=APP_ENABLEALTX);
		virtual ~App();
		/**
		When you define the application procedure also,
		you can send to this procedure any event by this method.
		*/
		static	void SendToApp(GuiEvent	*x);
		void	Run(MainHwnd hwnd=0);
		void	FGYield(void);
		/**
		Call this methods, for example, from menu item "EXIT".
		When it is once called, it will causes jump out of the main
		application loop - return from Run().
		*/
		static	void AppDone(void)
		{
			GuiEvent e(QUITEVENT);
			SendToApp(&e);
		}
		/**
		This is another call-back procedure that will be call per each 't' msec.
		Your user-defined procedure expects one int parameter - number
		of second from the start of your program.
		Call with parameter 0 switch this feature off.
		NOTE! resolution is cca. 20ms
		@param p the callback routine
		@param t number of milisec. between two calls the callback (default is 1 sec.)
		*/
		void	SetTimerProc(TimerProc p, int t=1000)
		{
			OnEverySecond =	p;
			ticks = t;
		}
		/**
		   Returns pointer to the ROOT window or 0 if one is not initialized.
		   Be sure before using this function, that you add switch APP_ROOTWINDOW
		   to the App constructor else program will be aborted.
		   You can draw using this pointer with standard window stuff (WindowLine (),  WindowText () etc.)
		   NOTE: You can't use any Controls into this window, it is only for drawing!
		   You can initialize one in App constructor time by passing APP_ROOTWINDOW as flags.
		   @return pointer to whole screen window ROOT.
		*/
		static Window * GetRootWindow(void)
		{
			return Root;
		}
		MODAL_RETURN RunModal(Window *which);
		//! Returns current mouse coordinate in X axis.
		static int	GetMouseX(void)
		{
			return mx;
		}
		//! Returns current mouse coordinate in Y axis.
		static int	GetMouseY(void)
		{
			return my;
		}

		static void SetCaption(char *new_name);

		FGMouseCursor * CursorLoad(FGMouseCursor *cur);

		int EnableBuffering(int mode);
		int Flip(void);
		void DisableBuffering(void);

		ENUM_OS GetOS(void);

        /**
        Set variables with actual data. It is dependent on the app flag APP_MAGNIFIFIER.
		@param x position of top left corner
        @param y position of top left corner
		@param w the width of selected rectangle that is showed when you drag mouse with button hold-down.
		@param h the height of selected rectangle that is showed when you drag mouse with button hold-down.
        */
		static void	GetDragVector(int &x, int &y, int &w, int &h);
        /**
        Set procedure that will be called when you drag&drop to draw out dragging object.
        This procedures has four argument x,y for position in the current window
        and offset_y, offset_x are current offset from [x,y]. To restore default,
        use with argument 0. For more see the RAD project - source rad_prj.cpp.
        */
		static void	SetDragShape(void(*a)(int,int,int,int)=0);
		static void SetRepeatDelay(int	c1,	int	c2);

		// Object-oriented API via overloaded methods
		//! overload this member when you want catch any keypress event.
		virtual void OnKeyPress(int) { }
		//! overload this member when you want catch any mousemove event.
		virtual void OnMouseMove(int , int) { }
		//! overload this member when you want catch the click on left mouse's button event.
		virtual void OnClick(int, int) { }
		//! overload this member when you want catch the click on right mouse's button event.
		virtual void OnContextPopup(int, int) { }
		//! overload this member when you want catch the start of mouse drag. By example for the change of the cursor shape.
		virtual void OnStartDrag(int , int, int) {}
		//! overload this member when you want catch the end of mouse drag.
		virtual void OnEndDrag(int ,int, int, int, int) {}
		//! overload this member when you want catch the cursor arrives screen boundary.
		virtual void OnCursorOut(int)
		{
		}
		/**
		The wrapper for an application idle proc.
		@return If you override this on in your class and you call
		some code from it you must return true from this callback to know that the system
		will not wait. Elsewhere your system may go slowdown.
		@note if call SetDelayProc() with non-null argument then this callback will not be called!
		@see SetDelayProc()
		*/
		virtual bool OnIdle(void)
		{
			return false;	// returns true if do some work
		}
		/**
		The wrapper for application timer proc.
		@note if call SetTimerProc() with non-null first argument then this callback will not be called!
		The default time slice is 1000 msec. You can change this by call SetTimerProc() fith first argument
		null and second with number of msecs for a new timeslice.
		@param secs in secs from the start of app.
		@see SetTimerProc()
		*/
		virtual void OnTimer(int secs)
		{
		}
};

/**
	The pointer to the current App object
	@ingroup Globals
*/
extern App *cApp;
/**
	The pointer to the applicatons Config object
	@ingroup Globals
*/
extern Config *cCfg;

/**
	Used for Pop-up menus.
	@see class Window
*/
class MenuWindow : public Window
{
		friend class Window;
		friend class BaseMenuItem;
		
		static GuiHwnd Proc;
		static MenuWindow *	CurrentMenu;
		static void	MenuWindowHandler(GuiEvent *p);
		int		offset;
		void _initmw(GuiHwnd proc)
		{
			offset = 0;
			Proc = proc;
			CurrentMenu	= this;
			set_font(FGFONT_MENU);
		}
		int	GetXM(void)	const
		{
			return 4;
		}
		int	GetYM(int a)
		{
			return (offset+=a,offset-a);
		}
	public:
		/**
		Creates pulldown menu.
		@param x x coordinate
		@param y y coordinate
		@param w a width of menu
		@param h a height of menu
		@param proc an optional window callback procedure
		*/
		MenuWindow(int x, int y, int w,	int	h, GuiHwnd proc=0):
			Window(0,x,y,w,h,"", MenuWindowHandler, CScheme->menuwindow_fore, CScheme->menuwindow_back,WFRAMED|WCLICKABLE|WNOPICTO|WMENUWINDOWTYPE)
		{
			draw();
			_initmw(proc);
		}
		/**
		Creates pulldown menu at current mouse position.
		@param w a width of menu
		@param h a height of menu
		@param proc an optional window callback procedure
		*/
		MenuWindow(int w, int h, GuiHwnd proc=0):
			Window(0,cApp->GetMouseX()+4,cApp->GetMouseY()+4,w,h, "",MenuWindowHandler,CScheme->menuwindow_fore,CScheme->menuwindow_back,WFRAMED|WCLICKABLE|WNOPICTO|WMENUWINDOWTYPE)
		{
			draw();
			_initmw(proc);
		}
		virtual ~MenuWindow()
		{
			CurrentMenu	= 0;
		}
		BaseMenuItem *AddMenu(char *nm,	int	key=0, ControlCall f=0);
		Label		*AddLabel(char *s, int key=0, ControlCall f=0, int i=-1, int p=-1);
		CheckButton	*AddCheckButton(char *nm, int key=0, int * variable=0, ControlCall f=0, int flag=BNORMAL);
		PointButton	*AddPointButton(char *nm, int key=0, int * variable=0, ControlCall f=0, int flag=BNORMAL);
		CheckButton	*AddCheckButtonMask(int, char *nm, int key=0, int * variable=0, ControlCall f=0, int flag=BNORMAL);
		PointButton	*AddPointButtonMask(int, char *nm, int key=0, int * variable=0, ControlCall f=0, int flag=BNORMAL);
		EditBox	*AddEditBox(int	ws1, int ws2, char *nm,	int	key, int *p,ControlCall f, int min, int max);
		EditBox	*AddEditBox(int	ws1, int ws2, char *nm,	int	key, int *p, ControlCall f=0);
		EditBox	*AddEditBox(int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f=0);
		EditBox	*AddEditBox(int sz, int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f=0);
		EditBox	*AddEditBox(int	ws1, int ws2, char *nm,	int	key, double	*p,	ControlCall f, double min, double max);
		EditBox	*AddEditBox(int	ws1, int ws2, char *nm,	int	key, double	*p,	ControlCall f=0);
		void Separator(void);
};

unsigned int CalculateCRC(unsigned StartCRC, void *Addr, unsigned Size);

#ifdef __BORLANDC__
#pragma option pop /* pop -a switch */
#pragma option pop /* pop -b */
#endif

#endif
