/*
    OpenGUI - Drawing & Windowing library

    Copyright (C) 1996,2003  Marian Krivos

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    nezmar@atlas.sk
	Get & translate user events

*/

/* TODO :
	Zakladna schema je vygenerovat nejaku udalost, tu posla oknu,
	pripadne i aplikacii via handler ak nejaky existuje.
	Nova schema by mala dodrzat staru a pridat pretazenie
	urcitych definovanych sprav. Tieto by mali rozne
	parametre. Pretazenie by zaroven odstavilo posielanie
	konkretnej spravy via handler. Podobne by sa mal spravat
	i aplikacny handler.
*/

#include <signal.h>

#if defined( __linux__) || defined( __rtems__) || defined(__sun__)
#include <sched.h>
#ifdef __rtems__
#include <rtems/kd.h>
#else
#include <sys/kd.h>
#endif
#include <sys/ioctl.h>
#endif

#if defined(__MSDOS__) || defined(_WIN32)
extern void sched_yield(void);
#endif

#include "config.h"
#include "fastgl.h"
#include "_fastgl.h"
#include "drivers.h"

int App::fulldrag = 1;
int App::startXdrag =	0;
int App::startYdrag =	0;
int App::endXdrag	= 0;
int App::endYdrag	= 0;
int App::mx=0, App::my=0;
int App::__fg_fps=0, App::__fg_fps_total=0;
FGMouseCursor * App::__fg_cursor = idc_normal;

#define		REPEAT_DELAY1			22
#define		REPEAT_DELAY2			4
#define		DRAG_TRESHOLD		3

int App::repeat_status=0, App::repeat_delay1=REPEAT_DELAY1, App::repeat_delay2=REPEAT_DELAY2;
static FGPixel m_save[MSZ*MSZ];
void (*App::DragShape)(int , int ,int	, int ) = 0;
int App::ctrl_break=0;
FGPixel App::__p4[8] = {0, 0, 0, 0, 1, 1, 1, 1};
Pattern App::__patt(__p4, 8, 1);

char *GuiEvent::eventName[]= {
	  "NOEVENT", "KEYEVENT", "MOUSEEVENT", "QUITEVENT", "ACCELEVENT",
	  "TERMINATEEVENT", "MOVEEVENT", "CLICKRIGHTEVENT",
	  "CLICKLEFTEVENT",	"DRAGLEFTEVENT", "DRAGRIGHTEVENT",
	  "GETFOCUSEVENT", "LOSTFOCUSEVENT", "INITEVENT",
	  "BUTTONHOLDEVENT", "WINDOWRESIZEEVENT","WINDOWMOVEEVENT",
	  "STARTDRAGLEFTEVENT", "STARTDRAGRIGHTEVENT", "REPAINTEVENT",
	  "NOTIFYEVENT", "ICONIZEEVENT", "CURSOROUTEVENT", "DBLCLICKRIGHTEVENT",
	  "DBLCLICKLEFTEVENT",
  // must by at the end
	  "NONSENSE"};

//! Set caption when application running in windowed mode (X11 or WIN32 GDI).
void App::SetCaption(char *new_name)
{
	__fg_driver->SetCaption(new_name);
}


/**
	Save the whole screen to the image file in the current directory,
	name is created as current time in HEX form.
	This function is called by application automagically when
	you press PRINT_SCREEN key.
*/
void App::SaveScreen(void)
{
	Bitmap *a =	new Bitmap(X_width, Y_width);
	char s[20];
	time_t num = time(0);
	get_block(0,0,X_width,Y_width,a->GetArray());
	sprintf(s,"%x",(unsigned int)num);
	if (verbose) printf("Bitmap '%s' saved\n", s);
	a->BitmapSave(s);
	delete a;
}

/**
	Use this for fine calibration for HOLD_EVENT event.
*/
void App::SetRepeatDelay(int c1, int c2)
{
	if (c1>0 &&	c1<2000)
		repeat_delay1 = c1;
	if (c2>0 &&	c2<500)
		repeat_delay2 = c2;
}

/**
	Test keyboard immediately.
	Try peek any KEY from the INPUT and	return 0 if none, or its keycode.
*/
int App::get_key(void)
{
	GuiEvent ev(0);
	GetUserEvent(ev, 1);
	if (ev.GetType()==KEYEVENT)
		return ev.GetKey();
	else return 0;
}

// start autorepeat routine
void App::AutoRepeatStart(Control *id)
{
	hold_event->key=id->GetId();
	hold_event->accel=id;
	hold_event->guiType=id->GetType(); // od buttonu typ ...
	repeat_status =	1;
	AutoRepeatDo();
	repeat_status =	REPEAT_DELAY1;
}

// end autorepeat routine
void App::AutoRepeatEnd(void)
{
	repeat_status =	0;
}

// process -||-
void App::AutoRepeatDo(void)
{
	if (--repeat_status	<= 0)
	{
		Window::GetCurrent()->SendToWindow(hold_event);
		repeat_status =	REPEAT_DELAY2;
	}
}

/**
	Return coordinates of last DRAG&DROP mouse operation.
*/
void App::GetDragVector(int &a, int &b, int &c, int &d)
{
	a =	startXdrag;
	b =	startYdrag;
	c =	endXdrag;
	d =	endYdrag;
}

/**
	Override current shape that is drawed on the screen
	when D&D operation is performed.
*/
void App::SetDragShape(void (*fnc)(int, int, int, int))
{
	DragShape = fnc;
}

void App::GetUserEvent(GuiEvent& EVENT, int waits)
{
	int	xx2, yy2, rc;
	xx2	= XX;
	yy2	= YY;

	EVENT.type = 0;

	if (mouseflag == 1)
		vector_mouse_cursor(XX, YY, TRUE, mouseflag);
	for (; waits; waits--)
	{
		rc = __fg_driver->get_event(EVENT.type,EVENT.key,EVENT.x,EVENT.y,EVENT.buttons);
#ifdef FG_MOUSEKEYS
		if (EVENT.type == KEYEVENT)
		{
			static int last=0, akcel=2, cbut=0;
			int doit=0;

			if (last != EVENT.key)	// reset accel
				akcel = 2;

			switch(EVENT.key)
			{
				case FG_MOUSEKEYS_UP:
					EVENT.type = MOUSEEVENT;
					EVENT.y = YY-accel;
					EVENT.x = XX;
					doit = 1;
					break;
				case FG_MOUSEKEYS_DOWN:
					EVENT.type = MOUSEEVENT;
					EVENT.y = YY+accel;
					EVENT.x = XX;
					doit = 1;
					break;
				case FG_MOUSEKEYS_LEFT:
					EVENT.type = MOUSEEVENT;
					EVENT.x = XX-accel;
					EVENT.y = YY;
					doit = 1;
					break;
				case FG_MOUSEKEYS_RIGHT:
					EVENT.type = MOUSEEVENT;
					EVENT.x = xx+accel;
					EVENT.y = YY;
					doit = 1;
					break;
				case FG_MOUSEKEYS_BUTTON1:
					EVENT.type = MOUSEEVENT;
					EVENT.x = XX;
					EVENT.y = YY;
					cbut ^= 1;
					doit = 1;
					break;
				case FG_MOUSEKEYS_BUTTON2:
					type = MOUSEEVENT;
					EVENT.x = XX;
					EVENT.y = YY;
					cbut ^= 2;
					doit = 1;
					break;
			}
			if (doit)
			{
				if (EVENT.akcel<16)		// max accel diff
					EVENT.akcel += 2;
				// round up co-ordinates
				if (EVENT.x<0)
					EVENT.x = 0;
				if (EVENT.x>(X_width - 1))
					EVENT.x = (X_width - 1);
				if (EVENT.y<0)
					EVENT.y = 0;
				if (EVENT.y>Y_max)
					EVENT.y = Y_max;
				EVENT.buttons = cbut;
			}
			last = EVENT.key;
		}
#endif
		if (rc==0)
		{
			sched_yield();
		}
		else
		{
			if (EVENT.type==KEYEVENT && EVENT.key==PRTRSC)
			{
				SaveScreen();
				continue;
			} // PrtScr
#if defined(__linux__) || defined(__rtems__) || defined(_WIN32)
			else if (EVENT.type==KEYEVENT && EVENT.key==0x0003 && ctrl_break==0)
			{
				raise(SIGINT);
				exit(2);
			}
#endif // LINUX
			XX = EVENT.x;
			YY = EVENT.y;
			break;
		}
		if (waits != 1)
		{
			App::CallDelayProc();			// call delay handler
			App::Timer();                   // call timer handler
			if (repeat_status>0)
				AutoRepeatDo();
		}
	}
//	printf("typ %d, key %x, x %d, y %d b %d\n", type,key,x,y,buttons);
	if (mouseflag == 1)
		vector_mouse_cursor(xx2, yy2,FALSE, mouseflag);
	if (EVENT.type ==	KEYEVENT)
		mouseflag =	0;
	else
		mouseflag =	1;
}

/**
	Redraw mouse pointer if needed.
*/
void App::UpdateMousePointer(void)
{
	if (mouseflag == 1)
		vector_mouse_cursor(mx, my, TRUE, mouseflag);
}

/**
	Hide mouse pointer if needed.
*/
int App::RemoveMousePointer(void)
{
	if (mouseflag == 1)
		vector_mouse_cursor(mx, my, FALSE, mouseflag);
	mouseflag =	0;
	return mouseflag;
}

/**
	Redraw mouse pointer unconditionally.
*/
void App::ShowMousePointer(void)
{
	mouseflag =	1;
	vector_mouse_cursor(mx, my, TRUE, mouseflag);
}

void App::blue_rect(int x, int y,	int	w, int h)
{
	draw_pattern_box(x,y, w-1, h-1, &__patt, _GNOT);
}

void App::blue_rect2(int x, int y, int w, int h)
{
	if (!DragShape) blue_rect(x,y,w,h);
	else DragShape(x,y,w,h);
}

int App::isgmkey(int k)
{
	Window *c =	Window::GetCurrent();

	switch (k)
	{
		case CTRL_TAB:			// ctrl+TAB
	 		if (c) Window::nextwnd();
			return 1;
		case ALT_X:			// alt+X
			if (flags	& APP_ENABLEALTX)
			{
				AppDone();
				return 1;
			}
			else return	0;
		case TAB:
			if (c) if (c->GetStatus() & WNOTEBOOK)
			{
				c->SetNextTabPage();
			}
			break;
		case ESC:
			if (c) if (!(c->GetStatus() & WESCAPE))
				break;
		case ALT_F04:			// alt+F4
			if (c)
				c->WindowClose();
			return 1;
		case ALT_F02:			// alt+F12
			if (c)
			{
				if (c->Window::WindowIconize()) return 1;
			}
			return 0;
	}
	return 0;
}

void App::DoubleClick(int& event)
{
	static int last=0, tmout=0;
	int now = FGClock();
	if (event == last && (now-tmout) <= FGDBLCLICKTIME)
		event += (DBLCLICKLEFTEVENT-CLICKLEFTEVENT);
	else
	{
		last = event;
		tmout = now;
	}
}

// event handling core
void App::TranslateUserEvent(GuiEvent& EVENT)
{
	Window *n =	0, *curr = Window::GetCurrent(), *ecurr=0, *idw=0;
	Control *idm=0;
	int	magn = flags&APP_MAGNIFIER;
	EditBox *ebox=0;

	EVENT.guiType = 0;

	GuiEvent app(NOEVENT); // events to  application

	if (curr)
		switch (EVENT.type)
		{
			case KEYEVENT:
				if (curr && !curr->IsInInput())	//ai nieje input
				{
					if (isgmkey(EVENT.key))
						return ;
					idc2 = Control::ButtonFind(&EVENT, 7);
					if (curr->IsVisible() && idc2)	//tak zisti buttony
					{
						idc2->ClickDown();
						EVENT.type = ACCELEVENT;	//bude to sprava
						EVENT.guiType	= idc2->GetType(); // od buttonu typ ...
						EVENT.accel = idc2;
						EVENT.key	= idc2->GetId();
						delay(100);
						Puk();
						idc2->ClickUp(TRUE);
						curr = Window::GetCurrent();
					}
					else // plain KEY - pass to AppProc
					{
						app.type = KEYEVENT;
						app.key  = EVENT.key;
					}
				}
				else
				{
					if (EVENT.key==TAB)
					{
						ebox = curr->FindNextEditBox();
						ecurr = curr;
						if (curr)
							curr->WindowInputChar(CR);
					}
					else
						if (curr)
							curr->WindowInputChar(EVENT.key);
					EVENT.type = NOEVENT;
				}
				break;
			case MOUSEEVENT:
				if (mousefirst)
				{
					mousefirst = 0;
					mx = oldx =	EVENT.x;
					my = oldy =	EVENT.y;
					mb = oldb =	EVENT.buttons;
				}
				mx = EVENT.x;
				my = EVENT.y;
				mb = EVENT.buttons;
				tmp_event = EVENT.type;
				EVENT.type = NOEVENT;
				// second phase for windowresize
				if (resize && ((mb & 1)	== 1))
				{
					if (fulldrag && !(curr->GetStatus()&WFASTMOVE))
					{
						curr->WindowResize(mx -	oldx, my - oldy);
					}
					else
					{
						blue_rect(curr->GetX(),	curr->GetY(), curr->GetW() + resizex, curr->GetH() + resizey);
						resizex	+= (mx - oldx);
						resizey	+= (my - oldy);
						blue_rect(curr->GetX(),	curr->GetY(), curr->GetW() + resizex, curr->GetH() + resizey);
					}
					resize++;
					break;
				}
				// last phase for windowresize
				else if	(resize	&& (!mb))
				{
					if (!fulldrag || curr->GetStatus()&WFASTMOVE)
					{
						blue_rect(curr->GetX(),	curr->GetY(), curr->GetW() + resizex, curr->GetH() + resizey);
						if (resizex	+ resizey)
						{
							curr->WindowResize(resizex,	resizey);
						}
					}
					resize = 0;
				}
				// second phase for windowmove
				else if	(drag && ((mb &	1) == 1))
				{
					if (fulldrag && !(curr->GetStatus()&WFASTMOVE))
					{
						curr->WindowMove(mx	- oldx,	my - oldy);
					}
					else
					{
						blue_rect(curr->GetX() + dragx,	curr->GetY() + dragy, curr->GetW(),	curr->GetH());
						dragx += (mx - oldx);
						dragy += (my - oldy);
						blue_rect(curr->GetX() + dragx,	curr->GetY() + dragy, curr->GetW(),	curr->GetH());
					}
					drag++;
				}
				// last phase for windowmove
				else if	(drag && (!mb))
				{
					if (!fulldrag || curr->GetStatus()&WFASTMOVE)
					{
						blue_rect(curr->GetX() + dragx,	curr->GetY() + dragy, curr->GetW(),	curr->GetH());
						if (dragx +	dragy)
							curr->WindowMove(dragx,	dragy);
					}
					drag = 0;
				}
				// release mousebutton on control
				else if	(action	&& (!mb))
				{
					action = 0;
					idcb->ClickUp(TRUE);
					AutoRepeatEnd();
					curr = Window::GetCurrent();
					if (curr->IsVisible() && ((idc2->ButtonFind(&EVENT, 3)) == idcb))
					{
						EVENT.type = ACCELEVENT;	//bude to sprava
						EVENT.guiType	= idcb->GetType();
						EVENT.accel = idcb;
					}
					else
						EVENT.type = NOEVENT;
					EVENT.key	= idcb->GetId();
					break;
				}
				// push mousebutton on control
				else if	(action	&& ((mb	& 1) ==	1))
				{
					int tt = idcb->GetType();
					if (curr->IsVisible() && ((idc2	= Control::ButtonFind(&EVENT,3)) != idcb) && tt!=Control::SLIDEBAR)
					{
						action = 0;
						AutoRepeatEnd();
						if (tt!=Control::POINTBUTTON && tt!=Control::CHECKBUTTON)
							idcb->ClickUp(FALSE);	// nevolat handler !!!!
						break;
					}
					else
					{
						idcb->ClickDown();
					}
				}
				else if	(vyrez && (!mb)) // vyrez end
				{
					if (drag_okno) // right click to wnd
					{
						startXdrag = clickx	- curr->GetX() - curr->GetXW();
						startYdrag = clicky	- curr->GetY() - curr->GetYW();
						endXdrag = clickw;
						endYdrag = clickh;
					}
					else
					{
						startXdrag = clickx;
						startYdrag = clicky;
						endXdrag = clickw;
						endYdrag = clickh;
					}
					if (vyrez > DRAG_TRESHOLD+1)
						if (magn) blue_rect2(clickx, clicky, clickw, clickh);
					if (clickw<0) {	clickx += clickw; clickw = -clickw;	}
					if (clickh<0) {	clicky += clickh; clickh = -clickh;	}
					if (drag_okno) // right click to wnd
					{
						if (vyrez >DRAG_TRESHOLD && oldb == 2)
							EVENT.type = DRAGRIGHTEVENT;
						else if	(vyrez <= DRAG_TRESHOLD	&& oldb	== 2)
						{
							DoubleClick(EVENT.type = CLICKRIGHTEVENT);
						}
						else if	(vyrez > DRAG_TRESHOLD && oldb == 1)
						{
							EVENT.type =	DRAGLEFTEVENT;
						}
						else if	(vyrez <= DRAG_TRESHOLD	&& oldb	== 1)
						{
							DoubleClick(EVENT.type = CLICKLEFTEVENT);
						}
						EVENT.w =	clickw;
						EVENT.h =	clickh;
						curr = drag_okno; // redirect it to right window
						EVENT.x =	clickx - curr->GetX() - curr->GetXW();
						EVENT.y =	clicky - curr->GetY() - curr->GetYW();

					}
					else  // right click to background
					{
						if (vyrez > DRAG_TRESHOLD && oldb == 2)	app.type = DRAGRIGHTEVENT;
						else if	(vyrez <= DRAG_TRESHOLD && oldb	== 2)
						{
							DoubleClick(app.type = CLICKRIGHTEVENT);

						}
						else if	(vyrez > DRAG_TRESHOLD && oldb == 1) app.type =	DRAGLEFTEVENT;
						else if	(vyrez <= DRAG_TRESHOLD && oldb	== 1)
						{
							DoubleClick(app.type = CLICKLEFTEVENT);
						}
						app.x =	clickx;
						app.y =	clicky;
						app.w =	clickw;
						app.h =	clickh;
					}
					vyrez =	0;
					break;
				}
				else if	(vyrez && (mb &	3))	// vyrez continue
				{
					if (vyrez == DRAG_TRESHOLD+1)	if (magn) blue_rect2(clickx,	clicky,	clickw,	clickh);
					if (vyrez > DRAG_TRESHOLD) if (magn) blue_rect2(clickx, clicky, clickw, clickh);
					clickw += (mx -	oldx);
					clickh += (my -	oldy);
					if (vyrez > DRAG_TRESHOLD) if (magn) blue_rect2(clickx, clicky, clickw, clickh);
					vyrez++;
					if (vyrez == DRAG_TRESHOLD+1)
					{
						int type2 =	mb==2?STARTDRAGRIGHTEVENT:STARTDRAGLEFTEVENT;
						if (drag_okno)
						{
							EVENT.type = type2;
							EVENT.x =	clickx - curr->GetX() - curr->GetXW();
							EVENT.y =	clicky - curr->GetY() - curr->GetYW();
						}
						else
						{
							app.type = type2;
							app.x = clickx;
							app.y = clicky;
						}
					}
					else // mouse move only!
					{
						// when ptr is over the window, its receive relative pos
						// app proc receive EACH movement as absolute pos
						if (drag_okno) EVENT.type = MOVEEVENT;
						app.type = MOVEEVENT;
						app.x = EVENT.x;
						app.y = EVENT.y;
						app.key = 0;
					}
				}
				// most first push button - what next?
				else if	((!action) && (!drag) && (!vyrez) && (!resize) && (mb == 1))
				{
					int	t;

					t =	curr->TitleFind(&EVENT);
					if (curr->IsVisible()
					   && ((idcb = Control::ButtonFind(&EVENT, 7)) !=	0) && t	!= 4)
					{
						action = 1;
						idcb->ClickDown();
						AutoRepeatStart(idcb);
					}
					else if	(t)
					{
						// windowmove start
						if (t == 1 && (!oldb) && !(curr->GetStatus() & WUNMOVED))
						{
							drag = 1;
							dragx =	dragy =	0;
							if (!fulldrag || curr->GetStatus()&WFASTMOVE)
							{
								blue_rect(curr->GetX() + dragx,	curr->GetY() + dragy, curr->GetW(),	curr->GetH());
							}
						}
						// close window
						else if	(t == 2)
						{
							if (!curr->WindowClose())
							{
								// return if deleted
								return ;
							}
							curr = Window::GetCurrent();
						}
						// hide
						else if	(t==3) curr->WindowIconize();
						// resize
						else if	(t == 4)
						{
							resize = 1;
							resizex	= resizey =	0;
							if (!fulldrag || curr->GetStatus()&WFASTMOVE)
							{
								blue_rect(curr->GetX(),	curr->GetY(), curr->GetW() + resizex, curr->GetH() + resizey);
							}
						}
					}
					else		// tuklo sa vedla titulku
					{
						n =	Window::WindowFind(&EVENT);
						// do okna
						// prepni na toto okno ak nieje aktivne
						if ((n != curr)	&& (n))
						{
							if (!(curr->GetStatus()	& WMODAL))
							{
								n->WindowFocus();
								EVENT.type = tmp_event;
								TranslateUserEvent(EVENT);
								curr = Window::GetCurrent();
								if (curr != n || !n) return; // window destroyed !!
							}
							else Snd(100,70);
						}
						else
						{
							vyrez =	1;
							drag_okno =	Window::WindowFind(&EVENT); // 0-back, 1-wnd
							clickx = EVENT.x;
							clicky = EVENT.y;
							clickw = clickh	= 0;
						}
					}
				}
				else if	(oldb == 0 && mb ==	2)	// right click down
				{
					vyrez =	1;
					drag_okno =	Window::WindowFind(&EVENT);  // find window for click
					if ((drag_okno != curr) && drag_okno)
					{
						if (!(curr->GetStatus()	& WMODAL))
						{
							drag_okno->WindowFocus();
							EVENT.type = tmp_event;
							TranslateUserEvent(EVENT);
							curr = Window::GetCurrent();
							if (curr != drag_okno || !drag_okno) return; // window destroyed !!
						}
						else Snd(100,70);
					}
					if (curr && curr->IsInInput())  // if input, force ESC
					{
						curr->WindowInputChar(ESC);
					}
					clickx = EVENT.x;
					clicky = EVENT.y;
					clickw = clickh	= 0;
				}
				else if	(mb == 0 &&	oldb ==	0)	// only mouse move
				{
					EVENT.type = MOVEEVENT;
					app.type = MOVEEVENT;
					app.x = EVENT.x;
					app.y = EVENT.y;
					app.key = 0;

					idm	= Control::ButtonFind(&EVENT, 3);

					if (idm	&& idm->GetOwner() != curr)
					{
						break;
					}
					if (idm)
					{
						EVENT.key = idm->GetId();
						if (idm->GetTrigger()) idm = 0;
					}
					else if	((idw =	Window::WindowFind(&EVENT)) != 0)
					{
						EVENT.key	= idw==curr?0:-1;
						curr = idw;		// sender is changed !!!
						EVENT.x =	EVENT.x-idw->GetX()-idw->GetXW();
						EVENT.y =	EVENT.y-idw->GetY()-idw->GetYW();
					}
					else
					{
						EVENT.type = NOEVENT;
						app.type = MOVEEVENT;
						app.x = EVENT.x;
						app.y = EVENT.y;
						app.key = 0;
					}
					if (curr)
					{
						if ((!curr->oldidm && idm) ||	((idm && curr->oldidm) &&	(idm !=	curr->oldidm)))
							idm->MoveIn();
						if ((curr->oldidm	&& !idm) ||	((idm && curr->oldidm) &&	(idm !=	curr->oldidm)))
						{
							curr->oldidm->MoveOut();
							curr->oldidm = 0;
						}
					}
					break;
				}
		}
	else
	{
		switch(EVENT.type)  // send & mouse keys only
		{
			case KEYEVENT:
				if (isgmkey(EVENT.key))
					break;	// exit when no window are
			case MOVEEVENT:
				App::SendToApp(&EVENT);
		}
		return ;
	}							// no window exists

	if (app.type==MOVEEVENT)
	{
		app.wnd = idw;
		EVENT.accel = idm;
    }
	
	if (!action && !drag && !vyrez && !resize && (EVENT.type==MOVEEVENT || app.type==MOVEEVENT))
	{
		int value=0;
		if (oldx==mx)
		{
			if (oldx==0) value |= 1;
			else if (oldx>=(X_width - 1)-1) value |= 2;
		}
		if (oldy==my)
		{
			if (oldy==0) value |= 4;
			if (oldy>=(Y_width-1)-1) value |= 8;
		}
		if (value)
		{
			app.type = CURSOROUTEVENT;
			app.key  = value;
		}
	}
	oldx = mx;
	oldy = my;
	oldb = mb;
	if (idm && curr) curr->oldidm = idm;

	// there are it's magic place where an events goes into the right place
	if (EVENT.type == ACCELEVENT) app = EVENT;

	int ii = curr ? curr->GetId() : -1;

	if (app.type !=	NOEVENT)
		App::SendToApp(&app); // send to app if needed

	if (EVENT.type != NOEVENT)   // send to Window if needed
	{
		if (EVENT.type == ACCELEVENT && curr->IsActive())
			curr->IsActive()->RefreshGroup();

		if (EVENT.type==MOVEEVENT && EVENT.key==-1)
		{
			EVENT.key = 0;
			if (idw) idw->SendToWindow(&EVENT);
		}
		else if ((curr=Window::GetCurrent()) != 0)
		{
			curr->SendToWindow(&EVENT);
		}
	}

	curr = Window::WindowFind(ii);

	if (curr)
		curr->RunHandler(); // call to control handler also if window still exist

	// at this point, call_handler is already NULL

	// switch to the next Editbox
	if (ebox && ecurr && curr==ecurr)
		ebox->ClickUp(1);
}

/**
	Load & set the new mouse pointer.
*/
FGMouseCursor * App::CursorLoad(FGMouseCursor *cur)
{
	if (cur == __fg_cursor)
		return cur;
	FGMouseCursor *old = __fg_cursor;
	int oldf = RemoveMousePointer();
	__fg_cursor = cur;
	if (oldf)
		ShowMousePointer();
	return old;
}

int	__MouseDraw(int x, int y, int a, int& mouseflag)
{
	int	old=mouseflag;
	int	i;
	FGPixel *maska, *mapa;
	static FGPixel work[sizeof(m_save)], tmp[sizeof(m_save)];

	mouseflag =	a;
	if (a) // draw cursor
	{
		VideoToRam(m_save, x-cApp->__fg_cursor->xoff, y-cApp->__fg_cursor->yoff, 0, 0, cApp->__fg_cursor->w, cApp->__fg_cursor->h, cApp->__fg_cursor->w, cApp->__fg_cursor->h);
		memcpy(work, m_save, sizeof(m_save));
		maska =	cApp->__fg_cursor->mask;
		mapa  = cApp->__fg_cursor->bitmap;
		for	(i = 0;	i <	(cApp->__fg_cursor->w * cApp->__fg_cursor->h); i++)
		{
			work[i]	&= *maska;
			work[i]	|= ((*mapa++) &	(FGPixel)(~*maska++));
		}
		RamToVideo(work, 0, 0, x-cApp->__fg_cursor->xoff, y-cApp->__fg_cursor->yoff, cApp->__fg_cursor->w, cApp->__fg_cursor->h, cApp->__fg_cursor->w, cApp->__fg_cursor->h, BITBLIT_COPY, _GSET);
	}
	else // restore background
	{
		VideoToRam(tmp, x-cApp->__fg_cursor->xoff, y-cApp->__fg_cursor->yoff, 0, 0, cApp->__fg_cursor->w, cApp->__fg_cursor->h, cApp->__fg_cursor->w, cApp->__fg_cursor->h);
		if (memcmp(tmp,	work, sizeof(m_save)) != 0) // yes, changed
		{
			for (i = 0; i < (int)sizeof(m_save)/bpp; i++)
			{
				if (tmp[i] != work[i])
				{
					m_save[i] = tmp[i];
				}
			}
		}
		RamToVideo(m_save, 0, 0, x-cApp->__fg_cursor->xoff, y-cApp->__fg_cursor->yoff, cApp->__fg_cursor->w, cApp->__fg_cursor->h, cApp->__fg_cursor->w, cApp->__fg_cursor->h, BITBLIT_COPY, _GSET);
	}
	return old;
}

// mouse cursor drawer
mouser_t   vector_mouse_cursor= __MouseDraw;

/**
	Enable drawing into the off-screen video RAM.
	This feature is intended for flickerless full screen
	animation. Parameter mode is one of these predefined:

	enum { FG_DOUBLEBUFFER, FG_TRIPLEBUFFER, FG_QUADBUFFER };

	More buffers may help for speed but is X-times memory
	hungry and not the all videocards supports it.
	On exit, return number of true allocated buffers (0 if error).
	You can still use BUFFERING mode when function fail but
	screen will flicker - all drawing goes to the visible screen.
*/
int App::EnableBuffering(int mode)
{
	if (__fg_driver)
		return __fg_driver->EnableBuffering(mode);
	return (ENUM_BUFFERING)false;
}

/**
	Call this routine when all objects of the scene are
	drawn. The buffer will be showed at the next screen
	vertical retrace. Return 0 if fail - no buffering is
	allowed/initialized.
*/
int App::Flip(void)
{
	if (__fg_driver)
	{
		__fg_fps++;
		if (verbose)
		{
			gprintf(CWHITE, CBLACK,16,16,"fps %3d", __fg_fps_total);
		}
		UpdateMousePointer();
		return __fg_driver->Flip();
	}
	return 0;
}

/**
	Disable video buffering and release the used video memory.
*/
void App::DisableBuffering(void)
{
	if (__fg_driver)
		__fg_driver->DisableBuffering();
}

/**
	Return current Operating system code.
	Code is one from these OS types.
*/
ENUM_OS App::GetOS(void)
{
#ifdef __QNX__
	return OSTYPE_QNX;
#endif
#ifdef __linux__
	return OSTYPE_LINUX;
#endif
#ifdef __MSDOS__
	return OSTYPE_MSDOS;
#endif
#ifdef _WIN32
	return OSTYPE_WIN32;
#endif
}


