/*
    OpenGUI - Drawing & Windowing library

    Copyright (C) 1996,2002  Marian Krivos

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    nezmar@atlas.sk
    declarations for widgets
*/

//
//	WIDGETS.CC		20.11.1998
//

#ifndef _WIDGETS_H_
#define _WIDGETS_H_

#pragma interface

#include "fastgl.h"
#include "listbox.h"

#ifdef __WATCOMC__
#ifdef __DOS__
#include <direct.h>
#else
#include <dirent.h>
#endif // dos
#elif !defined(_MSC_VER)
#include <dirent.h>
#endif // watcom

#ifdef __BORLANDC__
#include <dir.h>
#endif

#ifdef _MSC_VER
#include <direct.h>
#include "mydirent.h"
#endif

#ifdef __BORLANDC__
#pragma option push -b
#pragma option push -a4
#endif

#define	FDIALOG_OPEN	1
#define	FDIALOG_SAVE	2
#define	FDIALOG_MODAL	4
#define	FDIALOG_SAVEDIR	8

/**
	Internal helper structure for FileDialog
*/
struct mydirent
{
	struct dirent item;
	int		type;	// 0 for dir, 1 for file
};

/**
 A nice widget intended for the file selection by user for various things.
 filename length max. 255 chars
 max. visible is 20 chars
 @see FGFont FGFontManager

 @image html fdialog.png

*/
class FileDialog : public FGConnector
{
		char	path[256];
		char	tmppath[256];
		char 	filename[256];
		char	*filter;
		int 	mode;
		int		files,fls,dirs;
		int		maxfiles;
		FileDialogCallBack fileselect2;
		void	reload(void);
		Window	*wnd;
		EditBox *nameEBox;
		struct 	mydirent *filebuffer, *unused;
		int		param;
		FGListBox *list;
		PushButton *ok, *cancel, *mkdir, *home;
		int		mdir;
		int		Selected(int homedir=0);
		void	Refresh(char *);
		void 	_init(char *dir, char *flt, char *namewnd, int m, int ink, int paper);
		static 	void InputName(CallBack);
		static  void myproc(GuiEvent *p);
		static  void drawone(int flag, Window *wnd,int x,int y,void *str,int index);
	protected:
	public:
		/**
		Creates visual object for the file selection by user.
		@param filesel your callback, it is called when you enter or click a filename
		@param dir the start directory, default is current app's directory or last used (with flag FDIALOG_SAVEDIR)
		@param flt text string that contains file extension patterns separated by space, ".exe .com" by example
		@param namewnd the caption
		@param m mode of operation (SAVE or OPEN). See to SetMode()
		@param ink the Window foreground color
		@param paper the Window background color
		*/
		FileDialog(FileDialogCallBack filesel, char *dir=0, char *flt=0, char *namewnd="File Dialog", int m=FDIALOG_OPEN, int ink=0, int paper=PM);
		virtual ~FileDialog()
		{
			delete [] filebuffer;
			if (wnd) if (wnd->GetStatus() & WEXIST) delete wnd;
		}
		/**
		Returns the current browsed directory.
		*/
		char	*GetDir(void) { return path; }
		//! @deprecated - the filename is passed to the callback directly.
		char	*GetName(void) { return filename; }
		/**
		Change the current directory and reload it.
		*/
		void	SetDir(char *d) { strcpy(path, d); reload(); }
		/**
		Change the current input filter mask (".cpp .h" by example) and reload it.
		*/
		void	SetFilter(char *f) { filter=f; reload(); }
		/**
		Sets object's mode:
		-	FDIALOG_OPEN
		-	FDIALOG_SAVE
		-	FDIALOG_MODAL
		-	FDIALOG_SAVEDIR
		*/
		void	SetMode(int m) { mode=m; }
};

/**
An andvanced widget to edit the plain text files or create the new one.
It has a traditional Open, Save & Close menu and simply search engine.

@image html tedit.png

*/
class TextEditor : public FGConnector
{
	public:
		enum 	{ FGMAX_LINE=32768, FGMAX_LINESIZE=512 };
	private:
		class FGLINE
		{
			public:
			unsigned size;         		/* amount of memory allocated */
			unsigned char *text;         /* the characters */
		};

		class FGBUFFER
		{
			public:
			char name[256];              /* name of the file */
			char buffer[FGMAX_LINESIZE];   /* name of the file */
			int  posx, posy;
			int  line;      		        /* cursor position within the line */
			int  lines;                  // all lines
			int  flags;                  /* flags about the buffer state */
			int  top;                    /* top line on the screen */
			int  ovr;
			int  hscroll;                /* offset when screen is scrolled sideways */
			int  sel_line1;               /* line that the selection began at */
			int  sel_pos1;                /* selection offset within the line */
			int  sel_line2;               /* offset in lines from sel_line to cursor */
			int  sel_pos2;                /* selection offset within the line */
			FGBUFFER()
			{
				memset(this,0,sizeof(*this));
			}
		};

	protected:
		Window *Text_EditorPtr;
		Window *OptionsPtr;
		Window *String_SearchPtr;
		int lfonly, font_save;
		FGBUFFER buf;
		FGLINE text[FGMAX_LINE];
		char s[1028];
		int WX, WY;
		char srch_str[33];
		int last_found;
		int _font;
		int TABSIZE;
		int nodraw,inblock,ronly;
		int X, Y, W, H;
		TextEditor **self;
//
		void file(void);
		int  _memcmp(unsigned char *from, unsigned char *co, unsigned kolko, unsigned l);
		int  Find(int from, int count, int size);
		void Clear(void);
		void _goto(void);
		static void SearchProc(GuiEvent *p);
		void search(void);
		static void SetFont(CallBack);
		static void OptionsProc(GuiEvent *p);
		static void file_proc(CallBack);
		void options(void);
		static char * terminate(char *s);
		FGLINE *create_line(char *s, int at);
		void Init(void);
		void SaveAsBuffer2(void);
		void OpenBuffer2(void);
		static void OpenBuffer1(char *s, FileDialog *);
		static void SaveAsBuffer1(char *s, FileDialog *);
		void ShowCursor(int trigger);
		void ShowLine(int scr, char *s);
		int  isempty(char *s);
		void CKey(int k);
		void Key(int k);
		void SaveChanges(void);
		static void Text_EditorProc(GuiEvent *p);
		void Open(char *s);
		void Destruct(void);
	public:
		TextEditor(TextEditor **e=0, char *arg="Untitled", int font=FONT0816, int ink=CGRAY3, int paper=CBLACK);
		virtual ~TextEditor();
		void ShowBuffer(int top=0);
		void NewBuffer(void);
		void OpenBuffer(char *s);
		void ReopenBuffer(void);
		void SaveBuffer(void);
		void SaveAsBuffer(char *s);
		void Close(void);
		void Goto(int);
		/**
		Don't allow any changes on the file. View mode only.
		*/
		void ReadOnlyMode(void) { ronly = 1; Text_EditorPtr->ResetChange(); };
		/**
		Move the editor at the top of all opened windows.
		*/
		void SetFocus(void) { Text_EditorPtr->WindowFocus(); }
		/**
		Append this line at the end of text buffer.
		*/
		int  AppendLine(char *str);
};

/**
A nice widget to color selecting by user for various things.

@image html cdialog.png

*/
class ColorDialog : public FGConnector
{
		Window 	*wnd;
		ColorDialogCallBack fnc;
		int		curr;
		FGPixel color;
		SlideBarV *sl1, *sl2, *sl3;
		int		cd_r, cd_g, cd_b;
		static unsigned int custom_colors[256];
		void FocusTo(int a);
		static void CSliderFnc(CallBack cb);
		static void SetColorsProc(GuiEvent *p);
		void UpdatePalette(void);
	protected:
	public:
		virtual ~ColorDialog()
		{
			if (wnd) if (wnd->GetStatus() & WEXIST) delete wnd;
		}
		/**
		Creates a widget for easy and visual color selecting.
		@param capture a Window caption
		@param f the callback function, it is called with selected color
		@param bgc the background color of widget
		*/
		ColorDialog(char *capture, ColorDialogCallBack f, FGPixel bgc=PM);
		/**
		Saves the current palette to the file. Usable in 8bit color modes only.
		@param name a filename (with the path optionally)
		@return false if a file error is occured
		*/
		static int Import(char *name);
		/**
		Loads the current palette from the file. Usable in 8bit color modes only.
		@param name a filename (with the path optionally)
		@return false if a file error is occured
		*/
		static int Export(char *name);
		/**
		Returns the raw palette data.
		*/
		unsigned int * GetColors(void)
		{
			return custom_colors;
		}
};

#define	FONT_DIALOG_FIXED 		1
#define	FONT_DIALOG_VARIABLE 	2
#define	FONT_DIALOG_ALL			(FONT_DIALOG_FIXED | FONT_DIALOG_VARIABLE)

/**
 A nice widget intended for the font selection by user for various things.

 @image html fntdlg.png

*/
class FontDialog : public FGConnector
{
		Window 	*wnd;
		FontDialogCallBack userfnc;
		int		curr;
		FGListBox *lb;
		static void SetFontProc(GuiEvent *p);
		void ShowFont(void);
	public:
		/**
		Creates a widget for easy and visual color selecting.
		@param f the callback function, it is called with selected color
		@param which you can show to select only fixed, or only proportional or both fonts.
		Predefined consts are: FONT_DIALOG_FIXED, FONT_DIALOG_VARIABLE and FONT_DIALOG_ALL
		@param capture a Window caption
		@param bgc the background color of widget
		*/
		FontDialog(FontDialogCallBack f, int which=FONT_DIALOG_ALL, char *capture="Font selection dialog", FGPixel bgc=PM);
		virtual ~FontDialog()
		{
		}
};

/**
Use FGUpDown to add an up-down control to a Window. Up-down controls consist
of a pair of arrow buttons, such as the arrows that appear in a spin box.
Up-down controls allow users to change the size of a numerical value by clicking
on arrow buttons.
*/
class FGUpDown
{
		PushButton *up, *down;
		EditBox *eb;
		float step;
		static void handler1(CallBack);
		static void handler2(CallBack);

	public:
		FGUpDown(EditBox *eb, float _step);
};

/**
	Provides a container for multiple Enable/Disable of the Control
	in a Window.
	Use this if you want work with some Controls as with a group.
*/
class FGWidgetContainer
{
		void *ptr;
	public:
		FGWidgetContainer();
		~FGWidgetContainer();
		void DisableContainer(class FGWidgetContainer* which=0);
		void EnableContainer(class FGWidgetContainer* which=0);
		void AddToContainer(Control *);
};

#ifdef __BORLANDC__
#pragma option pop /* pop -a switch */
#pragma option pop /* pop -b */
#endif

#endif
